-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb1_board_lib, correlator_lib, filter_lib, fft_lib, rTwoSDF_lib, diag_lib, dp_lib, eth_lib, tech_tse_lib, tr_10GbE_lib, apertif_lib, wpfb_lib, bf_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE unb1_board_lib.unb1_board_pkg.ALL;
USE filter_lib.fil_pkg.ALL;
USE fft_lib.fft_pkg.ALL;    
USE rTwoSDF_lib.rTwoSDFPkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE eth_lib.eth_pkg.ALL;
USE tech_tse_lib.tech_tse_pkg.ALL;
USE tech_tse_lib.tb_tech_tse_pkg.ALL;
USE apertif_lib.apertif_udp_offload_pkg.ALL;
USE wpfb_lib.wpfb_pkg.ALL;   
USE bf_lib.bf_pkg.ALL;

-- Purpose:
-- . Calculate all cross- and autocorrelations of the incoming beamlets
-- Description:
-- . The incpoming beamlets (3x 10GbE from fn_beamformer) are distributed among the 8 FPGAs per board; each FPGA correlates
--   1/8th of the beamlets.
-- Remarks:
-- . The BG version (g_use_bg=TRUE) functionally works.
-- . The 10G-input version (no BG) needs the DDR3 transpose in fn_beamformer to be in place.
--   . Does the WPFB require SOP/EOP at its inputs? If so, we need to (re)set packet boundaries in the incoming stream of beamlet samples. --FIXME
--   . fn_beamformer now outputs 176/256 beamlets but this might be increased to 192/256 (9.6Gbps) again as its 10G issue is solved.
-- . URGENT:
--   . We need 9b filter coefficients before we can synthesize! --FIXME
--     . Note: the 18b coeffieciens *do* fit and meet timing with only BG (without 10G RX stage).
-- . Other:
--   . Synthesis incl. mesh terminals went OK but used 96% of the logic resources.
--   . Keep an eye on the FIXME marks.

ENTITY apertif_unb1_correlator IS
  GENERIC (
    g_design_name             : STRING  := "apertif_unb1_correlator";
    g_use_bg                  : BOOLEAN := FALSE; -- Overridden (TRUE) by TB but still a valid synthesis option; this replaces the 10GbE input stage with block gens.
    g_sim                     : BOOLEAN := FALSE; -- Overridden by TB
    g_sim_fast                : BOOLEAN := TRUE; -- TRUE = fast accumulator model and no inter-channel delay in the correlator output stream.
    g_sim_unb_nr              : NATURAL := 0;
    g_sim_node_nr             : NATURAL := 0;
    g_stamp_date              : NATURAL := 0;    -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time              : NATURAL := 0;    -- Time (HHMMSS)   -- set by QSF
    g_stamp_svn               : NATURAL := 0;    -- SVN revision    -- set by QSF
    g_use_dumb_mesh_terminals : BOOLEAN := FALSE -- Add dumb mesh terminals to the design. Non-functional, for synthesis results only.
  );
  PORT (
    -- GENERAL
    CLK           : IN    STD_LOGIC; -- System Clock
    PPS           : IN    STD_LOGIC; -- System Sync
    WDI           : OUT   STD_LOGIC; -- Watchdog Clear
    INTA          : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB          : INOUT STD_LOGIC; -- FPGA interconnect line
                  
    -- Others     
    VERSION       : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0);
    ID            : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0);
    TESTIO        : INOUT STD_LOGIC_VECTOR(c_unb1_board_aux.testio_w-1 DOWNTO 0);
    
    -- I2C Interface to Sensors
    sens_sc       : INOUT STD_LOGIC;
    sens_sd       : INOUT STD_LOGIC;
  
    -- 1GbE Control Interface
    ETH_clk       : IN    STD_LOGIC;
    ETH_SGIN      : IN    STD_LOGIC;
    ETH_SGOUT     : OUT   STD_LOGIC;

    -- Transceiver clocks
    SA_CLK        : IN  STD_LOGIC; -- SerDes Clock BN-BI / SI_FN
    SB_CLK        : IN  STD_LOGIC; -- SerDes clock FN-BN (tr_mesh)

    -- Mesh Serial I/O
    FN_BN_0_TX    : OUT STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
    FN_BN_0_RX    : IN  STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
    FN_BN_1_TX    : OUT STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
    FN_BN_1_RX    : IN  STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
    FN_BN_2_TX    : OUT STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
    FN_BN_2_RX    : IN  STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
    FN_BN_3_TX    : OUT STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);
    FN_BN_3_RX    : IN  STD_LOGIC_VECTOR(c_unb1_board_tr_mesh.bus_w-1 DOWNTO 0);

    -- Serial I/O: 10GbE receivers 
    SI_FN_0_TX    : OUT   STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);
    SI_FN_0_RX    : IN    STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);
    SI_FN_1_TX    : OUT   STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);
    SI_FN_1_RX    : IN    STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);
    SI_FN_2_TX    : OUT   STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);
    SI_FN_2_RX    : IN    STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);
    SI_FN_3_TX    : OUT   STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);
    SI_FN_3_RX    : IN    STD_LOGIC_VECTOR(c_unb1_board_ci.tr.bus_w-1 DOWNTO 0);

    SI_FN_0_CNTRL : INOUT STD_LOGIC_VECTOR(c_unb1_board_ci.tr.cntrl_w-1 DOWNTO 0);
    SI_FN_1_CNTRL : INOUT STD_LOGIC_VECTOR(c_unb1_board_ci.tr.cntrl_w-1 DOWNTO 0);
    SI_FN_2_CNTRL : INOUT STD_LOGIC_VECTOR(c_unb1_board_ci.tr.cntrl_w-1 DOWNTO 0);
    SI_FN_3_CNTRL : INOUT STD_LOGIC_VECTOR(c_unb1_board_ci.tr.cntrl_w-1 DOWNTO 0);
    SI_FN_RSTN    : OUT   STD_LOGIC := '1' -- ResetN is pulled up in the Vitesse chip, but pulled down again by external 1k resistor.
                                           -- So we need to assign a '1' to it.
  );
END apertif_unb1_correlator;


ARCHITECTURE str OF apertif_unb1_correlator IS

  -- Firmware version x.y
  CONSTANT c_fw_version             : t_unb1_board_fw_version := (0, 4);
  -- Enable block generators
  CONSTANT c_use_phy                : t_c_unb1_board_use_phy  := (1, sel_a_b(g_use_bg, 0, 1), 0, 0, 0, 0, 0, 1);
  
  CONSTANT c_nof_10GbE_streams      : NATURAL := 3;  -- The number of 10G input streams
  CONSTANT c_nof_bf_modules         : NATURAL := c_bf.nof_bf_units; -- We have 4 BF modules here because each sending fn_beamformer design has 4 bf units.
  CONSTANT c_compl_dat_w            : NATURAL := 8;
  
  -- BSN Alignern + FIFO's
  CONSTANT c_block_period           : NATURAL := 186; --256; --FIXME: first block size was 176/256. Now 128/186,18 (non-integer...), assuming 186 is OK.
  CONSTANT c_block_size             : NATURAL := 128; --176
  CONSTANT c_bsn_align_latency      : NATURAL := 3;
  CONSTANT c_bsn_align_sop_timeout  : NATURAL := (c_bsn_align_latency + 1) * c_block_period;  -- wait somewhat more than c_bsn_align_latency periods
  CONSTANT c_bsn_align_xoff_timeout : NATURAL :=  c_bsn_align_latency * 2  * c_block_period;  -- flush factor 2 longer than needed
  CONSTANT c_dp_fifo_size           : NATURAL := (c_bsn_align_latency + 5) * c_block_size;    -- be able to fit blocks for as long as sop time out;
  CONSTANT c_dp_fifo_fill           : NATURAL := c_block_size;
  CONSTANT c_nof_blocks_per_sync    : NATURAL := 800000;
  CONSTANT c_bsn_sync_time_out      : NATURAL := (c_block_period * c_nof_blocks_per_sync * 10)/8; --*10/8 as margin

  -- Re- and Deinterleaver
  CONSTANT c_use_complex            : BOOLEAN := TRUE;
    
  -- Correlator
  CONSTANT c_nof_inputs             : NATURAL := 24;
  CONSTANT c_nof_input_folds        : NATURAL := 1;
  CONSTANT c_nof_input_streams      : NATURAL := c_nof_inputs / pow2(c_nof_input_folds);
  CONSTANT c_nof_pre_mult_folds     : NATURAL := 1; 
  CONSTANT c_conjugate              : BOOLEAN := TRUE; --FIXME: Used for verification/validation (phase shifts) in first stages but should be set to FALSE for final version.
  CONSTANT c_nof_channels           : NATURAL := 64;
  CONSTANT c_integration_period     : NATURAL := c_nof_blocks_per_sync/c_nof_channels; --12207;
  CONSTANT c_nof_visibilities       : NATURAL := (c_nof_inputs*(c_nof_inputs+1))/2;
  -- . Inter Channel Delay:
  --   . With no inter channel output delay, channels are output back-to-back, creating a short full valid burst of 64 channels (12.8 Gbps burst, 158Mbps on average).
  --     We need to set an appropriate number of inter-channel delay cycles for a constant visibility buffer output rate.
  --   . The total accumulator output rate = (150 streams*200MHz / 12207) = 8192.021 channels/sec (157Mbps)
  --   . 8192 channels/sec = 128 * 64 channels * 300 visibilities
  --   . No   inter channel delay: 128 * (19200 valid cycles + 1543300 invalid cycles =)          1562500 cycles/sec = 200M        cycles/sec
  --   . With inter channel delay: 128 * 64 channels * (300 valid cycles + 24114 invalid cycles =)  24414 cycles/sec = 199.999.488 cycles/sec (less than 200M so we're fast enough)
  CONSTANT c_inter_channel_delay    : NATURAL := sel_a_b(g_sim, sel_a_b(g_sim_fast, 0, 4400), 24114); --                    ^^^^^--- 24414= 2*integration period.
                                                                                    -- 4400=~5.5x faster than 157Mbps = 860Mbps burst for sim (but still 157Mbps on average)
  -- WPFB
  CONSTANT c_wpfb_wb_factor         : NATURAL := 1;                    -- = default 1, wideband factor
  CONSTANT c_wpfb_nof_wb_streams    : NATURAL := c_nof_input_streams;  -- = 1, the number of parallel wideband streams. The fi
  CONSTANT c_wpfb_nof_chan          : NATURAL := 1;                    -- = default 0, defines the number of channels (=time-m
  CONSTANT c_wpfb_nof_points        : NATURAL := 64;                   -- = 1024, N point FFT
  CONSTANT c_wpfb_nof_taps          : NATURAL := 8;                    -- = 8 nof taps n the filter
  CONSTANT c_wpfb_coef_w            : NATURAL := 18;                   --FIXME: 9b does not work; needs updated coefficient files.
  CONSTANT c_wpfb_in_dat_w          : NATURAL := 8;                    --FIXME: should be 6b at some point
  CONSTANT c_wpfb_out_dat_w         : NATURAL := 9;                    --FIXME: Could be wider but is currently 9b to avoid the need for quantization: 9b -> mult -> 18b -> accu -> 32b at the corr output.
  CONSTANT c_wpfb_use_separate      : BOOLEAN := FALSE;                -- = false for complex input, true for two real inputs
  CONSTANT c_wpfb_use_reorder       : BOOLEAN := FALSE;

  CONSTANT c_wpfb : t_wpfb  := (c_wpfb_wb_factor, c_wpfb_nof_points, c_wpfb_nof_chan, c_wpfb_nof_wb_streams,
                                c_wpfb_nof_taps, c_wpfb_in_dat_w, 16, c_wpfb_coef_w,
                                c_wpfb_use_reorder, c_wpfb_use_separate, 16, c_wpfb_out_dat_w, 18, 2, true, 56, 2,                      
                                c_fft_pipeline, c_fft_pipeline, c_fil_ppf_pipeline);

  CONSTANT c_wpfb_coefs_file_prefix : STRING  := "mif/coefs_wide1_p64_t8";

  -- Block generator
  CONSTANT c_bg_block_size          : NATURAL := c_nof_channels*pow2(c_nof_input_folds); --128 -- Use 88+40(gap)/128 to match current 176/256 beamlet output of fn_beamformer.
  CONSTANT c_bg_gapsize             : NATURAL := 0; --40

  -- Indicate the integration period with the sync. In the correlator, the
  -- integration period is forced to a minimum of c_nof_visibilities to
  -- allow folding the output onto one stream.
  -- . The sync pulse is only there for the human eye (wave window) -
  --   it is not used by the correlator.
  CONSTANT c_bg_blocks_per_sync     : NATURAL := largest(c_integration_period, c_nof_visibilities);
  CONSTANT c_bg_ctrl                : t_diag_block_gen := (sel_a_b(g_sim, '1', '0'), -- enable: On by default in simulation; MM enable required on hardware.
                                                           '0',                      -- enable_sync        
                                                          TO_UVEC(     c_bg_block_size, c_diag_bg_samples_per_packet_w),
                                                          TO_UVEC(c_bg_blocks_per_sync, c_diag_bg_blocks_per_sync_w),
                                                          TO_UVEC(        c_bg_gapsize, c_diag_bg_gapsize_w),
                                                          TO_UVEC(                   0, c_diag_bg_mem_low_adrs_w),
                                                          TO_UVEC(   c_bg_block_size-1, c_diag_bg_mem_high_adrs_w),
                                                          TO_UVEC(                   0, c_diag_bg_bsn_init_w));

  -- System
  SIGNAL cs_sim                                      : STD_LOGIC;
  SIGNAL xo_clk                                      : STD_LOGIC;
  SIGNAL xo_rst                                      : STD_LOGIC;
  SIGNAL xo_rst_n                                    : STD_LOGIC;
  SIGNAL mm_clk                                      : STD_LOGIC;
  SIGNAL mm_locked                                   : STD_LOGIC;
  SIGNAL mm_rst                                      : STD_LOGIC;
                                                     
  SIGNAL dp_rst                                      : STD_LOGIC;
  SIGNAL dp_clk                                      : STD_LOGIC;  
  SIGNAL sa_rst                                      : STD_LOGIC;
                                                         
  -- PIOs                                                
  SIGNAL pout_wdi                                    : STD_LOGIC;
                                                     
  -- MM WDI override                                    
  SIGNAL reg_wdi_mosi                                : t_mem_mosi;
  SIGNAL reg_wdi_miso                                : t_mem_miso;

  -- MM PPSH
  SIGNAL reg_ppsh_mosi                               : t_mem_mosi;
  SIGNAL reg_ppsh_miso                               : t_mem_miso;
                                                     
  -- MM UniBoard system info                            
  SIGNAL reg_unb_system_info_mosi                    : t_mem_mosi;
  SIGNAL reg_unb_system_info_miso                    : t_mem_miso;
  SIGNAL rom_unb_system_info_mosi                    : t_mem_mosi;
  SIGNAL rom_unb_system_info_miso                    : t_mem_miso;
                                                     
  -- MM UniBoard I2C sens                               
  SIGNAL reg_unb_sens_mosi                           : t_mem_mosi;
  SIGNAL reg_unb_sens_miso                           : t_mem_miso;
                                                     
  -- MM eth1g                                           
  SIGNAL eth1g_tse_clk                               : STD_LOGIC;
  SIGNAL eth1g_mm_rst                                : STD_LOGIC;
  SIGNAL eth1g_tse_mosi                              : t_mem_mosi;  -- ETH TSE MAC registers
  SIGNAL eth1g_tse_miso                              : t_mem_miso;
  SIGNAL eth1g_reg_mosi                              : t_mem_mosi;  -- ETH control and status registers
  SIGNAL eth1g_reg_miso                              : t_mem_miso;
  SIGNAL eth1g_reg_interrupt                         : STD_LOGIC;   -- Interrupt
  SIGNAL eth1g_ram_mosi                              : t_mem_mosi;  -- ETH rx frame and tx frame memory
  SIGNAL eth1g_ram_miso                              : t_mem_miso;

  -- Block generator
  SIGNAL reg_diag_bg_mosi                            : t_mem_mosi;
  SIGNAL reg_diag_bg_miso                            : t_mem_miso;

  -- MM DP offload RX                                          
  SIGNAL reg_dp_offload_rx_hdr_dat_mosi              : t_mem_mosi;
  SIGNAL reg_dp_offload_rx_hdr_dat_miso              : t_mem_miso;  
  
  -- BSN monitors
  SIGNAL reg_bsn_monitor_mosi                        : t_mem_mosi;
  SIGNAL reg_bsn_monitor_miso                        : t_mem_miso;

  -- MM Filterbank
  SIGNAL ram_fil_coefs_mosi                          : t_mem_mosi;
  SIGNAL ram_fil_coefs_miso                          : t_mem_miso;
  
  -- MM  Databuffer                                                   
  SIGNAL ram_diag_data_buf_mosi                      : t_mem_mosi;
  SIGNAL ram_diag_data_buf_miso                      : t_mem_miso;
  SIGNAL reg_diag_data_buf_mosi                      : t_mem_mosi;
  SIGNAL reg_diag_data_buf_miso                      : t_mem_miso; 

  -- MM 1GbE visibility offload TX                                          
  SIGNAL reg_dp_offload_tx_hdr_dat_mosi              : t_mem_mosi;
  SIGNAL reg_dp_offload_tx_hdr_dat_miso              : t_mem_miso;  
                                                         
  -- Interface: 10GbE                                    
  SIGNAL xaui_tx_arr                                 : t_xaui_arr(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL xaui_rx_arr                                 : t_xaui_arr(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL unb_xaui_tx_arr                             : t_unb1_board_xaui_sl_2arr(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL unb_xaui_rx_arr                             : t_unb1_board_xaui_sl_2arr(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL mdio_mdc_arr                                : STD_LOGIC_VECTOR(c_nof_10GbE_streams-1 DOWNTO 0);  
  SIGNAL mdio_mdat_in_arr                            : STD_LOGIC_VECTOR(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL mdio_mdat_oen_arr                           : STD_LOGIC_VECTOR(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL reg_tr_10GbE_mosi                           : t_mem_mosi;
  SIGNAL reg_tr_10GbE_miso                           : t_mem_miso;
  SIGNAL reg_tr_xaui_mosi                            : t_mem_mosi;
  SIGNAL reg_tr_xaui_miso                            : t_mem_miso;
  SIGNAL reg_mdio_mosi_arr                           : t_mem_mosi_arr(c_unb1_board_nof_mdio-1 DOWNTO 0);
  SIGNAL reg_mdio_miso_arr                           : t_mem_miso_arr(c_unb1_board_nof_mdio-1 DOWNTO 0);
                                            
  -- DP offload RX                           
  SIGNAL dp_offload_rx_snk_in_arr                    : t_dp_sosi_arr(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL dp_offload_rx_snk_out_arr                   : t_dp_siso_arr(c_nof_10GbE_streams-1 DOWNTO 0);  
  SIGNAL dp_offload_rx_src_out_arr                   : t_dp_sosi_arr(c_nof_10GbE_streams-1 DOWNTO 0);                            
  SIGNAL dp_offload_rx_src_in_arr                    : t_dp_siso_arr(c_nof_10GbE_streams-1 DOWNTO 0) := (OTHERS=> c_dp_siso_rdy);
  SIGNAL dp_offload_rx_restored_src_out_arr          : t_dp_sosi_arr(c_nof_10GbE_streams-1 DOWNTO 0) := (OTHERS=> c_dp_sosi_rst);
  SIGNAL hdr_fields_out_arr                          : t_slv_1024_arr(c_nof_10GbE_streams-1 DOWNTO 0);
  
  -- BSN Aligner
  SIGNAL dp_fifo_fill_src_in_arr                     : t_dp_siso_arr(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL dp_fifo_fill_src_out_arr                    : t_dp_sosi_arr(c_nof_10GbE_streams-1 DOWNTO 0);
  SIGNAL dp_bsn_align_src_in_arr                     : t_dp_siso_arr(c_nof_10GbE_streams-1 DOWNTO 0) := (OTHERS=> c_dp_siso_rdy);
  SIGNAL dp_bsn_align_src_out_arr                    : t_dp_sosi_arr(c_nof_10GbE_streams-1 DOWNTO 0);
  
  -- BSN monitors
  SIGNAL dp_bsn_monitor_in_siso_arr                  : t_dp_siso_arr(c_nof_10GbE_streams+1+1-1 DOWNTO 0);
  SIGNAL dp_bsn_monitor_in_sosi_arr                  : t_dp_sosi_arr(c_nof_10GbE_streams+1+1-1 DOWNTO 0);

  -- De and Reinterleaver
  SIGNAL dp_deinterleave_snk_in_arr                  : t_dp_sosi_arr(c_nof_10GbE_streams*c_nof_bf_modules-1 DOWNTO 0);
  SIGNAL dp_pipeline_snk_in                          : t_dp_sosi;
  SIGNAL dp_pipeline_src_out                         : t_dp_sosi;
  SIGNAL interleaved_arr                             : t_dp_sosi_arr(c_nof_10GbE_streams*c_nof_bf_modules-1 DOWNTO 0);
  SIGNAL deinterleaved_arr                           : t_dp_sosi_arr(c_nof_inputs-1 DOWNTO 0);
  SIGNAL reinterleave_in_arr                         : t_dp_sosi_arr(c_nof_inputs-1 DOWNTO 0);
  SIGNAL reinterleave_out_arr                        : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);

  --Data buffers
  SIGNAL diag_data_buf_snk_in_arr                    : t_dp_sosi_arr(c_nof_10GbE_streams*c_nof_bf_modules-1 DOWNTO 0);

  -- dp_block_gen
  SIGNAL dp_block_gen_src_out_arr                    : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0); --FIXME should be 3, not 12
  
  -- Filterbank and Correlator  
  SIGNAL wpfb_snk_in_ctrl                            : t_dp_sosi;                    
  SIGNAL wpfb_snk_in_arr                             : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);
  SIGNAL wpfb_src_out_arr                            : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);
  SIGNAL correlator_snk_in_arr                       : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);
  SIGNAL correlator_src_out_arr                      : t_dp_sosi_arr(1-1 DOWNTO 0);
  
  -- 1GbE Visibility Offload
  SIGNAL dp_fifo_sc_snk_in                           : t_dp_sosi;
  SIGNAL dp_repack_data_snk_in                       : t_dp_sosi;
  SIGNAL dp_repack_data_snk_out                      : t_dp_siso;
  SIGNAL apertif_unb1_correlator_vis_offload_snk_in  : t_dp_sosi;
  SIGNAL apertif_unb1_correlator_vis_offload_snk_out : t_dp_siso;
  SIGNAL dp_offload_tx_src_out_arr                   : t_dp_sosi_arr(1-1 DOWNTO 0);
  SIGNAL dp_offload_tx_src_in_arr                    : t_dp_siso_arr(1-1 DOWNTO 0);

  -- Mesh terminals
  SIGNAL tx_serial_2arr                              : t_unb1_board_mesh_sl_2arr;
  SIGNAL rx_serial_2arr                              : t_unb1_board_mesh_sl_2arr;
  SIGNAL mesh_tx_snk_out_2arr                        : t_unb1_board_mesh_siso_2arr;
  SIGNAL mesh_tx_snk_in_2arr                         : t_unb1_board_mesh_sosi_2arr; 
  SIGNAL mesh_rx_src_in_2arr                         : t_unb1_board_mesh_siso_2arr; 
  SIGNAL mesh_rx_src_out_2arr                        : t_unb1_board_mesh_sosi_2arr;

BEGIN

  -----------------------------------------------------------------------------
  -- Block generators
  -----------------------------------------------------------------------------
  gen_block_gen : IF g_use_bg = TRUE GENERATE 
    u_mms_diag_block_gen : ENTITY diag_lib.mms_diag_block_gen
    GENERIC MAP (
      g_nof_streams        => c_nof_input_streams,
      g_buf_dat_w          => 2*c_wpfb_in_dat_w,
      g_buf_addr_w         => ceil_log2(TO_UINT(c_bg_ctrl.samples_per_packet)),
      g_file_name_prefix   => "hex/composite_signals",
      g_diag_block_gen_rst => c_bg_ctrl
    )
    PORT MAP (
      mm_rst           => mm_rst,
      mm_clk           => mm_clk,
    
      dp_rst           => dp_rst,
      dp_clk           => dp_clk,

      reg_bg_ctrl_mosi => reg_diag_bg_mosi,
      reg_bg_ctrl_miso => reg_diag_bg_miso,
    
      out_sosi_arr     => interleaved_arr
    );
  END GENERATE;
  
  gen_10GbE : IF g_use_bg = FALSE GENERATE 
    -----------------------------------------------------------------------------
    -- Interface : 10GbE
    -----------------------------------------------------------------------------
    -- Wire together different types
    gen_wires: FOR i IN 0 TO c_nof_10GbE_streams-1 GENERATE
      unb_xaui_tx_arr(i) <= xaui_tx_arr(i);
      xaui_rx_arr(i)     <= unb_xaui_rx_arr(i);
    END GENERATE;
    
    u_front_io : ENTITY unb1_board_lib.unb1_board_front_io
    GENERIC MAP (
      g_nof_xaui => c_nof_10GbE_streams
    )
    PORT MAP (
      xaui_tx_arr       => unb_xaui_tx_arr,
      xaui_rx_arr       => unb_xaui_rx_arr,
     
      mdio_mdc_arr      => mdio_mdc_arr,
      mdio_mdat_in_arr  => mdio_mdat_in_arr,
      mdio_mdat_oen_arr => mdio_mdat_oen_arr,
    
      -- Serial I/O
      SI_FN_0_TX        => SI_FN_0_TX,
      SI_FN_0_RX        => SI_FN_0_RX,
      SI_FN_1_TX        => SI_FN_1_TX,
      SI_FN_1_RX        => SI_FN_1_RX,
      SI_FN_2_TX        => SI_FN_2_TX,
      SI_FN_2_RX        => SI_FN_2_RX,
    
      SI_FN_0_CNTRL     => SI_FN_0_CNTRL,
      SI_FN_1_CNTRL     => SI_FN_1_CNTRL,
      SI_FN_2_CNTRL     => SI_FN_2_CNTRL,
      SI_FN_3_CNTRL     => SI_FN_3_CNTRL
    );
    
    u_areset_sa_rst : ENTITY common_lib.common_areset
    GENERIC MAP(
      g_rst_level => '1',
      g_delay_len => 4
    )
    PORT MAP(
      clk     => SA_CLK,
      in_rst  => '0',
      out_rst => sa_rst
    );    
    
    u_tr_10GbE: ENTITY tr_10GbE_lib.tr_10GbE
    GENERIC MAP(
      g_sim             => g_sim,
      g_sim_level       => 1,
      g_nof_macs        => c_nof_10GbE_streams,
      g_use_mdio        => TRUE
    )                      
    
    PORT MAP (  
      -- Transceiver PLL reference clock
      tr_ref_clk_156    => SA_CLK, 
      tr_ref_rst_156    => sa_rst,
    
      -- Calibration & reconfig clock
      cal_rec_clk       => mm_clk,
      
      -- MM interface       
      mm_rst            => mm_rst,  
      mm_clk            => mm_clk,
    
      reg_mac_mosi      => reg_tr_10GbE_mosi,
      reg_mac_miso      => reg_tr_10GbE_miso,
    
      xaui_mosi         => reg_tr_xaui_mosi,
      xaui_miso         => reg_tr_xaui_miso,
    
      mdio_mosi_arr     => reg_mdio_mosi_arr(c_nof_10GbE_streams-1 DOWNTO 0),
      mdio_miso_arr     => reg_mdio_miso_arr(c_nof_10GbE_streams-1 DOWNTO 0),
    
      -- DP interface
      dp_rst            => dp_rst,
      dp_clk            => dp_clk,
    
      src_out_arr       => dp_offload_rx_snk_in_arr,
      src_in_arr        => dp_offload_rx_snk_out_arr,
    
      -- Serial XAUI IO
      xaui_tx_arr       => xaui_tx_arr, 
      xaui_rx_arr       => xaui_rx_arr, 
    
      -- MDIO interface
      mdio_rst          => SI_FN_RSTN,
      mdio_mdc_arr      => mdio_mdc_arr,
      mdio_mdat_in_arr  => mdio_mdat_in_arr,
      mdio_mdat_oen_arr => mdio_mdat_oen_arr
    );
    
    -----------------------------------------------------------------------------
    -- RX: dp_offload_rx
    -----------------------------------------------------------------------------
    u_dp_offload_rx : ENTITY dp_lib.dp_offload_rx
    GENERIC MAP (
      g_nof_streams         => c_nof_10GbE_streams,
      g_data_w              => c_xgmii_data_w,
      g_hdr_field_arr       => c_apertif_udp_offload_hdr_field_arr,
      g_remove_crc          => FALSE,
      g_crc_nof_words       => 0 
     )
    PORT MAP (
      mm_rst                => mm_rst,
      mm_clk                => mm_clk,
      
      dp_rst                => dp_rst,
      dp_clk                => dp_clk,
    
      reg_hdr_dat_mosi      => reg_dp_offload_rx_hdr_dat_mosi,
      reg_hdr_dat_miso      => reg_dp_offload_rx_hdr_dat_miso,
    
      snk_in_arr            => dp_offload_rx_snk_in_arr,
      snk_out_arr           => dp_offload_rx_snk_out_arr,
                 
      src_out_arr           => dp_offload_rx_src_out_arr,
      src_in_arr            => (OTHERS=>c_dp_siso_rdy), --dp_offload_rx_src_in_arr,
    
      hdr_fields_out_arr    => hdr_fields_out_arr
      );
    
    gen_restore_bf_out_i : FOR i IN 0 TO c_nof_10GbE_streams-1 GENERATE
      dp_offload_rx_restored_src_out_arr(i).sync <=          sl(hdr_fields_out_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "dp_sync") DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "dp_sync" )));
      dp_offload_rx_restored_src_out_arr(i).bsn  <= RESIZE_UVEC(hdr_fields_out_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "dp_bsn" ) DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "dp_bsn"  )), c_dp_stream_bsn_w);
    
      dp_offload_rx_restored_src_out_arr(i).data  <= dp_offload_rx_src_out_arr(i).data;
      dp_offload_rx_restored_src_out_arr(i).valid <= dp_offload_rx_src_out_arr(i).valid;
      dp_offload_rx_restored_src_out_arr(i).sop   <= dp_offload_rx_src_out_arr(i).sop;
      dp_offload_rx_restored_src_out_arr(i).eop   <= dp_offload_rx_src_out_arr(i).eop;
      dp_offload_rx_restored_src_out_arr(i).err   <= dp_offload_rx_src_out_arr(i).err;
    END GENERATE;

    -----------------------------------------------------------------------------
    -- Re-assign SOP and EOP
    -- . We're receiving 176-word blocks
    -- . WPFB requires blocks of 128 words (2*64, interleaved)
    -----------------------------------------------------------------------------
    gen_dp_block_gen : FOR i IN 0 TO c_nof_10GbE_streams-1 GENERATE
      u_dp_block_gen : ENTITY dp_lib.dp_block_gen
      GENERIC MAP(
        g_use_src_in => FALSE,
        g_nof_data   => 128
      )
      PORT MAP(
        rst        => dp_rst,
        clk        => dp_clk,
       
        snk_in     => dp_offload_rx_restored_src_out_arr(i),
        src_out    => dp_block_gen_src_out_arr(i)
      );    
    END GENERATE;

    -----------------------------------------------------------------------------
    -- RX: BSN alignment
    -----------------------------------------------------------------------------
    gen_dp_fifo_fill: FOR i IN 0 TO c_nof_10GbE_streams-1 GENERATE
      u_dp_fifo_fill : ENTITY dp_lib.dp_fifo_fill
      GENERIC MAP (
        g_data_w         => c_xgmii_data_w,
        g_bsn_w          => c_dp_stream_bsn_w,
        g_channel_w      => c_dp_stream_channel_w,
        g_use_bsn        => TRUE,
        g_use_channel    => TRUE,
        g_use_error      => TRUE,
        g_use_sync       => TRUE,
        g_fifo_fill      => c_dp_fifo_fill,
        g_fifo_size      => c_dp_fifo_size
      )
      PORT MAP (
        rst       => dp_rst,
        clk       => dp_clk,
    
--        snk_out   => dp_offload_rx_src_in_arr(i),
        snk_in    => dp_block_gen_src_out_arr(i),
    
        src_in    => dp_fifo_fill_src_in_arr(i),
        src_out   => dp_fifo_fill_src_out_arr(i)
      );
    END GENERATE;
    
    u_dp_bsn_align : ENTITY dp_lib.dp_bsn_align
    GENERIC MAP (
      g_block_size           => c_block_size,
      g_block_period         => c_block_period, --FIXME: NOT EVEN USED INTERNALLY!!!
      g_nof_input            => c_nof_10GbE_streams,
      g_xoff_timeout         => c_bsn_align_xoff_timeout,
      g_sop_timeout          => c_bsn_align_sop_timeout,
      g_bsn_latency          => c_bsn_align_latency,
      g_bsn_request_pipeline => 2
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,
    
      snk_out_arr => dp_fifo_fill_src_in_arr,
      snk_in_arr  => dp_fifo_fill_src_out_arr,
    
      src_in_arr  => dp_bsn_align_src_in_arr,
      src_out_arr => dp_bsn_align_src_out_arr
    );
 
    -----------------------------------------------------------------------------
    -- Rewire: for each input stream, extract the 4 concatenated BF output streams
    --         and assign them to the interleaved array. 
    -----------------------------------------------------------------------------
    p_extract_bf_streams : PROCESS(dp_bsn_align_src_out_arr)
    BEGIN
      FOR i IN 0 TO c_nof_10GbE_streams-1 LOOP
        FOR j IN 0 TO c_nof_bf_modules-1 LOOP
          interleaved_arr(i*c_nof_bf_modules + j)      <= dp_bsn_align_src_out_arr(i); -- SOSI ctrl
          interleaved_arr(i*c_nof_bf_modules + j).data <= (OTHERS=>'0');
          interleaved_arr(i*c_nof_bf_modules + j).im   <= RESIZE_DP_DSP_DATA(dp_bsn_align_src_out_arr(i).data( (c_nof_complex*j+1)*c_compl_dat_w+c_compl_dat_w-1 DOWNTO c_nof_complex*j*c_compl_dat_w+c_compl_dat_w));
          interleaved_arr(i*c_nof_bf_modules + j).re   <= RESIZE_DP_DSP_DATA(dp_bsn_align_src_out_arr(i).data( (c_nof_complex*j+1)*c_compl_dat_w-1               DOWNTO c_nof_complex*j*c_compl_dat_w));
        END LOOP;
      END LOOP;
    END PROCESS;
  END GENERATE;

  -----------------------------------------------------------------------------
  -- RX: BSN monitors at several stages in the stream
  -----------------------------------------------------------------------------
  u_dp_bsn_monitor : ENTITY dp_lib.mms_dp_bsn_monitor
  GENERIC MAP (
    g_nof_streams        => 5,
    g_sync_timeout       => c_bsn_sync_time_out,
    g_log_first_bsn      => TRUE
  )
  PORT MAP (
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    reg_mosi    => reg_bsn_monitor_mosi,
    reg_miso    => reg_bsn_monitor_miso,
    
    dp_rst      => dp_rst,
    dp_clk      => dp_clk,
    in_siso_arr => dp_bsn_monitor_in_siso_arr,
    in_sosi_arr => dp_bsn_monitor_in_sosi_arr
  );

  --      0) Monitor the BSN-aligned output
  dp_bsn_monitor_in_sosi_arr(0) <= dp_bsn_align_src_out_arr(0);
  dp_bsn_monitor_in_siso_arr(0) <= dp_bsn_align_src_in_arr(0);
  --      1) Monitor the rewired/reinterleaver stage output / WPFB input
  dp_bsn_monitor_in_sosi_arr(1) <= wpfb_snk_in_arr(0);
  dp_bsn_monitor_in_siso_arr(1) <= c_dp_siso_rdy;
  --      2) Monitor the WPFB output / correlator input
  dp_bsn_monitor_in_sosi_arr(2) <= wpfb_src_out_arr(0);
  dp_bsn_monitor_in_siso_arr(2) <= c_dp_siso_rdy;
  --      3) Monitor the correlator output / visibility offload input  
  dp_bsn_monitor_in_sosi_arr(3) <= correlator_src_out_arr(0);
  dp_bsn_monitor_in_siso_arr(3) <= c_dp_siso_rdy;
  --      3) Monitor the correlator output / visibility offload input  
  dp_bsn_monitor_in_sosi_arr(4) <= dp_offload_tx_src_out_arr(0);
  dp_bsn_monitor_in_siso_arr(4) <= dp_offload_tx_src_in_arr(0);

  -----------------------------------------------------------------------------
  -- RX: Data buffers
  -----------------------------------------------------------------------------
  diag_data_buf_snk_in_arr <= interleaved_arr;

  u_diag_data_buffer : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams  => c_nof_10GbE_streams*c_nof_bf_modules, --12
    g_data_w       => 2*c_wpfb_in_dat_w, --16b (complex, 2*8b)
    g_buf_nof_data => 256, -- Used data width is still 32b so 256*32=8192=1 M9K
    g_buf_use_sync => TRUE,
    g_data_type    => e_complex
  )
  PORT MAP (
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,

    ram_data_buf_mosi => ram_diag_data_buf_mosi,
    ram_data_buf_miso => ram_diag_data_buf_miso,
    reg_data_buf_mosi => reg_diag_data_buf_mosi, --FIXME address space is wrong (1 bit)
    reg_data_buf_miso => reg_diag_data_buf_miso,

    in_sync           => diag_data_buf_snk_in_arr(0).sync, 
    in_sosi_arr       => diag_data_buf_snk_in_arr
  );

  -----------------------------------------------------------------------------
  -- Beamlet routing from 10GbE inputs T2..T0 to unfolder correlator inputs 23..0: ***48 VISIBILITY VERSION***
  -- . Beamlet indices between parentheses ()
  -- . Stream indices between brackets []            
  --                                                   
  -- T2(B6,B7)--[11]--deinter-[1]--(T2,B7)--[23]  [23]--(T2,B7)--[1]-reinter-[11]--(T1,B7),(T2,B7)--unfold-[1]--(T2,B7)-->[23]
  --                         \[0]--(T2,B6)--[22]  [15]--(T1,B7)--[0]/                                     \[0]--(T1,B7)-->[22]
  --   (B4,B5)--[10]--deinter-[1]--(T2,B5)--[21]  [07]--(T0,B7)--[1]-reinter-[10]--(T2,B6),(T0,B7)--unfold-[1]--(T0,B7)-->[21]
  --                         \[0]--(T2,B4)--[20]  [22]--(T2,B6)--[0]/                                     \[0]--(T2,B6)-->[20]
  --   (B2,B3)--[09]--deinter-[1]--(T2,B3)--[19]  [14]--(T1,B6)--[1]-reinter-[09]--(T0,B6),(T1,B6)--unfold-[1]--(T1,B6)-->[19]
  --                         \[0]--(T2,B2)--[18]  [06]--(T0,B6)--[0]/                                     \[0]--(T0,B6)-->[18]
  --   (B0,B1)--[08]--deinter-[1]--(T2,B1)--[17]  [21]--(T2,B5)--[1]-reinter-[08]--(T1,B5),(T2,B5)--unfold-[1]--(T2,B5)-->[17]
  --                         \[0]--(T2,B0)--[16]  [13]--(T1,B5)--[0]/                                     \[0]--(T1,B5)-->[16]
  -- T1(B6,B7)--[07]--deinter-[1]--(T1,B7)--[15]  [05]--(T0,B5)--[1]-reinter-[07]--(T2,B4),(T0,B5)--unfold-[1]--(T0,B5)-->[15]
  --                         \[0]--(T1,B6)--[14]  [20]--(T2,B4)--[0]/                                     \[0]--(T2,B4)-->[14]
  --   (B4,B5)--[06]--deinter-[1]--(T1,B5)--[13]  [12]--(T1,B4)--[1]-reinter-[06]--(T0,B4),(T1,B4)--unfold-[1]--(T1,B4)-->[13]
  --                         \[0]--(T1,B4)--[12]\/[04]--(T0,B4)--[0]/                                     \[0]--(T0,B4)-->[12]
  --   (B2,B3)--[05]--deinter-[1]--(T1,B3)--[11]/\[19]--(T2,B3)--[1]-reinter-[05]--(T1,B3),(T2,B3)--unfold-[1]--(T2,B3)-->[11]
  --                         \[0]--(T1,B2)--[10]  [11]--(T1,B3)--[0]/                                     \[0]--(T1,B3)-->[10]
  --   (B0,B1)--[04]--deinter-[1]--(T1,B1)--[09]  [03]--(T0,B3)--[1]-reinter-[04]--(T2,B2),(T0,B3)--unfold-[1]--(T0,B3)-->[09]
  --                         \[0]--(T1,B0)--[08]  [18]--(T2,B2)--[0]/                                     \[0]--(T2,B2)-->[08]
  -- T0(B6,B7)--[03]--deinter-[1]--(T0,B7)--[07]  [10]--(T1,B2)--[1]-reinter-[03]--(T0,B2),(T1,B2)--unfold-[1]--(T1,B2)-->[07]
  --                         \[0]--(T0,B6)--[06]  [02]--(T0,B2)--[0]/                                     \[0]--(T0,B2)-->[06]
  --   (B4,B5)--[02]--deinter-[1]--(T0,B5)--[05]  [17]--(T2,B1)--[1]-reinter-[02]--(T1,B1),(T2,B1)--unfold-[1]--(T2,B1)-->[05]
  --                         \[0]--(T0,B4)--[04]  [09]--(T1,B1)--[0]/                                     \[0]--(T1,B1)-->[04]
  --   (B2,B3)--[01]--deinter-[1]--(T0,B3)--[03]  [01]--(T0,B1)--[1]-reinter-[01]--(T2,B0),(T0,B1)--unfold-[1]--(T0,B1)-->[03]
  --                         \[0]--(T0,B2)--[02]  [16]--(T2,B0)--[0]/                                     \[0]--(T2,B0)-->[02]
  --   (B0,B1)--[00]--deinter-[1]--(T0,B1)--[01]  [08]--(T1,B0)--[1]-reinter-[00]--(T0,B0),(T1,B0)--unfold-[1]--(T1,B0)-->[01]
  --                         \[0]--(T0,B0)--[00]  [00]--(T0,B0)--[0]/                                     \[0]--(T0,B0)-->[00]
  --             ^                           ^     ^                          ^                                            ^
  --             interleaved_arr             |     reinterleave_in_arr        |                                            Unfolded correlator inputs
  --                                         deinterleaved_arr                reinterleave_out_arr,wpfb_out_arr,correlator_snk_in_arr
  ----------------------------------------------------------------------------- 

  -----------------------------------------------------------------------------
  -- Beamlet routing from 10GbE inputs T2..T0 to unfolder correlator inputs 23..0: ***FINAL MESH VERSION***
  -- . Beamlet indices between parentheses ()                    The exact indices depend on FPGA index--FIXME
  -- . Stream indices between brackets []                       /                 To do: Also BSN align these 24 streams --FIXME
  --                                              [<-----------/------------>]   /
  -- T2(B6,B7)--[11]--deinter-[1]--(T2,B7)--[23]--MESH_TX[3][2]  MESH_RX[3][2]--[23]--(T23,B0)--[1]-reinter-[11]--(T22,B0),(T23,B0)--unfold-[1]--(T23,B0)-->[23]
  --                         \[0]--(T2,B6)--[22]--MESH_TX[3][1]  MESH_RX[3][1]--[22]--(T22,B0)--[0]/                                       \[0]--(T22,B0)-->[22]
  --   (B4,B5)--[10]--deinter-[1]--(T2,B5)--[21]--MESH_TX[3][0]  MESH_RX[3][0]--[21]--(T21,B0)--[1]-reinter-[10]--(T20,B0),(T21,B0)--unfold-[1]--(T21,B0)-->[21]
  --                         \[0]--(T2,B4)--[20]--MESH_TX[2][2]  MESH_RX[2][2]--[20]--(T10,B0)--[0]/                                       \[0]--(T20,B0)-->[20]
  --   (B2,B3)--[09]--deinter-[1]--(T2,B3)--[19]--MESH_TX[2][1]  MESH_RX[2][1]--[19]--(T19,B0)--[1]-reinter-[09]--(T18,B0),(T19,B0)--unfold-[1]--(T19,B0)-->[19]
  --                         \[0]--(T2,B2)--[18]--MESH_TX[2][0]  MESH_RX[2][0]--[18]--(T18,B0)--[0]/                                       \[0]--(T18,B0)-->[18]
  --   (B0,B1)--[08]--deinter-[1]--(T2,B1)--[17]--MESH_TX[1][2]  MESH_RX[1][2]--[17]--(T17,B0)--[1]-reinter-[08]--(T16,B0),(T17,B0)--unfold-[1]--(T17,B0)-->[17]
  --                         \[0]--(T2,B0)--[16]--MESH_TX[1][1]  MESH_RX[1][1]--[16]--(T16,B0)--[0]/                                       \[0]--(T16,B0)-->[16]
  -- T1(B6,B7)--[07]--deinter-[1]--(T1,B7)--[15]--MESH_TX[1][0]  MESH_RX[1][0]--[15]--(T15,B0)--[1]-reinter-[07]--(T14,B0),(T15,B0)--unfold-[1]--(T15,B0)-->[15]
  --                         \[0]--(T1,B6)--[14]--MESH_TX[0][2]  MESH_RX[0][2]--[14]--(T14,B0)--[0]/                                       \[0]--(T14,B0)-->[14]
  --   (B4,B5)--[06]--deinter-[1]--(T1,B5)--[13]--MESH_TX[0][1]  MESH_RX[0][1]--[13]--(T13,B0)--[1]-reinter-[06]--(T12,B0),(T13,B0)--unfold-[1]--(T13,B0)-->[13]
  --                         \[0]--(T1,B4)--[12]--MESH_TX[0][0]\/MESH_RX[0][0]--[12]--(T12,B0)--[0]/                                       \[0]--(T12,B0)-->[12]
  --   (B2,B3)--[05]--deinter-[1]--(T1,B3)--[11]--MESH_TX[3][2]/\MESH_RX[3][2]--[11]--(T11,B0)--[1]-reinter-[05]--(T10,B0),(T11,B0)--unfold-[1]--(T11,B0)-->[11]
  --                         \[0]--(T1,B2)--[10]--MESH_TX[3][1]  MESH_RX[3][1]--[10]--(T10,B0)--[0]/                                       \[0]--(T10,B0)-->[10]
  --   (B0,B1)--[04]--deinter-[1]--(T1,B1)--[09]--MESH_TX[3][0]  MESH_RX[3][0]--[09]--(T09,B0)--[1]-reinter-[04]--(T08,B0),(T09,B0)--unfold-[1]--(T09,B0)-->[09]
  --                         \[0]--(T1,B0)--[08]--MESH_TX[2][2]  MESH_RX[2][2]--[08]--(T08,B0)--[0]/                                       \[0]--(T08,B0)-->[08]
  -- T0(B6,B7)--[03]--deinter-[1]--(T0,B7)--[07]--MESH_TX[2][1]  MESH_RX[2][1]--[07]--(T07,B0)--[1]-reinter-[03]--(T06,B0),(T07,B0)--unfold-[1]--(T07,B0)-->[07]
  --                         \[0]--(T0,B6)--[06]--MESH_TX[2][0]  MESH_RX[2][0]--[06]--(T06,B0)--[0]/                                       \[0]--(T06,B0)-->[06]
  --   (B4,B5)--[02]--deinter-[1]--(T0,B5)--[05]--MESH_TX[1][2]  MESH_RX[1][2]--[05]--(T05,B0)--[1]-reinter-[02]--(T04,B0),(T05,B0)--unfold-[1]--(T05,B0)-->[05]
  --                         \[0]--(T0,B4)--[04]--MESH_TX[1][1]  MESH_RX[1][1]--[04]--(T04,B0)--[0]/                                       \[0]--(T04,B0)-->[04]
  --   (B2,B3)--[01]--deinter-[1]--(T0,B3)--[03]--MESH_TX[1][0]  MESH_RX[1][0]--[03]--(T03,B0)--[1]-reinter-[01]--(T02,B0),(T03,B0)--unfold-[1]--(T03,B0)-->[03]
  --                         \[0]--(T0,B2)--[02]--MESH_TX[0][2]  MESH_RX[0][2]--[02]--(T02,B0)--[0]/                                       \[0]--(T02,B0)-->[02]
  --   (B0,B1)--[00]--deinter-[1]--(T0,B1)--[01]--MESH_TX[0][1]  MESH_RX[0][1]--[01]--(T01,B0)--[1]-reinter-[00]--(T00,B0),(T01,B0)--unfold-[1]--(T01,B0)-->[01] -- Note that we're also TX'ing the locally RX'ed telescope inputs T0..T2 as
  --                         \[0]--(T0,B0)--[00]--MESH_TX[0][0]  MESH_RX[0][0]--[00]--(T00,B0)--[0]/                                       \[0]--(T00,B0)-->[00] -- it appears to be easier with the re-interleaving and BSN alignment.
  --             ^                           ^            ^                                                  ^                                               ^
  --             interleaved_arr             |            mesh_tx_snk_in_2arr                                |                                               Unfolded correlator inputs
  --                                         deinterleaved_arr                                               reinterleave_out_arr,wpfb_out_arr,correlator_snk_in_arr
  ----------------------------------------------------------------------------- 

  gen_deinterleave_snk_in_arr: IF g_use_dumb_mesh_terminals=FALSE GENERATE
    dp_pipeline_snk_in <= interleaved_arr(0);
    dp_deinterleave_snk_in_arr <= interleaved_arr;
  END GENERATE;

  gen_deinterleave : FOR i IN 0 TO c_nof_10GbE_streams*c_nof_bf_modules-1 GENERATE
    u_deinterleave : ENTITY dp_lib.dp_deinterleave
    GENERIC MAP (
      g_nof_out      => 2,
      g_block_size   => 1,
      g_dat_w        => c_nof_complex*c_wpfb_in_dat_w,
      g_use_ctrl     => FALSE,
      g_align_out    => TRUE,
      g_use_complex  => c_use_complex
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,
      
      snk_in      => dp_deinterleave_snk_in_arr(i), 
      src_out_arr => deinterleaved_arr(2*i+1 DOWNTO 2*i)
    );  
  END GENERATE;

  p_reorder_array : PROCESS(deinterleaved_arr)
  BEGIN
    FOR i IN 0 TO 2*c_nof_bf_modules-1 LOOP
      FOR j IN 0 TO c_nof_10GbE_streams-1 LOOP
        reinterleave_in_arr(i*c_nof_10GbE_streams + j) <= deinterleaved_arr(j*2*c_nof_bf_modules + i); 
      END LOOP;
    END LOOP;   
  END PROCESS;
  
  gen_reinterleave : FOR i IN 0 TO c_nof_10GbE_streams*c_nof_bf_modules-1 GENERATE
    u_reinterleave : ENTITY dp_lib.dp_reinterleave
    GENERIC MAP (
      g_nof_in            => 2,
      g_deint_block_size  => 1,
      g_nof_out           => 1,
      g_inter_block_size  => 1,
      g_use_ctrl          => FALSE,
      g_dat_w             => c_nof_complex*c_wpfb_in_dat_w,
      g_use_complex       => c_use_complex,
      g_align_out         => TRUE
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,
      snk_in_arr  => reinterleave_in_arr(2*i+1 DOWNTO 2*i), 
      src_out_arr => reinterleave_out_arr(i DOWNTO i)
    );
  END GENERATE;
  
  u_st_ctrl_pipe : ENTITY dp_lib.dp_pipeline
  GENERIC MAP(
    g_pipeline  => 3  -- 0 for wires, > 0 for registers, 
  )
  PORT MAP (
    rst          => dp_rst,
    clk          => dp_clk,

    snk_in       => dp_pipeline_snk_in,
    src_out      => dp_pipeline_src_out
  );

  p_add_st_ctrl : PROCESS(reinterleave_out_arr, dp_pipeline_src_out)
  BEGIN
    FOR i IN 0 TO c_nof_input_streams-1 LOOP
      wpfb_snk_in_arr(i)      <= dp_pipeline_src_out; -- SOSI ctrl
      wpfb_snk_in_arr(i).data <= reinterleave_out_arr(i).data;
      wpfb_snk_in_arr(i).im   <= reinterleave_out_arr(i).im;
      wpfb_snk_in_arr(i).re   <= reinterleave_out_arr(i).re;
    END LOOP;
  END PROCESS;

  -----------------------------------------------------------------------------
  -- Preliminary mesh terminals to explore resource usage of final design:
  -- . With g_use_dumb_mesh_terminals=TRUE, the mesh terminals:
  --   . Transmit 21/24 beamlets to the other 7 FPGAs via the mesh.
  --   . Receive 21/24 beamlets from the other FPGAs via the mesh.
  -- . These are DUMB terminals because this will synthesize but not 
  --   functionally work because the lack of alignment and proper rewiring.
  --   . FOR EXPLORATION OF RESOURCE UTILIZATION ONLY!
  -----------------------------------------------------------------------------
  gen_terminals_mesh : if g_use_dumb_mesh_terminals=TRUE GENERATE
    mesh_tx_snk_in_2arr(0)(0) <= interleaved_arr(0);
    mesh_tx_snk_in_2arr(0)(1) <= interleaved_arr(1);
    mesh_tx_snk_in_2arr(0)(2) <= interleaved_arr(2);
    mesh_tx_snk_in_2arr(1)(0) <= interleaved_arr(3);
    mesh_tx_snk_in_2arr(1)(1) <= interleaved_arr(4);
    mesh_tx_snk_in_2arr(1)(2) <= interleaved_arr(5);
    mesh_tx_snk_in_2arr(2)(0) <= interleaved_arr(6);
    mesh_tx_snk_in_2arr(2)(1) <= interleaved_arr(7);
    mesh_tx_snk_in_2arr(2)(2) <= interleaved_arr(8);
    mesh_tx_snk_in_2arr(3)(0) <= interleaved_arr(9);
    mesh_tx_snk_in_2arr(3)(1) <= interleaved_arr(10);
    mesh_tx_snk_in_2arr(3)(2) <= interleaved_arr(11);

    dp_deinterleave_snk_in_arr(0)  <= mesh_rx_src_out_2arr(0)(0);
    dp_deinterleave_snk_in_arr(1)  <= mesh_rx_src_out_2arr(0)(1);
    dp_deinterleave_snk_in_arr(2)  <= mesh_rx_src_out_2arr(0)(2);
    dp_deinterleave_snk_in_arr(3)  <= mesh_rx_src_out_2arr(1)(0);
    dp_deinterleave_snk_in_arr(4)  <= mesh_rx_src_out_2arr(1)(1);
    dp_deinterleave_snk_in_arr(5)  <= mesh_rx_src_out_2arr(1)(2);
    dp_deinterleave_snk_in_arr(6)  <= mesh_rx_src_out_2arr(2)(0);
    dp_deinterleave_snk_in_arr(7)  <= mesh_rx_src_out_2arr(2)(1);
    dp_deinterleave_snk_in_arr(8)  <= mesh_rx_src_out_2arr(2)(2);
    dp_deinterleave_snk_in_arr(9)  <= mesh_rx_src_out_2arr(3)(0);
    dp_deinterleave_snk_in_arr(10) <= mesh_rx_src_out_2arr(3)(1);
    dp_deinterleave_snk_in_arr(11) <= mesh_rx_src_out_2arr(3)(2);

    dp_pipeline_snk_in  <= mesh_rx_src_out_2arr(0)(0);

    u_terminals_mesh : ENTITY unb1_board_lib.unb1_board_terminals_mesh
    GENERIC MAP (
      g_sim                  => g_sim,
      g_sim_level            => 1,
      -- System
      g_node_type            => e_fn,
      g_nof_bus              => 4,
      -- User
      g_usr_use_complex      => c_use_complex,
      g_usr_data_w           => 16,
      g_usr_frame_len        => 176, --FIXME: current fn_beamformer output. Mismatch with expected (by WPFB) back-to-back 128 word blocks.
      g_usr_nof_streams      => 4, -- Actually 3 but has to be 4 to prevent errors.
      -- Phy
      g_phy_nof_serial       => 3,
      g_phy_gx_mbps          => 6250,
      g_phy_rx_fifo_size     => 176,
      g_phy_ena_reorder      => TRUE,
      -- Tx
      g_use_tx               => TRUE,
      g_tx_input_use_fifo    => TRUE,
      -- Rx
      g_use_rx               => TRUE, 
      g_rx_output_use_fifo   => TRUE,
      g_rx_output_fifo_size  => 176,  --FIXME: current fn_beamformer output. Mismatch with expected (by WPFB) back-to-back 128 word blocks.
      g_rx_output_fifo_fill  => 0,
      g_rx_timeout_w         => 0,
      -- Monitoring
      g_mon_select           => 0
    )
    PORT MAP (
      chip_id                => ID(2 DOWNTO 0),
      
      mm_rst                 => mm_rst,
      mm_clk                 => mm_clk,
      dp_rst                 => dp_rst,
      dp_clk                 => dp_clk,
--      dp_sync                => dp_pps,
      tr_clk                 => SB_clk,
      cal_clk                => mm_clk,  
      
      -- Received beamlets from other nodes 
--      rx_usr_siso_2arr       => mesh_rx_src_in_2arr,
      rx_usr_sosi_2arr       => mesh_rx_src_out_2arr,

      -- Beamlets transmitted to other nodes
--        tx_usr_siso_2arr       => mesh_tx_snk_out_2arr,
      tx_usr_sosi_2arr       => mesh_tx_snk_in_2arr,
      
      -- Mesh interface level (4 nodes)(4 lanes)
      tx_serial_2arr         => tx_serial_2arr, 
      rx_serial_2arr         => rx_serial_2arr 
      
      -- MM Control
--      reg_tr_nonbonded_mosi  => reg_tr_nonbonded_mosi,
--      reg_tr_nonbonded_miso  => reg_tr_nonbonded_miso,
--      reg_diagnostics_mosi   => reg_diagnostics_mosi,
--      reg_diagnostics_miso   => reg_diagnostics_miso,
--      ram_diag_data_buf_mosi => ram_mesh_diag_data_buf_mosi,
--      ram_diag_data_buf_miso => ram_mesh_diag_data_buf_miso
    );

    u_unb1_board_mesh_io : ENTITY unb1_board_lib.unb1_board_mesh_io
    PORT MAP (
      tx_serial_2arr => tx_serial_2arr,
      rx_serial_2arr => rx_serial_2arr,
      
      -- Serial I/O
      FN_BN_0_TX     => FN_BN_0_TX,
      FN_BN_0_RX     => FN_BN_0_RX,
      FN_BN_1_TX     => FN_BN_1_TX,
      FN_BN_1_RX     => FN_BN_1_RX,
      FN_BN_2_TX     => FN_BN_2_TX,
      FN_BN_2_RX     => FN_BN_2_RX,
      FN_BN_3_TX     => FN_BN_3_TX,
      FN_BN_3_RX     => FN_BN_3_RX
    );
  END GENERATE; -- g_use_dumb_mesh_terminals

  -----------------------------------------------------------------------------
  -- WPFB
  -----------------------------------------------------------------------------
  u_wpfb_unit : ENTITY wpfb_lib.wpfb_unit
  GENERIC MAP(
    g_wpfb              => c_wpfb,     
    g_stats_ena         => FALSE,
    g_use_bg            => FALSE,    
    g_file_index_arr    => array_init(0, 128, 1),    
    g_coefs_file_prefix => c_wpfb_coefs_file_prefix 
  )
  PORT MAP(
    dp_rst             => dp_rst,
    dp_clk             => dp_clk,
    mm_rst             => mm_rst,
    mm_clk             => mm_clk,  
    ram_fil_coefs_mosi => ram_fil_coefs_mosi,
    ram_fil_coefs_miso => ram_fil_coefs_miso,
    ram_st_sst_mosi    => c_mem_mosi_rst,
    ram_st_sst_miso    => OPEN, 
    reg_bg_ctrl_mosi   => c_mem_mosi_rst,
    reg_bg_ctrl_miso   => OPEN,
    ram_bg_data_mosi   => c_mem_mosi_rst,
    ram_bg_data_miso   => OPEN,
    in_sosi_arr        => wpfb_snk_in_arr,     
    out_sosi_arr       => wpfb_src_out_arr
  ); 

  -----------------------------------------------------------------------------
  -- Workaround for simulation: the first WPFB output block is not usable! --FIXME
  -- Forward only blocks with BSN 5+
  -----------------------------------------------------------------------------
  p_bsn_filter : PROCESS(wpfb_src_out_arr)
  BEGIN
    correlator_snk_in_arr <= (OTHERS=>c_dp_sosi_rst);
    IF TO_UINT(wpfb_src_out_arr(0).bsn)>=5 THEN
      correlator_snk_in_arr <= wpfb_src_out_arr;
    END IF;
  END PROCESS;

  -----------------------------------------------------------------------------
  -- Stream recorder to record the WPFB output stream to a file
  -----------------------------------------------------------------------------
  u_dp_stream_rec_play_wpfb_unit : ENTITY dp_lib.dp_stream_rec_play
  GENERIC MAP (
    g_sim            => TRUE,
    g_pass_through   => FALSE,
    g_rec_not_play   => TRUE,
    g_rec_play_file  => "../../../../applications/apertif/designs/apertif_unb1_correlator/tb/rec/wpfb_src_out_arr0.rec",
    g_record_invalid => FALSE
  )
  PORT MAP (
    dp_clk  => dp_clk,
    snk_in  => wpfb_src_out_arr(0),
    snk_out => OPEN,
    src_out => OPEN,
    src_in  => c_dp_siso_rdy
  );

  -----------------------------------------------------------------------------
  -- Correlator
  -----------------------------------------------------------------------------
  u_correlator : ENTITY correlator_lib.correlator
  GENERIC MAP (
    g_sim                 => g_sim,
    g_nof_input_streams   => c_nof_input_streams,
    g_input_unfold_factor => c_nof_input_folds,
    g_nof_pre_mult_folds  => c_nof_pre_mult_folds,
    g_data_w              => c_wpfb_out_dat_w,
    g_conjugate           => c_conjugate,
    g_nof_channels        => c_nof_channels,
    g_integration_period  => c_integration_period,
    g_inter_channel_delay => c_inter_channel_delay,
    g_sim_fast_acc_model  => g_sim AND g_sim_fast
  )
  PORT MAP (
    clk         => dp_clk,
    rst         => dp_rst,

    snk_in_arr  => correlator_snk_in_arr,
    src_out_arr => correlator_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Wire complex fields into data field
  -- . Real & imag. Real part is at the MS end so is put on the (1GbE) line first.
  -- . For fast simulation, we output all channels back-to-back and write them to
  --   a file (dp_stream_rec_play). In that case we don't want to use (and overload) 
  --   the visibility offload stage.
  -----------------------------------------------------------------------------
  gen_wires_complex : IF g_sim=FALSE or g_sim_fast=FALSE GENERATE
    p_wires_complex : PROCESS(correlator_src_out_arr)
    BEGIN
      dp_fifo_sc_snk_in <= correlator_src_out_arr(0);
      dp_fifo_sc_snk_in.data(64-1 DOWNTO 0) <= correlator_src_out_arr(0).re(64/2-1 DOWNTO 0) & correlator_src_out_arr(0).im(64/2-1 DOWNTO 0);
    END PROCESS;
  END GENERATE;

  -----------------------------------------------------------------------------
  -- 64b -> 32b 
  -----------------------------------------------------------------------------
  u_dp_fifo_sc : ENTITY dp_lib.dp_fifo_sc
  GENERIC MAP (
    g_data_w         => 64, 
    g_use_ctrl       => TRUE,
    g_use_bsn        => TRUE,
    g_use_channel    => TRUE,
    g_bsn_w          => 64,
    g_channel_w      => 16,
    g_fifo_size      => 300 + 20
  )
  PORT MAP (
    rst         => dp_rst,
    clk         => dp_clk,
  
    snk_in      => dp_fifo_sc_snk_in,

    src_in      => dp_repack_data_snk_out,
    src_out     => dp_repack_data_snk_in
  );

  u_dp_repack_data : ENTITY dp_lib.dp_repack_data
  GENERIC MAP (
    g_in_dat_w      => 64,
    g_in_nof_words  => 1,
    g_out_dat_w     => 32,
    g_out_nof_words => 2
  )
  PORT MAP (
    rst            => dp_rst,
    clk            => dp_clk,

    snk_out        => dp_repack_data_snk_out,
    snk_in         => dp_repack_data_snk_in,
                   
    src_in         => apertif_unb1_correlator_vis_offload_snk_out,
    src_out        => apertif_unb1_correlator_vis_offload_snk_in
  );

  -----------------------------------------------------------------------------
  -- 1GbE visivility offload
  -----------------------------------------------------------------------------
  u_apertif_unb1_correlator_vis_offload : ENTITY work.apertif_unb1_correlator_vis_offload
  PORT MAP (
    mm_rst                          => mm_rst,
    mm_clk                          => mm_clk,
    
    dp_rst                          => dp_rst,
    dp_clk                          => dp_clk,
     
    reg_dp_offload_tx_hdr_dat_mosi  => reg_dp_offload_tx_hdr_dat_mosi,
    reg_dp_offload_tx_hdr_dat_miso  => reg_dp_offload_tx_hdr_dat_miso,

    snk_in                          => apertif_unb1_correlator_vis_offload_snk_in,
    snk_out                         => apertif_unb1_correlator_vis_offload_snk_out,

    src_out_arr                     => dp_offload_tx_src_out_arr,
    src_in_arr                      => dp_offload_tx_src_in_arr,

    ID                              => ID
  );

  -----------------------------------------------------------------------------
  -- Stream recorder to record the correlator output stream to a file
  -- . The data buffer can only take snapshots.
  -----------------------------------------------------------------------------
  gen_dp_stream_rec_play_correlator : IF g_sim=TRUE GENERATE
    u_dp_stream_rec_play_correlator : ENTITY dp_lib.dp_stream_rec_play
    GENERIC MAP (
      g_sim            => TRUE,
      g_pass_through   => FALSE,
      g_rec_not_play   => TRUE,
      g_rec_play_file  => "../../../../applications/apertif/designs/apertif_unb1_correlator/tb/rec/correlator_src_out_arr0.rec",
      g_record_invalid => FALSE
    )
    PORT MAP (
      dp_clk  => dp_clk,
      snk_in  => correlator_src_out_arr(0),
      snk_out => OPEN,
      src_out => OPEN,
      src_in  => c_dp_siso_rdy
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl : ENTITY unb1_board_lib.ctrl_unb1_board
  GENERIC MAP (
    g_sim                     => g_sim,
    g_sim_flash_model         => FALSE,
    g_design_name             => g_design_name,
    g_stamp_date              => g_stamp_date,
    g_stamp_time              => g_stamp_time, 
    g_stamp_svn               => g_stamp_svn, 
    g_fw_version              => c_fw_version,
    g_mm_clk_freq             => c_unb1_board_mm_clk_freq_50M,
    g_udp_offload             => TRUE,
    g_udp_offload_nof_streams => 1,
    g_use_phy                 => c_use_phy,
    g_aux                     => c_unb1_board_aux,
    g_dp_clk_use_pll          => TRUE 
  )
  PORT MAP (
    -- Clock an reset signals
    cs_sim                   => cs_sim,
    xo_clk                   => xo_clk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk                   => mm_clk,
    mm_locked                => mm_locked,
    mm_rst                   => mm_rst,

    epcs_clk                 => '0',
    
    dp_rst                   => dp_rst,
    dp_clk                   => dp_clk,
    dp_pps                   => OPEN,
    dp_rst_in                => dp_rst,
    dp_clk_in                => dp_clk,
    
    -- Toggle WDI
    pout_wdi                 => pout_wdi,

    -- MM buses
    -- . Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,
    
    -- . System_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 
    
    -- . UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,    
    
    -- . PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,

    -- eth1g
    eth1g_tse_clk            => eth1g_tse_clk,   -- 125 MHz from xo_clk PLL in SOPC system
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,
     -- eth1g UDP streaming ports
    udp_tx_sosi_arr          => dp_offload_tx_src_out_arr,
    udp_tx_siso_arr          => dp_offload_tx_src_in_arr,
       
    -- FPGA pins
    -- . General
    CLK                      => CLK,
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    -- . Others
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,
    -- . I2C Interface to Sensors
    sens_sc                  => sens_sc,
    sens_sd                  => sens_sd,        
    -- . 1GbE Control Interface
    ETH_clk                  => ETH_clk,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  );

  -----------------------------------------------------------------------------
  -- MM master
  -----------------------------------------------------------------------------
  u_mmm : ENTITY work.mmm_apertif_unb1_correlator
  GENERIC MAP (
    g_sim           => g_sim,
    g_sim_unb_nr    => g_sim_unb_nr,
    g_sim_node_nr   => g_sim_node_nr,
    g_wpfb          => c_wpfb,
    g_hdr_field_arr => c_apertif_udp_offload_hdr_field_arr
   )
  PORT MAP(  
    xo_clk                         => xo_clk,       
    xo_rst_n                       => xo_rst_n,     
    xo_rst                         => xo_rst,       

    mm_rst                         => mm_rst,
    mm_clk                         => mm_clk,       
    mm_locked                      => mm_locked,    

    -- PIOs
    pout_wdi                       => pout_wdi,

    -- Manual WDI override
    reg_wdi_mosi                   => reg_wdi_mosi,
    reg_wdi_miso                   => reg_wdi_miso,

    -- system_info
    reg_unb_system_info_mosi       => reg_unb_system_info_mosi,
    reg_unb_system_info_miso       => reg_unb_system_info_miso,
    rom_unb_system_info_mosi       => rom_unb_system_info_mosi,
    rom_unb_system_info_miso       => rom_unb_system_info_miso, 

    -- UniBoard I2C sensors
    reg_unb_sens_mosi              => reg_unb_sens_mosi,
    reg_unb_sens_miso              => reg_unb_sens_miso, 
 
    -- PPSH
    reg_ppsh_mosi                  => reg_ppsh_mosi,
    reg_ppsh_miso                  => reg_ppsh_miso, 

    -- Block generator
    reg_diag_bg_mosi               => reg_diag_bg_mosi,
    reg_diag_bg_miso               => reg_diag_bg_miso,
  
    -- 10 GbE
    reg_tr_10GbE_mosi              => reg_tr_10GbE_mosi,
    reg_tr_10GbE_miso              => reg_tr_10GbE_miso,
    reg_tr_xaui_mosi               => reg_tr_xaui_mosi,
    reg_tr_xaui_miso               => reg_tr_xaui_miso,
    reg_mdio_mosi_arr              => reg_mdio_mosi_arr,
    reg_mdio_miso_arr              => reg_mdio_miso_arr,

    -- DP offload RX    
    reg_dp_offload_rx_hdr_dat_mosi => reg_dp_offload_rx_hdr_dat_mosi,
    reg_dp_offload_rx_hdr_dat_miso => reg_dp_offload_rx_hdr_dat_miso,

    reg_bsn_monitor_mosi           => reg_bsn_monitor_mosi,
    reg_bsn_monitor_miso           => reg_bsn_monitor_miso,
    
    -- . Data buffers
    reg_diag_data_buf_mosi         => reg_diag_data_buf_mosi,
    reg_diag_data_buf_miso         => reg_diag_data_buf_miso,
    ram_diag_data_buf_mosi         => ram_diag_data_buf_mosi,
    ram_diag_data_buf_miso         => ram_diag_data_buf_miso,    

    -- 1GbE visibility offload TX    
    reg_dp_offload_tx_hdr_dat_mosi => reg_dp_offload_tx_hdr_dat_mosi,
    reg_dp_offload_tx_hdr_dat_miso => reg_dp_offload_tx_hdr_dat_miso,

    -- Filter coefficients
    ram_fil_coefs_mosi             => ram_fil_coefs_mosi,
    ram_fil_coefs_miso             => ram_fil_coefs_miso,
  
    -- eth1g
    eth1g_tse_clk                  => eth1g_tse_clk,
    eth1g_mm_rst                   => eth1g_mm_rst,
    eth1g_tse_mosi                 => eth1g_tse_mosi,
    eth1g_tse_miso                 => eth1g_tse_miso,
    eth1g_reg_mosi                 => eth1g_reg_mosi,
    eth1g_reg_miso                 => eth1g_reg_miso,
    eth1g_reg_interrupt            => eth1g_reg_interrupt,
    eth1g_ram_mosi                 => eth1g_ram_mosi,
    eth1g_ram_miso                 => eth1g_ram_miso
  );

END str;

