-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose:
-- . Wrapper containing dp_offload_tx and header field assignments.
-- Description:
-- . The actual header field definition is located in fn_beamformer_udp_offload_pkg,
--   because the same header field definition is also used by dp_offload_rx in 
--   design apertif_fn_qdcor.

LIBRARY IEEE, common_lib, unb1_board_lib, tech_tse_lib, dp_lib, apertif_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL; 
USE unb1_board_lib.unb1_board_pkg.ALL;
USE unb1_board_lib.unb1_board_peripherals_pkg.ALL;
USE tech_tse_lib.tech_tse_pkg.ALL;
USE tech_tse_lib.tb_tech_tse_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE apertif_lib.apertif_udp_offload_pkg.ALL;

ENTITY apertif_unb1_fn_beamformer_udp_offload IS
  PORT (
    mm_rst                         : IN  STD_LOGIC;
    mm_clk                         : IN  STD_LOGIC; 
                                   
    dp_rst                         : IN  STD_LOGIC;
    dp_clk                         : IN  STD_LOGIC;  
                                   
    snk_in_arr                     : IN  t_dp_sosi_arr(3 DOWNTO 0); -- 4 BF unit outputs           
                                   
    src_out_arr                    : OUT t_dp_sosi_arr(0 DOWNTO 0); 
    src_in_arr                     : IN  t_dp_siso_arr(0 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
                                   
    ID                             : IN  STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0);
                                   
    reg_dp_offload_tx_mosi         : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_dp_offload_tx_miso         : OUT t_mem_miso := c_mem_miso_rst;
                                    
    reg_dp_offload_tx_hdr_dat_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_dp_offload_tx_hdr_dat_miso : OUT t_mem_miso := c_mem_miso_rst
  );
END apertif_unb1_fn_beamformer_udp_offload;


ARCHITECTURE wrap OF apertif_unb1_fn_beamformer_udp_offload IS

  CONSTANT c_nof_offload_streams       : NATURAL := 1; 

  -- Override ('1') only the Ethernet fields so we can use MM defaults there.
  CONSTANT c_hdr_field_ovr_init : STD_LOGIC_VECTOR(c_apertif_udp_offload_nof_hdr_fields-1 DOWNTO 0) := "101"&"111111111101"&"0011"&"100";

  CONSTANT c_bf_out_compl_dat_w        : NATURAL := 8;
  CONSTANT c_data_w                    : NATURAL := 64; -- 4 BF units * 8b complex

  CONSTANT c_header_overhead_bytes     : NATURAL := field_slv_out_len(c_apertif_udp_offload_hdr_field_arr) / c_byte_w;

  CONSTANT c_max_frame_len             : NATURAL := 9018;
  CONSTANT c_frame_nof_words           : NATURAL := (c_max_frame_len * c_byte_w ) / c_data_w;

  CONSTANT c_max_udp_payload_len       : NATURAL := c_max_frame_len - c_header_overhead_bytes - c_network_eth_crc_len;
  CONSTANT c_max_udp_payload_nof_words : NATURAL := (c_max_udp_payload_len * c_byte_w) / c_data_w;

  CONSTANT c_max_nof_words_per_block   : NATURAL := 256;
  CONSTANT c_min_nof_words_per_block   : NATURAL := 1;
  CONSTANT c_max_nof_blocks_per_packet : NATURAL := c_max_udp_payload_nof_words/c_min_nof_words_per_block;

  CONSTANT c_def_nof_words_per_block   : NATURAL := 176; --176+7(header words)/256 + 1 CRC word = 184/256 64b words@200MHz = 9.2Gbps user rate. Incl. inter-frame gap (96b) and preamble (64b): 9.325Gbps.
  CONSTANT c_def_nof_blocks_per_packet : NATURAL := 1;

  SIGNAL hdr_fields_in_arr             : t_slv_1024_arr(c_nof_offload_streams-1 DOWNTO 0);

  SIGNAL id_backplane                  : STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
  SIGNAL id_chip                       : STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);

  SIGNAL dp_fifo_sc_snk_in             : t_dp_sosi;
  SIGNAL dp_offload_tx_snk_in_arr      : t_dp_sosi_arr(c_nof_offload_streams-1 DOWNTO 0);
  SIGNAL dp_offload_tx_snk_out_arr     : t_dp_siso_arr(c_nof_offload_streams-1 DOWNTO 0);

BEGIN

  --------------------------------------------------------------------------------------- 
  -- 4 BF unit outputs: 2*8b complex; concatenate these into one 64b word.
  ---------------------------------------------------------------------------------------
  p_connect : PROCESS(snk_in_arr)
  BEGIN
    dp_fifo_sc_snk_in <= snk_in_arr(0); -- Control 
    dp_fifo_sc_snk_in.data(c_data_w-1 DOWNTO 0) <= snk_in_arr(3).im(c_bf_out_compl_dat_w-1 DOWNTO 0) & snk_in_arr(3).re(c_bf_out_compl_dat_w-1 DOWNTO 0) &
                                                   snk_in_arr(2).im(c_bf_out_compl_dat_w-1 DOWNTO 0) & snk_in_arr(2).re(c_bf_out_compl_dat_w-1 DOWNTO 0) &
                                                   snk_in_arr(1).im(c_bf_out_compl_dat_w-1 DOWNTO 0) & snk_in_arr(1).re(c_bf_out_compl_dat_w-1 DOWNTO 0) &
                                                   snk_in_arr(0).im(c_bf_out_compl_dat_w-1 DOWNTO 0) & snk_in_arr(0).re(c_bf_out_compl_dat_w-1 DOWNTO 0);
  END PROCESS;

  --------------------------------------------------------------------------------------- 
  -- FIFO to provide a flow controllable buffer while dp_offload_tx inserts headers
  ---------------------------------------------------------------------------------------
  u_dp_fifo_sc : ENTITY dp_lib.dp_fifo_sc
  GENERIC MAP (
    g_data_w         => c_data_w,
    g_bsn_w          => 64,
    g_use_sync       => TRUE,
    g_use_bsn        => TRUE,
    g_fifo_size      => 1024
  )
  PORT MAP (
    rst         => dp_rst,
    clk         => dp_clk,

    snk_out     => OPEN,
    snk_in      => dp_fifo_sc_snk_in,

    src_in      => dp_offload_tx_snk_out_arr(0),
    src_out     => dp_offload_tx_snk_in_arr(0)
  );

  --------------------------------------------------------------------------------------- 
  -- dp_offload_tx
  ---------------------------------------------------------------------------------------
  u_dp_offload_tx : ENTITY dp_lib.dp_offload_tx
  GENERIC MAP (
    g_nof_streams               => c_nof_offload_streams,
    g_data_w                    => c_data_w,
    g_use_complex               => FALSE,
    g_max_nof_words_per_block   => c_max_nof_words_per_block,
    g_def_nof_words_per_block   => c_def_nof_words_per_block,
    g_max_nof_blocks_per_packet => c_max_nof_blocks_per_packet,
    g_def_nof_blocks_per_packet => c_def_nof_blocks_per_packet,
    g_hdr_field_arr             => c_apertif_udp_offload_hdr_field_arr,
    g_hdr_field_sel             => c_hdr_field_ovr_init
   )
  PORT MAP (
    mm_rst                => mm_rst,
    mm_clk                => mm_clk,
    
    dp_rst                => dp_rst,
    dp_clk                => dp_clk,

    reg_mosi              => reg_dp_offload_tx_mosi,
    reg_miso              => reg_dp_offload_tx_miso,

    reg_hdr_dat_mosi      => reg_dp_offload_tx_hdr_dat_mosi,
    reg_hdr_dat_miso      => reg_dp_offload_tx_hdr_dat_miso,

    snk_in_arr            => dp_offload_tx_snk_in_arr,
    snk_out_arr           => dp_offload_tx_snk_out_arr,

    src_out_arr           => src_out_arr,
    src_in_arr            => src_in_arr,

    hdr_fields_in_arr     => hdr_fields_in_arr
  );

  --------------------------------------------------------------------------------------- 
  -- Extract the chip and backplane numbers from ID
  ---------------------------------------------------------------------------------------
  id_backplane <= RESIZE_UVEC(ID(7 DOWNTO 3), c_byte_w);
  id_chip      <= RESIZE_UVEC(ID(2 DOWNTO 0), c_byte_w);

  --------------------------------------------------------------------------------------- 
  -- Wire the hardwired header fields to DP signals and ID
  ---------------------------------------------------------------------------------------
  gen_slv_hard_fields : FOR i IN 0 TO c_nof_offload_streams-1 GENERATE
    hdr_fields_in_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "eth_src_mac"     ) DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "eth_src_mac"     )) <= x"00228608" & id_backplane & id_chip;
    hdr_fields_in_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "udp_src_port"    ) DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "udp_src_port"    )) <= x"D0" & ID;
    hdr_fields_in_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "udp_dst_port"    ) DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "udp_dst_port"    )) <= x"D0" & ID;
    hdr_fields_in_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "ip_src_addr"     ) DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "ip_src_addr"     )) <= x"0A63" & id_backplane & INCR_UVEC(id_chip, 1);
                                                                                                                     
    hdr_fields_in_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "dp_sync"         ) DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "dp_sync"         )) <= slv(dp_offload_tx_snk_in_arr(i).sync);
    hdr_fields_in_arr(i)(field_hi(c_apertif_udp_offload_hdr_field_arr, "dp_bsn"          ) DOWNTO field_lo(c_apertif_udp_offload_hdr_field_arr, "dp_bsn"          )) <=     dp_offload_tx_snk_in_arr(i).bsn(63 DOWNTO 0); 
  END GENERATE;

END wrap;
