###############################################################################
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

"""
Purpose:
. Instantiate mms_diag_block_gen.vhd in myHDL and generate a new VHDL wrapper
  around it.
Description:
. This a first test to integrate myHDL-generated code in the existing RadioHDL
  environment. DS used following directory for reference/examples:
  . RadioHDL/trunk/tools/oneclick/prestudy/myHDL
Remarks:
. The following ports of the VHDL entity are not connected (left at default):
  . en_sync
  . all MM buses
  . usr_siso_arr, usr_sosi_arr
  . out_siso_arr
. MyHDL does not support our types such as t_mem_mosi and t_dp_sosi. This is 
  not a problem as we include user-defined code to map our types onto myHDL
  compatible types. This code is literally pasted into the file and therefor
  falls outside of the myHDL scope.
  . We use separate DP signals for now, but we may also combine them into a
    Python record as this is fully supported my myHDL.
Usage:
. python gen_mms_diag_block_gen.py
. VHDL files are generated in the same dir.
"""

from myhdl import *

###############################################################################
# Constants/generics
# . NOTE: use extra quotes around strings!
###############################################################################
G_NOF_STREAMS = 4
G_BUF_DAT_W = 16
G_BUF_ADDR_W = 256
G_FILE_NAME_PREFIX = '"hex/bg_data"' # Note: extra quotes
G_DIAG_BLOCK_GEN_RST = '"FIXME"'     # Note: extra quotes

###############################################################################
# MyHDL function definition
# . VHDL entity I/O ports
#   . We need to declare the same I/O ports as exist on the VHDL entity. We
#     declare them as function arguments.
# . VHDL Architecture
#   . Not described here (pass). We set the .vhdl_code attribute later on to 
#     insert literal VHDL.
###############################################################################
def myhdl_mms_diag_block_gen(g_nof_streams, g_buf_dat_w, g_buf_addr_w, g_file_name_prefix, g_diag_block_gen_rst, mm_rst,mm_clk,dp_rst,dp_clk,out_sosi_arr):

    @always(mm_rst,mm_clk,dp_rst,dp_clk)
    def wrap():
        # Nothing to do here.
        pass

    # Important: explicitely say these outputs are driven, otherwise myHDL makes
    # these inputs.
    out_sosi_arr.driven = "wire"

    # Workaround for lack of user defined type support: define the substrings here
    out_sosi_arr_substrings = ""
    for i in range(g_nof_streams):
        out_sosi_arr_substring_data  = "    out_sosi_arr(%d).data(%d-1 DOWNTO 0) => out_sosi_arr(%d).data,\n" %(i,g_buf_dat_w,i)
        out_sosi_arr_substring_sop   = "    out_sosi_arr(%d).sop   => out_sosi_arr(%d).sop,\n" %(i,i)
        out_sosi_arr_substring_eop   = "    out_sosi_arr(%d).eop   => out_sosi_arr(%d).eop,\n" %(i,i)
        out_sosi_arr_substring_valid = "    out_sosi_arr(%d).valid => out_sosi_arr(%d).valid,\n" %(i,i)

        out_sosi_arr_substrings+=(out_sosi_arr_substring_data+\
                                  out_sosi_arr_substring_sop+\
                                  out_sosi_arr_substring_eop+\
                                  out_sosi_arr_substring_valid)
    # Remove the last ',' (and the last newline char) from the string
    out_sosi_arr_substrings=out_sosi_arr_substrings[:-2]
    return wrap

###############################################################################
# Architecture VHDL code: Instantiate mms_diag_block_gen
# . Note the $ prefix for all myHDL function (myhdl_mms_diag_block_gen)
#   arguments.
###############################################################################
myhdl_mms_diag_block_gen.vhdl_code =\
"""
  u_mms_diag_block_gen : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP (
    g_nof_streams        => $g_nof_streams,
    g_buf_dat_w          => $g_buf_dat_w,
    g_buf_addr_w         => $g_buf_addr_w,
    g_file_name_prefix   => $g_file_name_prefix,
    g_diag_block_gen_rst => $g_diag_block_gen_rst
  )
  PORT MAP (
    mm_rst           => $mm_rst,
    mm_clk           => $mm_clk,
  
    dp_rst           => $dp_rst,
    dp_clk           => $dp_clk,

$out_sosi_arr_substrings
  );

"""

###############################################################################
# Convert to VHDL
###############################################################################
mm_clk = Signal(bool(0))
mm_rst = Signal(bool(0))
dp_clk = Signal(bool(0))
dp_rst = Signal(bool(0))
out_sosi_arr = Signal(intbv(0)[8:]) #FIXME

toVHDL(myhdl_mms_diag_block_gen, G_NOF_STREAMS, G_BUF_DAT_W, G_BUF_ADDR_W, G_FILE_NAME_PREFIX, G_DIAG_BLOCK_GEN_RST,mm_rst,mm_clk,dp_rst,dp_clk,out_sosi_arr)




