###############################################################################
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

# Purpose:
# . Generate HEX files for the 4 block generators, emulating BF output.
# Description:
# . The BG output block size is 256 words, without any gaps;
# . The BGs emulate transposed BF output (b=beamlet, t=time);
# . The 256-word blocks are 2 interleaved streams of 128 words:
#   . stream 0) b0t0,b1t0, .. b0t127,b1t127 (repeated)
#   . stream 1) b2t0,b3t0, .. b2t127,b3t127 (repeated)
#   . stream 2) b4t0,b5t0, .. b4t127,b5t127 (repeated)
#   . stream 3) b6t0,b7t0, .. b6t127,b7t127 (repeated)
# . Each beamlet signal repeats after 128 samples and consists of an integer 
#   number of sine wave periods.
# . A full integration period spans 800000 timesamples:
#   . 6250 * 128 timesamples
#   . 3125 * 256-word BG output blocks.

from common import *
from mem_init_file import list_to_hex
import matplotlib
matplotlib.use('TkAgg')
from common_dsp import *
import matplotlib.pyplot as plt
import numpy as np
from scipy.fftpack import fft,ifft, fftfreq, fftshift

NOF_INPUTS = 12 # 4 transposed BF output streams
NOF_INPUT_SIGNALS = 24 # 4 streams * 2 interleaved signals
COMPLEX_WIDTH = 8
NOF_WORDS_PER_BLOCK = 64

PATH = os.environ['RADIOHDL']+'/applications/apertif/designs/apertif_unb1_fn_bf_emu/src/hex'
FILENAME = "bg_data"

###############################################################################
# Create composite waves
###############################################################################
N = NOF_WORDS_PER_BLOCK
t=range(N)

# sample spacing
T = 1.0 / N

x = np.linspace(0.0, N*T, N)

AMPL_MAX = 127
CHANNELS = [1,5,12,17,21,25,28,34,41,47,54,55,60]
NOF_CHANNELS = len(CHANNELS)
input_signals_concat_inter
input_signals = []
for input_signal_nr in range(NOF_INPUT_SIGNALS):
    phase_shift_deg = input_signal_nr
    channel_signals = []
    for bin_nr in CHANNELS:
        phase_shift_rad = math.radians(phase_shift_deg)
        ampl=AMPL_MAX/NOF_CHANNELS
        signal = ampl * np.exp( (bin_nr) * 1.j * (2.0*np.pi*x+(phase_shift_rad/bin_nr)) )

        channel_signals.append( signal )
    composite_signal=numpy.sum(channel_signals, axis=0)
    input_signals.append(composite_signal)

s=input_signals[1]

###############################################################################
# Convert the float values to n-bits
###############################################################################
s_bits = []
for fword in s:
    re_signed = to_signed(fword.real, COMPLEX_WIDTH)
    im_signed = to_signed(fword.imag, COMPLEX_WIDTH)

    s_bits.append( complex(re_signed, im_signed) )

s = numpy.array(s_bits)
print s

###############################################################################
# Plot the signal
###############################################################################
plt.plot(t, s.real, 'b-', t, s.imag, 'r--')
plt.legend(('real', 'imaginary'))
plt.show()

###############################################################################
# Plot FFT
###############################################################################
yf = fft(s)

xf = fftfreq(N, T)
xf = range(NOF_WORDS_PER_BLOCK)
xf = fftshift(xf)
yplot = fftshift(yf)
plt.bar(xf, 1.0/N * np.abs(yplot))
plt.grid()
plt.show()

###############################################################################
# Convert complex floats to concatenated integers
###############################################################################
input_signals_concat = []
for input_signal in input_signals:
    input_signals_concat.append( concat_complex(input_signal, COMPLEX_WIDTH) )

###############################################################################
# 48-visibility system: reorder the 24 streams as described in
# apertif_unb1_correlator.vhd
###############################################################################
input_signals_concat_reordered = 24*[None]

input_signals_concat_reordered[23] = input_signals_concat[23]
input_signals_concat_reordered[15] = input_signals_concat[22]
input_signals_concat_reordered[ 7] = input_signals_concat[21]
input_signals_concat_reordered[22] = input_signals_concat[20]
input_signals_concat_reordered[14] = input_signals_concat[19]
input_signals_concat_reordered[ 6] = input_signals_concat[18]
input_signals_concat_reordered[21] = input_signals_concat[17]
input_signals_concat_reordered[13] = input_signals_concat[16]
input_signals_concat_reordered[ 5] = input_signals_concat[15]
input_signals_concat_reordered[20] = input_signals_concat[14]
input_signals_concat_reordered[12] = input_signals_concat[13]
input_signals_concat_reordered[ 4] = input_signals_concat[12]
input_signals_concat_reordered[19] = input_signals_concat[11]
input_signals_concat_reordered[11] = input_signals_concat[10]
input_signals_concat_reordered[ 3] = input_signals_concat[ 9]
input_signals_concat_reordered[18] = input_signals_concat[ 8]
input_signals_concat_reordered[10] = input_signals_concat[ 7]
input_signals_concat_reordered[ 2] = input_signals_concat[ 6]
input_signals_concat_reordered[17] = input_signals_concat[ 5]
input_signals_concat_reordered[ 9] = input_signals_concat[ 4]
input_signals_concat_reordered[ 1] = input_signals_concat[ 3]
input_signals_concat_reordered[16] = input_signals_concat[ 2]
input_signals_concat_reordered[ 8] = input_signals_concat[ 1]
input_signals_concat_reordered[ 0] = input_signals_concat[ 0]

###############################################################################
# Interleave 2 lists into one
###############################################################################
input_signals_concat_inter = []
for i in range(NOF_INPUTS):
    input_signals_concat_inter.append( interleave([input_signals_concat_reordered[2*i],input_signals_concat_reordered[2*i+1]]) )

###############################################################################
# Use this list for each block generator
###############################################################################
bg_lists = input_signals_concat_inter

###############################################################################
# Write the HEX files
###############################################################################
MEM_WIDTH  = COMPLEX_WIDTH*2
MEM_DEPTH  = 2*NOF_WORDS_PER_BLOCK # We're interleaving, hence twice the depth

for input_nr in range(NOF_INPUTS):
    list_to_hex( bg_lists[input_nr], PATH+"/"+FILENAME+'_'+str(input_nr)+".hex", MEM_WIDTH, MEM_DEPTH)

