#! /usr/bin/env python
###############################################################################
#
# Copyright (C) 2013
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

# Purpose:  Test LCU - loop back via DUT
# Description:
#
#   LCU = Local Control Unit (models the PC)
#   DUT = Device Under Test
#
# flow diagram:
#
#   LCU = FN 0               1GbE link     DUT = FN 1
#   BG --> dp_offload_tx ----------------> dp_offload_rx ---\
#                                                           |   DP loopback in DUT
#   DB <-- dp_offload_rx <---------------- dp_offload_tx <--/
#
# Usage when starting from the beginning:
#   > cd $RADIOHDL
#   > rm -rf build/
#   > python tools/oneclick/base/modelsim_config.py
#   > python tools/oneclick/base/quartus_config.py
#   > run_sopc unb1 compaan_unb1_dp_offload
#   > run_modelsim unb1 &
#     . lp compaan_unb1_dp_offload
#     . mk all
#     double click tb_compaan_unb1_dp_offload_2ins simulation icon
#     . as 8
#     . run 2 us
#   > python tc_compaan_unb1_dp_offload_2ins.py --unb 0 --fn 0,1 --sim
#     . run 100 us  (or run -a)
#
# Usage when rebuilding the simulation:
#   > cd $RADIOHDL
#   > rm -rf build/
#   > python tools/oneclick/base/modelsim_config.py
#   > run_modelsim unb1 &
#   ...etc
#
# Usage on hardware:
#   > python tc_compaan_unb1_dp_offload_2ins.py --unb 0 --fn 0,1

from common import *

import test_case
import node_io
import pi_common
import pi_dp_offload_tx
import pi_dp_offload_tx_hdr_dat_unb_dp_offload
import pi_dp_offload_tx_hdr_ovr_unb_dp_offload
import time
import sys
import os

import pi_diag_block_gen
import pi_eth

###############################################################################
# Setup
###############################################################################

NOF_STREAMS           = 1
STREAM_ID             = 0
NOF_WORDS_PER_BLOCK   = 10
NOF_BLOCKS_PER_PACKET = 1


tc = test_case.Testcase('TB - ', '')
io = node_io.NodeIO(tc.nodeImages, tc.base_ip)

tc.set_result('PASSED')
tc.append_log(3, '>>>')
tc.append_log(1, '>>> Title : Test case for design unb_dp_offload. Targets: %s' % tc.unb_nodes_string())
tc.append_log(3, '>>>')
tc.append_log(3, '')

eth           = pi_eth.PiEth(tc, io) 
dpotx         = pi_dp_offload_tx.PiDpOffloadTx(tc, io, nof_inst=NOF_STREAMS)
dpotx_hdr_dat = pi_dp_offload_tx_hdr_dat_unb_dp_offload.PiDpOffloadTxHdrDatUnbDpOffload(tc, io, nof_inst=NOF_STREAMS)
dpotx_hdr_ovr = pi_dp_offload_tx_hdr_ovr_unb_dp_offload.PiDpOffloadTxHdrOvrUnbDpOffload(tc, io, nof_inst=NOF_STREAMS)

###############################################################################
# Task
###############################################################################

# Enable the UDP port on the 1GbE RX side for stream 0
eth.write_udp_port_en(STREAM_ID)

# Write the dp_offload_tx settings
registers = [('nof_words_per_block', NOF_WORDS_PER_BLOCK), ('nof_blocks_per_packet', NOF_BLOCKS_PER_PACKET)]
dpotx.write(inst_nrs=STREAM_ID, registers=registers, regmap=dpotx.regmap)


# Set dst MAC of node[0] to MAC of node[1] 
dpotx_hdr_dat.write(node_nrs=tc.nodeNrs[0], inst_nrs=tc.gpNumbers, registers=[('eth_dst_mac', 0x2286080000 + tc.nodeNrs[1])], regmap=dpotx_hdr_dat.regmap)
# Set dst MAC of node[1] to MAC of node[0] 
dpotx_hdr_dat.write(node_nrs=tc.nodeNrs[1], inst_nrs=tc.gpNumbers, registers=[('eth_dst_mac', 0x2286080000 + tc.nodeNrs[0])], regmap=dpotx_hdr_dat.regmap)

# Now override the eth_dst_mac field so it is no longer read from the data path but from the MM register we've just written to.
dpotx_hdr_ovr.write(inst_nrs=tc.gpNumbers, registers=[('eth_dst_mac', 1)], regmap=dpotx_hdr_ovr.regmap)    

###############################################################################
# end
###############################################################################
tc.set_section_id('')
tc.append_log(0, '>>> Test Case result: %s' % tc.get_result())

sys.exit(tc.get_result())
