-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb1_board_lib, dp_lib, eth_lib, technology_lib, tech_ddr_lib, diag_lib, reorder_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE unb1_board_lib.unb1_board_pkg.ALL;
USE unb1_board_lib.unb1_board_peripherals_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE eth_lib.eth_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE reorder_lib.reorder_pkg.ALL;   

ENTITY unb1_reorder IS
  GENERIC (
    g_sim         : BOOLEAN      := FALSE; 
    g_sim_unb_nr  : NATURAL      := 0;
    g_sim_node_nr : NATURAL      := 4;
    -- Stamps are passed via QIP at compile start if $UNB_COMPILE_STAMPS is set
    g_stamp_date  : NATURAL      := 0;     -- Date (YYYYMMDD)
    g_stamp_time  : NATURAL      := 0;     -- Time (HHMMSS)
    g_stamp_svn   : NATURAL      := 0;     -- SVN revision  
    g_use_MB_I    : NATURAL      := 1;              -- 1: use MB_I  0: do not use
    g_tech_ddr    : t_c_tech_ddr := c_tech_ddr3_4g_single_rank_800m_master;
    g_aux         : t_c_unb1_board_aux  := c_unb1_board_aux
  );
  PORT (
    -- GENERAL
    CLK           : IN    STD_LOGIC; -- System Clock 
    PPS           : IN    STD_LOGIC; -- System Sync
    WDI           : OUT   STD_LOGIC; -- Watchdog Clear   
    INTA          : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB          : INOUT STD_LOGIC; -- FPGA interconnect line
    
    -- Others
    VERSION       : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0);
    ID            : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0);
    TESTIO        : INOUT STD_LOGIC_VECTOR(c_unb1_board_aux.testio_w-1 DOWNTO 0);
  
    -- I2C Interface to Sensors
    sens_sc       : INOUT STD_LOGIC;
    sens_sd       : INOUT STD_LOGIC;

    -- 1GbE Control Interface
    ETH_clk       : IN    STD_LOGIC;  
    ETH_SGIN      : IN    STD_LOGIC;
    ETH_SGOUT     : OUT   STD_LOGIC;
    
    -- SO-DIMM Memory Bank I
    MB_I_IN       : IN    t_tech_ddr3_phy_in;     
    MB_I_IO       : INOUT t_tech_ddr3_phy_io;     
    MB_I_OU       : OUT   t_tech_ddr3_phy_ou      
  );
END unb1_reorder;


ARCHITECTURE str OF unb1_reorder IS

  -- Constant definitions for ctrl_unb_common
  CONSTANT c_design_name      : STRING := "unb1_reorder";  
  CONSTANT c_design_note      : STRING := "Reference Reorder";
  CONSTANT c_fw_version       : t_unb1_board_fw_version := (0, 13);  -- firmware version x.y
  CONSTANT c_use_phy          : t_c_unb1_board_use_phy := (1, 0, 0, 0, 1, 0, 0, 1);
  CONSTANT c_aux              : t_c_unb1_board_aux     := c_unb1_board_aux;
  CONSTANT c_app_led_en       : BOOLEAN                := TRUE;
                              
  CONSTANT c_wr_chunksize     : POSITIVE := 256;
  CONSTANT c_wr_nof_chunks    : POSITIVE := 1;  
  CONSTANT c_rd_chunksize     : POSITIVE := 32;
  CONSTANT c_rd_nof_chunks    : POSITIVE := 8;  
  CONSTANT c_rd_interval      : POSITIVE := 32;
  CONSTANT c_gapsize          : NATURAL  := 0;
  CONSTANT c_nof_blocks       : POSITIVE := 32;
  CONSTANT c_nof_streams      : POSITIVE := 4;
  CONSTANT c_in_dat_w         : POSITIVE := 8;
  CONSTANT c_ena_pre_transp   : BOOLEAN  := TRUE;
  CONSTANT c_frame_size_in    : NATURAL  := 256;
  CONSTANT c_blocksize        : POSITIVE := c_wr_nof_chunks * c_wr_chunksize;  
  CONSTANT c_reorder_seq_conf : t_reorder_seq := (c_wr_chunksize, 
                                                  c_rd_chunksize, 
                                                  c_rd_nof_chunks,  
                                                  c_rd_interval,
                                                  c_gapsize,      
                                                  c_nof_blocks);   
  -- Block generator
  CONSTANT c_bg_block_size      : NATURAL := c_blocksize * c_rd_chunksize;
  CONSTANT c_bg_gapsize         : NATURAL := 0;

  CONSTANT c_bg_buf_dat_w           : POSITIVE := c_nof_complex*c_in_dat_w;
  CONSTANT c_bg_buf_adr_w           : POSITIVE := ceil_log2(c_bg_block_size);
  CONSTANT c_bg_data_file_prefix    : STRING   := "hex/tb_bg_dat";
  CONSTANT c_bg_data_file_index_arr : t_nat_natural_arr := array_init(0, 128, 1);


  -- Configuration of the databuffers:
  CONSTANT c_db_data_w              : POSITIVE := c_diag_db_max_data_w;
  CONSTANT c_db_buf_nof_data        : POSITIVE := c_bg_block_size;
  CONSTANT c_db_buf_use_sync        : BOOLEAN  := FALSE;
  CONSTANT c_db_data_type_re        : t_diag_data_type_enum := e_real;
  CONSTANT c_db_data_type_im        : t_diag_data_type_enum := e_imag;

  -- System
  SIGNAL cs_sim                     : STD_LOGIC;
  SIGNAL xo_clk                     : STD_LOGIC;
  SIGNAL xo_rst                     : STD_LOGIC;
  SIGNAL xo_rst_n                   : STD_LOGIC;   
  SIGNAL cal_clk                    : STD_LOGIC;
  SIGNAL mm_clk                     : STD_LOGIC;
  SIGNAL mm_locked                  : STD_LOGIC;
  SIGNAL mm_rst                     : STD_LOGIC;
  
  SIGNAL dp_rst                     : STD_LOGIC;
  SIGNAL dp_clk                     : STD_LOGIC;
  SIGNAL dp_pps                     : STD_LOGIC;      
  SIGNAL ddr_ref_rst                : STD_LOGIC;      
  
  SIGNAL this_chip_id               : STD_LOGIC_VECTOR(c_unb1_board_nof_chip_w-1 DOWNTO 0);  -- [2:0], so range 0-3 for FN and range 4-7 BN
  
  SIGNAL app_led_red                : STD_LOGIC := '0';
  SIGNAL app_led_green              : STD_LOGIC := '1';
  
  -- PIOs
  SIGNAL pout_debug_wave            : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL pout_wdi                   : STD_LOGIC;

  -- WDI override
  SIGNAL reg_wdi_mosi               : t_mem_mosi;
  SIGNAL reg_wdi_miso               : t_mem_miso;

  -- PPSH
  SIGNAL reg_ppsh_mosi              : t_mem_mosi;
  SIGNAL reg_ppsh_miso              : t_mem_miso;
  
  -- UniBoard system info
  SIGNAL reg_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL reg_unb_system_info_miso   : t_mem_miso;
  SIGNAL rom_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL rom_unb_system_info_miso   : t_mem_miso;

  -- Blockgenerator
  SIGNAL reg_diag_bg_mosi           : t_mem_mosi;
  SIGNAL reg_diag_bg_miso           : t_mem_miso;
                                    
  SIGNAL ram_diag_bg_mosi           : t_mem_mosi;
  SIGNAL ram_diag_bg_miso           : t_mem_miso;

  -- Databuffer                               
  SIGNAL ram_diag_data_buf_re_mosi  : t_mem_mosi;
  SIGNAL ram_diag_data_buf_re_miso  : t_mem_miso;
                                    
  SIGNAL reg_diag_data_buf_re_mosi  : t_mem_mosi;
  SIGNAL reg_diag_data_buf_re_miso  : t_mem_miso;
                                    
  SIGNAL ram_diag_data_buf_im_mosi  : t_mem_mosi;
  SIGNAL ram_diag_data_buf_im_miso  : t_mem_miso;
                                    
  SIGNAL reg_diag_data_buf_im_mosi  : t_mem_mosi;
  SIGNAL reg_diag_data_buf_im_miso  : t_mem_miso;

  -- eth1g
  SIGNAL eth1g_tse_clk              : STD_LOGIC;
  SIGNAL eth1g_mm_rst               : STD_LOGIC;
  SIGNAL eth1g_tse_mosi             : t_mem_mosi := c_mem_mosi_rst;  -- ETH TSE MAC registers
  SIGNAL eth1g_tse_miso             : t_mem_miso;
  SIGNAL eth1g_reg_mosi             : t_mem_mosi := c_mem_mosi_rst;  -- ETH control and status registers
  SIGNAL eth1g_reg_miso             : t_mem_miso;
  SIGNAL eth1g_reg_interrupt        : STD_LOGIC;   -- Interrupt
  SIGNAL eth1g_ram_mosi             : t_mem_mosi := c_mem_mosi_rst;  -- ETH rx frame and tx frame memory
  SIGNAL eth1g_ram_miso             : t_mem_miso;

  -- Reorder Transpose
  SIGNAL ram_ss_ss_transp_mosi      : t_mem_mosi;
  SIGNAL ram_ss_ss_transp_miso      : t_mem_miso;       
  
  -- IO DDR register map 
  SIGNAL reg_io_ddr_mosi            : t_mem_mosi;       
  SIGNAL reg_io_ddr_miso            : t_mem_miso;       
  
   -- . UniBoard I2C sens                         
  SIGNAL reg_unb_sens_mosi          : t_mem_mosi;
  SIGNAL reg_unb_sens_miso          : t_mem_miso;
  
  SIGNAL reg_bsn_monitor_mosi       : t_mem_mosi;
  SIGNAL reg_bsn_monitor_miso       : t_mem_miso;
 
  SIGNAL block_gen_src_out_arr      : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0);
  SIGNAL block_gen_src_in_arr       : t_dp_siso_arr(c_nof_streams-1 DOWNTO 0) := (OTHERS=> c_dp_siso_rdy);

  SIGNAL data_buf_snk_in_sosi_arr   : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0);
  SIGNAL data_buf_snk_out_siso_arr  : t_dp_siso_arr(c_nof_streams-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  
--  SIGNAL debug_out_arr              : t_dp_sosi_arr(4 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl : ENTITY unb1_board_lib.ctrl_unb1_board
  GENERIC MAP (
    -- General
    g_sim            => g_sim,
    g_stamp_date     => g_stamp_date,
    g_stamp_time     => g_stamp_time, 
    g_stamp_svn      => g_stamp_svn,
    g_design_name    => c_design_name,
    g_design_note    => c_design_note, 
    g_fw_version     => c_fw_version,
    g_mm_clk_freq    => c_unb1_board_mm_clk_freq_50M,
    g_dp_clk_use_pll => FALSE,
    g_app_led_red    => c_app_led_en,
    g_app_led_green  => c_app_led_en,
    g_use_phy        => c_use_phy,
    g_aux            => c_aux
  )
  PORT MAP (
    --
    -- >>> SOPC system with conduit peripheral MM bus
    --
    -- System
    cs_sim                   => cs_sim,
    xo_clk                   => xo_clk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,
                             
    mm_clk                   => mm_clk,
    mm_locked                => mm_locked,
    mm_rst                   => mm_rst,
                             
    dp_rst                   => OPEN, --dp_rst, 
    dp_clk                   => OPEN, --dp_clk, -- dp_clk is now generated in the DDR controller
    dp_pps                   => OPEN,
    dp_rst_in                => dp_rst,
    dp_clk_in                => dp_clk,
                             
    this_chip_id             => this_chip_id,
    this_bck_id              => OPEN,
                             
    app_led_red              => app_led_red,
    app_led_green            => app_led_green,
                             
    -- PIOs                  
    pout_debug_wave          => pout_debug_wave,
    pout_wdi                 => pout_wdi,
                             
    -- Manual WDI override   
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,

    -- . PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,

    -- system_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 
     
     -- UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,
                             
    -- eth1g                 
    eth1g_tse_clk            => eth1g_tse_clk,
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,
 
    --
    -- >>> Ctrl FPGA pins
    --
    -- General
    CLK                      => CLK,
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
                             
    -- Others                
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO, 
    
    -- I2C Interface to Sensors                        
    sens_sc                  => sens_sc,
    sens_sd                  => sens_sd,
                             
    ETH_clk                  => ETH_clk,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  ); 
  
  u_mmm : ENTITY work.mmm_unb1_reorder
  GENERIC MAP(
    g_sim         => g_sim,         
    g_sim_unb_nr  => g_sim_unb_nr,
    g_sim_node_nr => g_sim_node_nr, 
    g_nof_streams => c_nof_streams,
    g_reorder_seq => c_reorder_seq_conf
  )
  PORT MAP (
    -- GENERAL
    xo_clk                    => xo_clk,   
    xo_rst_n                  => xo_rst_n, 
    xo_rst                    => xo_rst,   
                                           
    mm_rst                    => mm_rst,   
    mm_clk                    => mm_clk,   
    mm_locked                 => mm_locked,
    cal_clk                   => cal_clk,  
                              
    pout_wdi                  => pout_wdi,                 
                                                           
    -- Manual WDI override                                 
    reg_wdi_mosi              => reg_wdi_mosi,             
    reg_wdi_miso              => reg_wdi_miso,             
                                                           
    -- system_info                                         
    reg_unb_system_info_mosi  => reg_unb_system_info_mosi, 
    reg_unb_system_info_miso  => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi  => rom_unb_system_info_mosi, 
    rom_unb_system_info_miso  => rom_unb_system_info_miso, 
                                                           
    -- UniBoard I2C sensors                                
    reg_unb_sens_mosi         => reg_unb_sens_mosi,        
    reg_unb_sens_miso         => reg_unb_sens_miso,        

    -- PPSH                   
    reg_ppsh_mosi             => reg_ppsh_mosi,
    reg_ppsh_miso             => reg_ppsh_miso, 

    -- Blockgenerator         
    reg_diag_bg_mosi          => reg_diag_bg_mosi,
    reg_diag_bg_miso          => reg_diag_bg_miso,
    ram_diag_bg_mosi          => ram_diag_bg_mosi,
    ram_diag_bg_miso          => ram_diag_bg_miso,

    -- DDR3 transpose         
    ram_ss_ss_transp_mosi     => ram_ss_ss_transp_mosi,
    ram_ss_ss_transp_miso     => ram_ss_ss_transp_miso,

    -- eth1g
    eth1g_tse_clk             => eth1g_tse_clk,      
    eth1g_mm_rst              => eth1g_mm_rst,       
    eth1g_tse_mosi            => eth1g_tse_mosi,     
    eth1g_tse_miso            => eth1g_tse_miso,     
    eth1g_reg_mosi            => eth1g_reg_mosi,     
    eth1g_reg_miso            => eth1g_reg_miso,     
    eth1g_reg_interrupt       => eth1g_reg_interrupt,
    eth1g_ram_mosi            => eth1g_ram_mosi,     
    eth1g_ram_miso            => eth1g_ram_miso,      
                             
    -- Databuffers            
    ram_diag_data_buf_im_mosi => ram_diag_data_buf_im_mosi,
    ram_diag_data_buf_im_miso => ram_diag_data_buf_im_miso,
    reg_diag_data_buf_im_mosi => reg_diag_data_buf_im_mosi,
    reg_diag_data_buf_im_miso => reg_diag_data_buf_im_miso,
    ram_diag_data_buf_re_mosi => ram_diag_data_buf_re_mosi,
    ram_diag_data_buf_re_miso => ram_diag_data_buf_re_miso,
    reg_diag_data_buf_re_mosi => reg_diag_data_buf_re_mosi,
    reg_diag_data_buf_re_miso => reg_diag_data_buf_re_miso,

    -- BSN monitor
    reg_bsn_monitor_mosi      => reg_bsn_monitor_mosi,
    reg_bsn_monitor_miso      => reg_bsn_monitor_miso,

    -- IO DDR register map 
    reg_io_ddr_mosi           => reg_io_ddr_mosi,
    reg_io_ddr_miso           => reg_io_ddr_miso
    
  );
  
  u_areset_ddr_ref_rst : ENTITY common_lib.common_areset
  GENERIC MAP(
    g_rst_level => '1',
    g_delay_len => 40
  )
  PORT MAP(
    clk     => CLK,
    in_rst  => '0',
    out_rst => ddr_ref_rst
  );    

  u_node : ENTITY work.node_unb1_reorder
  GENERIC MAP(
    g_sim            => g_sim,       
    g_use_MB_I       => g_use_MB_I,  
    g_tech_ddr       => g_tech_ddr,       
    g_nof_streams    => c_nof_streams,   
    g_in_dat_w       => c_in_dat_w,      
    g_ena_pre_transp => c_ena_pre_transp,
    g_reorder_seq    => c_reorder_seq_conf
  )
  PORT MAP(
    -- System
    mm_rst                => mm_rst,
    mm_clk                => mm_clk,
    dp_rst                => dp_rst,
    dp_clk                => dp_clk,
    
    ddr_ref_clk           => CLK,      -- Provide external 200 MHZ clk to DDR controller 
    ddr_ref_rst           => ddr_ref_rst,
    
    -- Clock outputs
    ddr_out_clk           => dp_clk,   
    ddr_out_rst           => dp_rst,   -- dp_clk is generated by DDR controller
    
    -- ST sinks          
    in_siso_arr           => block_gen_src_in_arr,
    in_sosi_arr           => block_gen_src_out_arr, 

    -- ST source         
    out_siso_arr          => data_buf_snk_out_siso_arr,
    out_sosi_arr          => data_buf_snk_in_sosi_arr,
    
    -- IO DDR register map 
    reg_io_ddr_mosi       => reg_io_ddr_mosi,
    reg_io_ddr_miso       => reg_io_ddr_miso,
    
    -- Reorder transpose      
    ram_ss_ss_transp_mosi => ram_ss_ss_transp_mosi,
    ram_ss_ss_transp_miso => ram_ss_ss_transp_miso,
    
    -- BSN monitor
    reg_bsn_monitor_mosi      => reg_bsn_monitor_mosi,
    reg_bsn_monitor_miso      => reg_bsn_monitor_miso,

    -- SO-DIMM Memory Bank I = ddr3_I
    MB_I_in               =>  MB_I_IN, 
    MB_I_io               =>  MB_I_IO, 
    MB_I_ou               =>  MB_I_OU
  );

  -----------------------------------------------------------------------------
  -- TX: Block generator
  -----------------------------------------------------------------------------
  u_mms_diag_block_gen : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP (
    g_nof_streams        => c_nof_streams,
    g_buf_dat_w          => c_bg_buf_dat_w,
    g_buf_addr_w         => c_bg_buf_adr_w,
    g_file_index_arr     => c_bg_data_file_index_arr,
    g_file_name_prefix   => c_bg_data_file_prefix
  )
  PORT MAP (
    mm_rst           => mm_rst,
    mm_clk           => mm_clk,

    dp_rst           => dp_rst,
    dp_clk           => dp_clk,
    en_sync          => '1',

    reg_bg_ctrl_mosi => reg_diag_bg_mosi,
    reg_bg_ctrl_miso => reg_diag_bg_miso,
    ram_bg_data_mosi => ram_diag_bg_mosi,
    ram_bg_data_miso => ram_diag_bg_miso,

    out_siso_arr     => block_gen_src_in_arr,
    out_sosi_arr     => block_gen_src_out_arr
  );
    
  ----------------------------------------------------------------------------
  -- Sink: data buffer real 
  ---------------------------------------------------------------------------- 
  u_data_buf_re : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams     => c_nof_streams, 
    g_data_type       => c_db_data_type_re,   
    g_data_w          => c_db_data_w,      
    g_buf_nof_data    => c_db_buf_nof_data,
    g_buf_use_sync    => c_db_buf_use_sync
  )
  PORT MAP (
    -- System
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,
     -- MM interface
    ram_data_buf_mosi => ram_diag_data_buf_re_mosi,
    ram_data_buf_miso => ram_diag_data_buf_re_miso,
    reg_data_buf_mosi => reg_diag_data_buf_re_mosi,
    reg_data_buf_miso => reg_diag_data_buf_re_miso,
    -- ST interface
    in_sync           => data_buf_snk_in_sosi_arr(0).sync,
    in_sosi_arr       => data_buf_snk_in_sosi_arr
  );

  ----------------------------------------------------------------------------
  -- Sink: data buffer imag 
  ---------------------------------------------------------------------------- 
  u_data_buf_im : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams     => c_nof_streams, 
    g_data_type       => c_db_data_type_im,   
    g_data_w          => c_db_data_w,      
    g_buf_nof_data    => c_db_buf_nof_data,
    g_buf_use_sync    => c_db_buf_use_sync
  )
  PORT MAP (
    -- System
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,
    -- MM interface
    ram_data_buf_mosi => ram_diag_data_buf_im_mosi,
    ram_data_buf_miso => ram_diag_data_buf_im_miso,
    reg_data_buf_mosi => reg_diag_data_buf_im_mosi,
    reg_data_buf_miso => reg_diag_data_buf_im_miso,
    -- ST interface
    in_sync           => data_buf_snk_in_sosi_arr(0).sync,
    in_sosi_arr       => data_buf_snk_in_sosi_arr
  );

END str;

