-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb1_board_lib, mm_lib, dp_lib, eth_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE unb1_board_lib.unb1_board_pkg.ALL;
USE unb1_board_lib.unb1_board_peripherals_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE eth_lib.eth_pkg.ALL;

ENTITY mmm_unb1_ddr3 IS
  GENERIC (
    g_sim         : BOOLEAN := TRUE; --FALSE: use SOPC; TRUE: use mm_file I/O
    g_sim_unb_nr  : NATURAL := 0;
    g_sim_node_nr : NATURAL := 0
  );
  PORT (
    -- GENERAL
    xo_clk                   : IN  STD_LOGIC;
    xo_rst_n                 : IN  STD_LOGIC;
    xo_rst                   : IN  STD_LOGIC;

    mm_rst                   : IN  STD_LOGIC;
    mm_clk                   : OUT STD_LOGIC;
    mm_locked                : OUT STD_LOGIC;
    cal_clk                  : OUT STD_LOGIC;

    pout_wdi                 : OUT STD_LOGIC;

    -- Manual WDI override
    reg_wdi_mosi             : OUT t_mem_mosi;
    reg_wdi_miso             : IN  t_mem_miso;

    -- PPSH
    reg_ppsh_mosi            : OUT t_mem_mosi;
    reg_ppsh_miso            : IN  t_mem_miso;

    -- system_info
    reg_unb_system_info_mosi : OUT t_mem_mosi;
    reg_unb_system_info_miso : IN  t_mem_miso;
    rom_unb_system_info_mosi : OUT t_mem_mosi;
    rom_unb_system_info_miso : IN  t_mem_miso;

    -- UniBoard I2C sensors
    reg_unb_sens_mosi        : OUT t_mem_mosi;
    reg_unb_sens_miso        : IN  t_mem_miso;

    -- DDR3
    reg_io_ddr_mosi          : OUT  t_mem_mosi;
    reg_io_ddr_miso          : IN   t_mem_miso;

    -- Data Buffer Control
    reg_diag_data_buf_mosi   : OUT t_mem_mosi;
    reg_diag_data_buf_miso   : IN  t_mem_miso;

    -- Data Buffer Data
    ram_diag_data_buf_mosi   : OUT t_mem_mosi;
    ram_diag_data_buf_miso   : IN  t_mem_miso;

    -- Block Generator Control
    reg_diag_bg_ctrl_mosi    : OUT t_mem_mosi;
    reg_diag_bg_ctrl_miso    : IN  t_mem_miso;

    -- Block Generator Data
    ram_diag_bg_data_mosi    : OUT t_mem_mosi;
    ram_diag_bg_data_miso    : IN  t_mem_miso;

    -- TX Sequencer
    reg_diag_tx_seq_mosi     : OUT t_mem_mosi;
    reg_diag_tx_seq_miso     : IN  t_mem_miso;

    -- RX Sequencer
    reg_diag_rx_seq_mosi     : OUT t_mem_mosi;
    reg_diag_rx_seq_miso     : IN  t_mem_miso;

    -- eth1g
    eth1g_tse_clk            : OUT STD_LOGIC;
    eth1g_mm_rst             : OUT STD_LOGIC;
    eth1g_tse_mosi           : OUT t_mem_mosi;
    eth1g_tse_miso           : IN  t_mem_miso;
    eth1g_reg_mosi           : OUT t_mem_mosi;
    eth1g_reg_miso           : IN  t_mem_miso;
    eth1g_reg_interrupt      : IN  STD_LOGIC;
    eth1g_ram_mosi           : OUT t_mem_mosi;
    eth1g_ram_miso           : IN  t_mem_miso
  );
END mmm_unb1_ddr3;


ARCHITECTURE str OF mmm_unb1_ddr3 IS

  -- Actual MM address widths, the MM data width is fixed at the default c_word_w=32
  CONSTANT c_mm_reg_io_ddr_addr_w        : NATURAL := 4;
  CONSTANT c_mm_reg_diag_data_buf_addr_w : NATURAL := 5;
  CONSTANT c_mm_ram_diag_data_buf_addr_w : NATURAL := 14;
  CONSTANT c_mm_reg_diag_bg_ctrl_addr_w  : NATURAL := 3;
  CONSTANT c_mm_ram_diag_bg_data_addr_w  : NATURAL := 3;
  CONSTANT c_mm_reg_diag_tx_seq_addr_w   : NATURAL := 2;
  CONSTANT c_mm_reg_diag_rx_seq_addr_w   : NATURAL := 3;

  -- Simulation
  CONSTANT c_mm_clk_period             : TIME := 8 ns; --500 ps;
  CONSTANT c_tse_clk_period            : TIME := 8 ns;
  CONSTANT c_cal_clk_period            : TIME := 25 ns;

  CONSTANT c_sim_node_type             : STRING(1 TO 2):= sel_a_b(g_sim_node_nr<4, "FN", "BN");
  CONSTANT c_sim_node_nr               : NATURAL := sel_a_b(c_sim_node_type="BN", g_sim_node_nr-4, g_sim_node_nr);

  -- PIOs
  SIGNAL pout_debug_wave : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL pin_pps         : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);

  SIGNAL i_mm_clk        : STD_LOGIC := '1';
  SIGNAL i_tse_clk       : STD_LOGIC := '1';
  SIGNAL i_cal_clk       : STD_LOGIC := '1';

  ----------------------------------------------------------------------------
  -- mm_file component
  ----------------------------------------------------------------------------
  COMPONENT mm_file
  GENERIC(
    g_file_prefix       : STRING;
    g_mm_clk_period     : TIME := c_mm_clk_period;
    g_update_on_change  : BOOLEAN := FALSE;
    g_mm_rd_latency     : NATURAL := 1
  );
  PORT (
    mm_rst        : IN  STD_LOGIC;
    mm_clk        : IN  STD_LOGIC;
    mm_master_out : OUT t_mem_mosi;
    mm_master_in  : IN  t_mem_miso 
  );
  END COMPONENT;

  CONSTANT c_dut_src_mac       : STD_LOGIC_VECTOR(c_network_eth_mac_slv'RANGE) := X"002286080001";
  SIGNAL eth_psc_access        : STD_LOGIC;

  CONSTANT c_dut_control_rx_en : NATURAL := 2**c_eth_mm_reg_control_bi.rx_en;

  SIGNAL i_eth1g_reg_mosi      : t_mem_mosi;
  SIGNAL i_eth1g_reg_miso      : t_mem_miso;

  SIGNAL eth1g_reg_proc_mosi   : t_mem_mosi;
  SIGNAL eth1g_reg_proc_miso   : t_mem_miso;

  SIGNAL mm_bus_switch         : STD_LOGIC;

BEGIN

  mm_clk        <= i_mm_clk;
  eth1g_tse_clk <= i_tse_clk;
  cal_clk       <= i_cal_clk;

  ----------------------------------------------------------------------------
  -- MM <-> file I/O for simulation. The files are created in $UPE/sim.
  ----------------------------------------------------------------------------
  gen_mm_file_io : IF g_sim = TRUE GENERATE

    i_mm_clk  <= NOT i_mm_clk AFTER c_mm_clk_period/2;
    mm_locked <= '0', '1' AFTER c_mm_clk_period*5;

    i_tse_clk    <= NOT i_tse_clk AFTER c_tse_clk_period/2;
    eth1g_mm_rst <= '1', '0' AFTER c_tse_clk_period*5;

    i_cal_clk <= NOT i_cal_clk AFTER c_cal_clk_period/2;

    u_mm_file_reg_unb_system_info : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "PIO_SYSTEM_INFO")
                                               PORT MAP(mm_rst, i_mm_clk, reg_unb_system_info_mosi, reg_unb_system_info_miso );

    u_mm_file_rom_unb_system_info : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "ROM_SYSTEM_INFO")
                                               PORT MAP(mm_rst, i_mm_clk, rom_unb_system_info_mosi, rom_unb_system_info_miso );

    u_mm_file_reg_wdi             : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "REG_WDI")
                                               PORT MAP(mm_rst, i_mm_clk, reg_wdi_mosi, reg_wdi_miso );

    u_mm_file_reg_unb_sens        : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "REG_UNB_SENS")
                                               PORT MAP(mm_rst, i_mm_clk, reg_unb_sens_mosi, reg_unb_sens_miso );                          

    u_mm_file_reg_ppsh            : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "PIO_PPS")                                               
                                               PORT MAP(mm_rst, i_mm_clk, reg_ppsh_mosi, reg_ppsh_miso );                              
                                               
    u_mm_file_reg_io_ddr          : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "REG_IO_DDR")
                                               PORT MAP(mm_rst, i_mm_clk, reg_io_ddr_mosi, reg_io_ddr_miso );

    u_mm_file_reg_diag_data_buf   : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "REG_DIAG_DATA_BUF")
                                               PORT MAP(mm_rst, i_mm_clk, reg_diag_data_buf_mosi, reg_diag_data_buf_miso );

    u_mm_file_ram_diag_data_buf   : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "RAM_DIAG_DATA_BUF")
                                               PORT MAP(mm_rst, i_mm_clk, ram_diag_data_buf_mosi, ram_diag_data_buf_miso );

    u_mm_file_reg_diag_bg_ctrl    : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "REG_DIAG_BG_CTRL")
                                               PORT MAP(mm_rst, i_mm_clk, reg_diag_bg_ctrl_mosi, reg_diag_bg_ctrl_miso );

    u_mm_file_ram_diag_bg_data    : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "RAM_DIAG_BG_DATA")
                                               PORT MAP(mm_rst, i_mm_clk, ram_diag_bg_data_mosi, ram_diag_bg_data_miso );

    u_mm_file_reg_diag_tx_seq     : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "REG_DIAG_TX_SEQ")
                                               PORT MAP(mm_rst, i_mm_clk, reg_diag_tx_seq_mosi, reg_diag_tx_seq_miso );

    u_mm_file_reg_diag_rx_seq     : mm_file GENERIC MAP(mmf_unb_file_prefix(g_sim_unb_nr, c_sim_node_nr, c_sim_node_type) & "REG_DIAG_RX_SEQ")
                                               PORT MAP(mm_rst, i_mm_clk, reg_diag_rx_seq_mosi, reg_diag_rx_seq_miso );


    ----------------------------------------------------------------------------
    -- 1GbE setup sequence normally performed by unb_os@NIOS
    ----------------------------------------------------------------------------
    p_eth_setup : PROCESS
    BEGIN
      mm_bus_switch <= '1';

      eth1g_tse_mosi.wr <= '0';
      eth1g_tse_mosi.rd <= '0';
      WAIT FOR 400 ns;
      WAIT UNTIL rising_edge(i_mm_clk);

      -- Enable RX
      proc_mem_mm_bus_wr(c_eth_reg_control_wi+0, c_dut_control_rx_en, i_mm_clk, eth1g_reg_miso, eth1g_reg_proc_mosi);  -- control rx en
      mm_bus_switch <= '0';

      WAIT;
    END PROCESS;

    p_switch : PROCESS(mm_bus_switch, eth1g_reg_proc_mosi, i_eth1g_reg_mosi)
    BEGIN
      IF mm_bus_switch = '1' THEN 
        eth1g_reg_mosi <= eth1g_reg_proc_mosi;
      ELSE
        eth1g_reg_mosi <= i_eth1g_reg_mosi;
      END IF;
    END PROCESS;

    ----------------------------------------------------------------------------
    -- Procedure that polls a sim control file that can be used to e.g. get
    -- the simulation time in ns
    ----------------------------------------------------------------------------
    mmf_poll_sim_ctrl_file(c_mmf_unb_file_path & "sim.ctrl", c_mmf_unb_file_path & "sim.stat");

  END GENERATE;

  ----------------------------------------------------------------------------
  -- SOPC for synthesis
  ----------------------------------------------------------------------------
  gen_sopc : IF g_sim = FALSE GENERATE
    u_sopc : ENTITY work.sopc_unb1_ddr3
    PORT MAP (
      clk_0                                          => xo_clk,        -- PLL reference = 25 MHz from ETH_clk pin
      reset_n                                        => xo_rst_n,
      mm_clk                                         => i_mm_clk,      -- PLL clk[0] = 125 MHz system clock that the NIOS2 and the MM bus run on
      cal_clk                                        => i_cal_clk,     -- PLL clk[1] =  40 MHz calibration clock for the IO reconfiguration
      tse_clk                                        => i_tse_clk,     -- PLL clk[2] = 125 MHz dedicated clock for the 1 Gbit Ethernet unit

      -- the_altpll_0
      areset_to_the_altpll_0                         => '0',
      locked_from_the_altpll_0                       => mm_locked,
      phasedone_from_the_altpll_0                    => OPEN,

      -- the_avs_eth_0
      coe_clk_export_from_the_avs_eth_0              => OPEN,
      coe_reset_export_from_the_avs_eth_0            => eth1g_mm_rst,
      coe_tse_address_export_from_the_avs_eth_0      => eth1g_tse_mosi.address(c_unb1_board_peripherals_mm_reg_default.reg_tse_adr_w-1 DOWNTO 0),
      coe_tse_write_export_from_the_avs_eth_0        => eth1g_tse_mosi.wr,
      coe_tse_writedata_export_from_the_avs_eth_0    => eth1g_tse_mosi.wrdata(c_word_w-1 DOWNTO 0),
      coe_tse_read_export_from_the_avs_eth_0         => eth1g_tse_mosi.rd,
      coe_tse_readdata_export_to_the_avs_eth_0       => eth1g_tse_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_tse_waitrequest_export_to_the_avs_eth_0    => eth1g_tse_miso.waitrequest,
      coe_reg_address_export_from_the_avs_eth_0      => eth1g_reg_mosi.address(c_unb1_board_peripherals_mm_reg_default.reg_eth_adr_w-1 DOWNTO 0),
      coe_reg_write_export_from_the_avs_eth_0        => eth1g_reg_mosi.wr,
      coe_reg_writedata_export_from_the_avs_eth_0    => eth1g_reg_mosi.wrdata(c_word_w-1 DOWNTO 0),
      coe_reg_read_export_from_the_avs_eth_0         => eth1g_reg_mosi.rd,
      coe_reg_readdata_export_to_the_avs_eth_0       => eth1g_reg_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_irq_export_to_the_avs_eth_0                => eth1g_reg_interrupt,
      coe_ram_address_export_from_the_avs_eth_0      => eth1g_ram_mosi.address(c_unb1_board_peripherals_mm_reg_default.ram_eth_adr_w-1 DOWNTO 0),
      coe_ram_write_export_from_the_avs_eth_0        => eth1g_ram_mosi.wr,
      coe_ram_writedata_export_from_the_avs_eth_0    => eth1g_ram_mosi.wrdata(c_word_w-1 DOWNTO 0),
      coe_ram_read_export_from_the_avs_eth_0         => eth1g_ram_mosi.rd,
      coe_ram_readdata_export_to_the_avs_eth_0       => eth1g_ram_miso.rddata(c_word_w-1 DOWNTO 0),

      -- the_reg_unb_sens
      coe_clk_export_from_the_reg_unb_sens           => OPEN,
      coe_reset_export_from_the_reg_unb_sens         => OPEN,
      coe_address_export_from_the_reg_unb_sens       => reg_unb_sens_mosi.address(c_unb1_board_peripherals_mm_reg_default.reg_unb_sens_adr_w-1 DOWNTO 0),
      coe_read_export_from_the_reg_unb_sens          => reg_unb_sens_mosi.rd,
      coe_readdata_export_to_the_reg_unb_sens        => reg_unb_sens_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_reg_unb_sens         => reg_unb_sens_mosi.wr,
      coe_writedata_export_from_the_reg_unb_sens     => reg_unb_sens_mosi.wrdata(c_word_w-1 DOWNTO 0),      

      -- the_pio_debug_wave
      out_port_from_the_pio_debug_wave               => pout_debug_wave,

      -- the_pio_pps
      coe_clk_export_from_the_pio_pps                => OPEN,
      coe_reset_export_from_the_pio_pps              => OPEN,
      coe_address_export_from_the_pio_pps            => reg_ppsh_mosi.address(c_unb1_board_peripherals_mm_reg_default.reg_ppsh_adr_w-1),  -- 1 bit address width so must use (0) instead of (0 DOWNTO 0)
      coe_read_export_from_the_pio_pps               => reg_ppsh_mosi.rd,
      coe_readdata_export_to_the_pio_pps             => reg_ppsh_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_pio_pps              => reg_ppsh_mosi.wr,
      coe_writedata_export_from_the_pio_pps          => reg_ppsh_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_pio_system_info: actually a avs_common_mm instance
      coe_clk_export_from_the_pio_system_info        => OPEN,
      coe_reset_export_from_the_pio_system_info      => OPEN,
      coe_address_export_from_the_pio_system_info    => reg_unb_system_info_mosi.address(c_unb1_board_peripherals_mm_reg_default.reg_unb_system_info_adr_w-1 DOWNTO 0), 
      coe_read_export_from_the_pio_system_info       => reg_unb_system_info_mosi.rd,
      coe_readdata_export_to_the_pio_system_info     => reg_unb_system_info_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_pio_system_info      => reg_unb_system_info_mosi.wr,
      coe_writedata_export_from_the_pio_system_info  => reg_unb_system_info_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_rom_system_info
      coe_clk_export_from_the_rom_system_info        => OPEN,
      coe_reset_export_from_the_rom_system_info      => OPEN,
      coe_address_export_from_the_rom_system_info    => rom_unb_system_info_mosi.address(c_unb1_board_peripherals_mm_reg_default.rom_unb_system_info_adr_w-1 DOWNTO 0), 
      coe_read_export_from_the_rom_system_info       => rom_unb_system_info_mosi.rd,
      coe_readdata_export_to_the_rom_system_info     => rom_unb_system_info_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_rom_system_info      => rom_unb_system_info_mosi.wr,
      coe_writedata_export_from_the_rom_system_info  => rom_unb_system_info_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_pio_wdi
      out_port_from_the_pio_wdi                      => pout_wdi,

      -- the_reg_wdi: Manual WDI override; causes FPGA reconfiguration if WDI is enabled (g_use_phy).
      coe_clk_export_from_the_reg_wdi                => OPEN,
      coe_reset_export_from_the_reg_wdi              => OPEN,
      coe_address_export_from_the_reg_wdi            => reg_wdi_mosi.address(0),
      coe_read_export_from_the_reg_wdi               => reg_wdi_mosi.rd,
      coe_readdata_export_to_the_reg_wdi             => reg_wdi_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_reg_wdi              => reg_wdi_mosi.wr,
      coe_writedata_export_from_the_reg_wdi          => reg_wdi_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_reg_io_ddr
      coe_clk_export_from_the_reg_io_ddr             => OPEN,
      coe_reset_export_from_the_reg_io_ddr           => OPEN,
      coe_address_export_from_the_reg_io_ddr         => reg_io_ddr_mosi.address(c_mm_reg_io_ddr_addr_w-1 DOWNTO 0),
      coe_read_export_from_the_reg_io_ddr            => reg_io_ddr_mosi.rd,
      coe_readdata_export_to_the_reg_io_ddr          => reg_io_ddr_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_reg_io_ddr           => reg_io_ddr_mosi.wr,
      coe_writedata_export_from_the_reg_io_ddr       => reg_io_ddr_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_ram_diag_bg_data
      coe_clk_export_from_the_ram_diag_bg_data       => OPEN,
      coe_reset_export_from_the_ram_diag_bg_data     => OPEN,
      coe_address_export_from_the_ram_diag_bg_data   => ram_diag_bg_data_mosi.address(c_mm_ram_diag_bg_data_addr_w-1 DOWNTO 0),
      coe_read_export_from_the_ram_diag_bg_data      => ram_diag_bg_data_mosi.rd,
      coe_readdata_export_to_the_ram_diag_bg_data    => ram_diag_bg_data_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_ram_diag_bg_data     => ram_diag_bg_data_mosi.wr,
      coe_writedata_export_from_the_ram_diag_bg_data => ram_diag_bg_data_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_ram_diag_data_buf
      coe_clk_export_from_the_ram_diag_data_buf       => OPEN,
      coe_reset_export_from_the_ram_diag_data_buf     => OPEN,
      coe_address_export_from_the_ram_diag_data_buf   => ram_diag_data_buf_mosi.address(c_mm_ram_diag_data_buf_addr_w-1 DOWNTO 0),
      coe_read_export_from_the_ram_diag_data_buf      => ram_diag_data_buf_mosi.rd,
      coe_readdata_export_to_the_ram_diag_data_buf    => ram_diag_data_buf_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_ram_diag_data_buf     => ram_diag_data_buf_mosi.wr,
      coe_writedata_export_from_the_ram_diag_data_buf => ram_diag_data_buf_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_reg_diag_bg_ctrl
      coe_clk_export_from_the_reg_diag_bg_ctrl       => OPEN,
      coe_reset_export_from_the_reg_diag_bg_ctrl     => OPEN,
      coe_address_export_from_the_reg_diag_bg_ctrl   => reg_diag_bg_ctrl_mosi.address(c_mm_reg_diag_bg_ctrl_addr_w-1 DOWNTO 0),
      coe_read_export_from_the_reg_diag_bg_ctrl      => reg_diag_bg_ctrl_mosi.rd,
      coe_readdata_export_to_the_reg_diag_bg_ctrl    => reg_diag_bg_ctrl_miso.rddata(c_word_w-1 DOWNTO 0),
      coe_write_export_from_the_reg_diag_bg_ctrl     => reg_diag_bg_ctrl_mosi.wr,
      coe_writedata_export_from_the_reg_diag_bg_ctrl => reg_diag_bg_ctrl_mosi.wrdata(c_word_w-1 DOWNTO 0),

      -- the_reg_diag_data_buf
      coe_clk_export_from_the_reg_diag_data_buf       => OPEN,
      coe_reset_export_from_the_reg_diag_data_buf     => OPEN,
      coe_address_export_from_the_reg_diag_data_buf   => reg_diag_data_buf_mosi.address(c_mm_reg_diag_data_buf_addr_w-1 DOWNTO 0), 
      coe_read_export_from_the_reg_diag_data_buf      => reg_diag_data_buf_mosi.rd,                                              
      coe_readdata_export_to_the_reg_diag_data_buf    => reg_diag_data_buf_miso.rddata(c_word_w-1 DOWNTO 0),                     
      coe_write_export_from_the_reg_diag_data_buf     => reg_diag_data_buf_mosi.wr,                                              
      coe_writedata_export_from_the_reg_diag_data_buf => reg_diag_data_buf_mosi.wrdata(c_word_w-1 DOWNTO 0),                     

      -- the_reg_diag_rx_seq
      coe_clk_export_from_the_reg_diag_rx_seq       => OPEN,
      coe_reset_export_from_the_reg_diag_rx_seq     => OPEN,
      coe_address_export_from_the_reg_diag_rx_seq   => reg_diag_rx_seq_mosi.address(c_mm_reg_diag_rx_seq_addr_w-1 DOWNTO 0), 
      coe_read_export_from_the_reg_diag_rx_seq      => reg_diag_rx_seq_mosi.rd,                                              
      coe_readdata_export_to_the_reg_diag_rx_seq    => reg_diag_rx_seq_miso.rddata(c_word_w-1 DOWNTO 0),                     
      coe_write_export_from_the_reg_diag_rx_seq     => reg_diag_rx_seq_mosi.wr,                                              
      coe_writedata_export_from_the_reg_diag_rx_seq => reg_diag_rx_seq_mosi.wrdata(c_word_w-1 DOWNTO 0),                     

      -- the_reg_diag_tx_seq
      coe_clk_export_from_the_reg_diag_tx_seq       => OPEN,
      coe_reset_export_from_the_reg_diag_tx_seq     => OPEN,
      coe_address_export_from_the_reg_diag_tx_seq   => reg_diag_tx_seq_mosi.address(c_mm_reg_diag_tx_seq_addr_w-1 DOWNTO 0), 
      coe_read_export_from_the_reg_diag_tx_seq      => reg_diag_tx_seq_mosi.rd,                                              
      coe_readdata_export_to_the_reg_diag_tx_seq    => reg_diag_tx_seq_miso.rddata(c_word_w-1 DOWNTO 0),                     
      coe_write_export_from_the_reg_diag_tx_seq     => reg_diag_tx_seq_mosi.wr,                                              
      coe_writedata_export_from_the_reg_diag_tx_seq => reg_diag_tx_seq_mosi.wrdata(c_word_w-1 DOWNTO 0)                     
    );
  END GENERATE;
END str;


