-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Verify DDR3 module using stimuli that are applied via python script

-- Description:

-- Usage:
--   Modelsim:
--   > load simulation configuration: tb_unb_ddr3
--   > run 100 us
--   Terminal:
--   > python tc_unb1_ddr3_seq.py --sim --unb 0 --fn 3 --rep 3

LIBRARY IEEE, common_lib, dp_lib, unb1_board_lib, diag_lib, tech_ddr_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE IEEE.MATH_REAL.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE unb1_board_lib.unb1_board_pkg.ALL;
USE unb1_board_lib.unb1_board_peripherals_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE tech_ddr_lib.tech_ddr_mem_model_component_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE diag_lib.diag_pkg.ALL;


ENTITY tb_unb1_ddr3 IS
  GENERIC (
    g_design_name : STRING  := "unb1_ddr3"
  );
END tb_unb1_ddr3;

ARCHITECTURE tb OF tb_unb1_ddr3 IS

  -- UniBoard
  CONSTANT c_sim             : BOOLEAN := TRUE;
  CONSTANT c_unb_nr          : NATURAL := 0; -- UniBoard 0
  CONSTANT c_node_nr         : NATURAL := 3; -- Front node 3
  CONSTANT c_id              : STD_LOGIC_VECTOR(7 DOWNTO 0) := TO_UVEC(c_unb_nr, c_unb1_board_nof_uniboard_w) & TO_UVEC(c_node_nr, c_unb1_board_nof_chip_w);

  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_fw_version      : t_unb1_board_fw_version := (1, 0);
  
  CONSTANT c_ddr             : t_c_tech_ddr := c_tech_ddr3_4g_800m_master;

  CONSTANT c_st_dat_w        : NATURAL := 64; -- Any power of two 8..256
  
  CONSTANT c_cable_delay     : TIME := 12 ns;
  CONSTANT c_eth_clk_period  : TIME := 40 ns;  -- 25 MHz XO on UniBoard
  CONSTANT c_ext_clk_period  : TIME := 5 ns;
  CONSTANT c_pps_period      : NATURAL := 1000; 
  
   -- SO-DIMM Memory Bank I = ddr3_I
  SIGNAL MB_I_in             : t_tech_ddr3_phy_in;
  SIGNAL MB_I_io             : t_tech_ddr3_phy_io;
  SIGNAL MB_I_ou             : t_tech_ddr3_phy_ou;

  -- DUT  
  SIGNAL clk                 : STD_LOGIC := '0';
  SIGNAL pps                 : STD_LOGIC := '0';
  SIGNAL pps_rst             : STD_LOGIC := '0';
  
  SIGNAL WDI                 : STD_LOGIC;
  SIGNAL INTA                : STD_LOGIC;
  SIGNAL INTB                : STD_LOGIC;

  SIGNAL eth_clk             : STD_LOGIC := '0';
  SIGNAL eth_txp             : STD_LOGIC;
  SIGNAL eth_rxp             : STD_LOGIC;

  SIGNAL VERSION             : STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0) := c_version; 
  SIGNAL ID                  : STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0)      := c_id;
  SIGNAL TESTIO              : STD_LOGIC_VECTOR(c_unb1_board_aux.testio_w-1 DOWNTO 0);
  
  SIGNAL sens_scl            : STD_LOGIC;
  SIGNAL sens_sda            : STD_LOGIC;     
  
BEGIN

   ----------------------------------------------------------------------------
  -- System setup
  ----------------------------------------------------------------------------
  clk     <= NOT clk AFTER c_ext_clk_period/2;          -- External clock (200 MHz)
  eth_clk <= NOT eth_clk AFTER c_eth_clk_period/2;  -- Ethernet ref clock (25 MHz)
  
  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up

  sens_scl <= 'H';  -- pull up
  sens_sda <= 'H';  -- pull up

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(1, c_pps_period, '1', pps_rst, clk, pps);

  ------------------------------------------------------------------------------
  -- 1GbE Loopback model
  ------------------------------------------------------------------------------  
  eth_rxp <= TRANSPORT eth_txp AFTER c_cable_delay;
  
  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------
  
  dut : ENTITY work.unb1_ddr3
  GENERIC MAP (
    g_sim           => c_sim,     
    g_sim_unb_nr    => c_unb_nr, 
    g_sim_node_nr   => c_node_nr,
    -- Stamps are passed via QIP at compile start if $UNB_COMPILE_STAMPS is set
    g_st_dat_w      => c_st_dat_w

  )
  PORT MAP (
    -- GENERAL
    CLK         => clk,
    PPS         => pps,
    WDI         => WDI,
    INTA        => INTA,
    INTB        => INTB,

    sens_sc     => sens_scl,
    sens_sd     => sens_sda,

    -- Others
    VERSION     => VERSION,
    ID          => ID,
    TESTIO      => TESTIO,

    -- 1GbE Control Interface
    ETH_clk     => eth_clk,
    ETH_SGIN    => eth_rxp,
    ETH_SGOUT   => eth_txp,
    
    -- SO-DIMM Memory Bank I = ddr3_I
    MB_I_IN     => MB_I_in,
    MB_I_IO     => MB_I_io,
    MB_I_OU     => MB_I_ou
   );

  ------------------------------------------------------------------------------
  -- DDR3 memory model
  ------------------------------------------------------------------------------
  u_tech_ddr_memory_model : ENTITY tech_ddr_lib.tech_ddr_memory_model
  GENERIC MAP (
    g_tech_ddr => c_ddr
  )
  PORT MAP (
    mem3_in => MB_I_ou,
    mem3_io => MB_I_io,
    mem3_ou => MB_I_in
  );
 
END tb;

