-------------------------------------------------------------------------------
--
-- Copyright (C) 2012-2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

--  RO  read only  (no VHDL present to access HW in write mode)
--  WO  write only (no VHDL present to access HW in read mode)
--  WE  write event (=WO)
--  WR  write control, read control
--  RW  read status, write control  
--  RC  read, clear on read
--  FR  FIFO read
--  FW  FIFO write
--
--  wi  Bits    R/W Name          Default  Description      |REG_UNB2_BOARD_SYSTEM_INFO|
--  =============================================================================
--  0   [23..0] RO  info          
--  1   [7..0]  RO  0
--  2   [31..0] RO  design_name
--  .   ..      .   ..
--  9   [31..0] RO  design name
--  10  [31..0] RO  date stamp             (YYYYMMDD)
--  11  [31..0] RO  time stamp             (HHMMSS)
--  12  [31..0] RO  SVN  stamp         
--  13  [31..0] RO  note
--  .   .       .   ..
--  20  [31..0] RO  note
--  =============================================================================

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.unb2_board_pkg.ALL;

ENTITY unb2_board_system_info_reg IS
  GENERIC (
    g_design_name : STRING;
    g_stamp_date  : NATURAL := 0;
    g_stamp_time  : NATURAL := 0;
    g_stamp_svn   : NATURAL := 0;
    g_design_note : STRING
  );
  PORT (
    -- Clocks and reset
    mm_rst      : IN  STD_LOGIC;    
    mm_clk      : IN  STD_LOGIC;
    
    -- Memory Mapped Slave 
    sla_in      : IN  t_mem_mosi; 
    sla_out     : OUT t_mem_miso; 
    
    info        : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0)
    );
END unb2_board_system_info_reg;


ARCHITECTURE rtl OF unb2_board_system_info_reg IS

  CONSTANT c_nof_fixed_regs       : NATURAL := 2; -- info
  CONSTANT c_nof_design_name_regs : NATURAL := 8; -- design_name
  CONSTANT c_nof_stamp_regs       : NATURAL := 3; -- date, time, svn rev
  CONSTANT c_nof_design_note_regs : NATURAL := 8; -- note

  CONSTANT c_nof_regs             : NATURAL := c_nof_fixed_regs + c_nof_design_name_regs + c_nof_stamp_regs + c_nof_design_note_regs;

  CONSTANT c_mm_reg        : t_c_mem := (latency  => 1,
                                         adr_w    => ceil_log2(c_nof_regs),
                                         dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                         nof_dat  => c_nof_regs,
                                         init_sl  => '0');   

  CONSTANT c_use_phy_w     : NATURAL := 8; -- FIXME: to be removed
  CONSTANT c_use_phy       : STD_LOGIC_VECTOR(c_use_phy_w-1 DOWNTO 0) := (OTHERS=> '0'); -- FIXME: to be removed

  CONSTANT c_design_name    : t_slv_32_arr(0 TO c_nof_design_name_regs-1) := str_to_ascii_slv_32_arr(g_design_name, c_nof_design_name_regs);
  CONSTANT c_design_note    : t_slv_32_arr(0 TO c_nof_design_note_regs-1) := str_to_ascii_slv_32_arr(g_design_note, c_nof_design_note_regs);
 
BEGIN

  p_mm_reg : PROCESS (mm_rst, mm_clk)
    VARIABLE vA : NATURAL := 0;
  BEGIN
    IF mm_rst = '1' THEN
      -- Read access
      sla_out   <= c_mem_miso_rst;
    ELSIF rising_edge(mm_clk) THEN
      -- Read access defaults
      sla_out.rdval <= '0';
           
      -- Read access: get register value
      IF sla_in.rd = '1' THEN
        sla_out       <= c_mem_miso_rst;    -- set unused rddata bits to '0' when read
        sla_out.rdval <= '1';               -- c_mm_reg.latency = 1

        vA := TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0));
        IF vA = 0 THEN         
          sla_out.rddata(c_word_w-1 DOWNTO 0) <= info;
          -- Use bit 11 to indicate that we're using the MM bus (not the info SLV).
          -- Using the MM bus enables user to also read use_phy, design_name etc.
          sla_out.rddata(11) <= '1';
        ELSIF vA = 1 THEN  
          sla_out.rddata(c_use_phy_w-1 DOWNTO 0) <= c_use_phy;
        ELSIF vA < c_nof_fixed_regs + c_nof_design_name_regs THEN      
          sla_out.rddata(c_word_w-1 DOWNTO 0) <= c_design_name(vA-c_nof_fixed_regs);

        ELSIF vA = c_nof_fixed_regs + c_nof_design_name_regs THEN      
          sla_out.rddata(c_word_w-1 DOWNTO 0) <= TO_UVEC(g_stamp_date, c_word_w);

        ELSIF vA = c_nof_fixed_regs + c_nof_design_name_regs+1 THEN      
          sla_out.rddata(c_word_w-1 DOWNTO 0) <= TO_UVEC(g_stamp_time, c_word_w);

        ELSIF vA = c_nof_fixed_regs + c_nof_design_name_regs+2 THEN      
          sla_out.rddata(c_word_w-1 DOWNTO 0) <= TO_UVEC(g_stamp_svn, c_word_w);

        ELSIF vA < c_nof_fixed_regs + c_nof_design_name_regs+c_nof_stamp_regs+c_nof_design_note_regs THEN      
          sla_out.rddata(c_word_w-1 DOWNTO 0) <= c_design_note(vA-c_nof_fixed_regs-c_nof_design_name_regs-c_nof_stamp_regs);

        END IF;

      END IF;
    END IF;
  END PROCESS;

 
END rtl;

