-------------------------------------------------------------------------------
--
-- Copyright (C) 2012-2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: 
--   Manually override WDI to initiate reconfiguratioon of the FPGA.
--   Write 0xB007FAC7 to address 0x0.

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY unb2_board_wdi_reg IS
  PORT (
    -- Clocks and reset
    mm_rst            : IN  STD_LOGIC;   -- reset synchronous with mm_clk
    mm_clk            : IN  STD_LOGIC;   -- memory-mapped bus clock
     
    -- Memory Mapped Slave in mm_clk domain
    sla_in            : IN  t_mem_mosi;  -- actual ranges defined by c_mm_reg
    sla_out           : OUT t_mem_miso;  -- actual ranges defined by c_mm_reg
    
    -- MM registers in st_clk domain
    wdi_override      : OUT STD_LOGIC
 );
END unb2_board_wdi_reg;


ARCHITECTURE rtl OF unb2_board_wdi_reg IS

  -- Define the actual size of the MM slave register
  CONSTANT c_mm_reg : t_c_mem := (latency  => 1,
                                  adr_w    => ceil_log2(1),
                                  dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                  nof_dat  => 1,
                                  init_sl  => '0');                       

  -- For safety, WDI override requires the following word to be written:
  CONSTANT c_cmd_reconfigure : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0 ) := x"B007FAC7"; -- "Boot factory"  

BEGIN

  p_mm_reg : PROCESS (mm_rst, mm_clk)
  BEGIN
    IF mm_rst = '1' THEN
      -- Read access
      sla_out   <= c_mem_miso_rst;
      -- Write access, register values
        wdi_override <= '0';
    ELSIF rising_edge(mm_clk) THEN
      -- Read access defaults: unused
      sla_out   <= c_mem_miso_rst;   
   
      -- Write access: set register value
      IF sla_in.wr = '1' THEN
        CASE TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0)) IS
          -- Write Block Sync
          WHEN 0 => 
            IF sla_in.wrdata(c_word_w-1 DOWNTO 0) = c_cmd_reconfigure THEN
              wdi_override <= '1';
            ELSE
              wdi_override <= '0';
            END IF;
          WHEN OTHERS => NULL;  -- unused MM addresses
        END CASE;
      END IF;
  
    END IF;
  END PROCESS;

END rtl;

