--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose:
--   Prepend one word with empty symbols at the head of a frame
-- Description:
--   This function is typically used to change the word boundary in a frame by
--   inserting g_nof_padding symbols at the head of the frame.
--   E.g. the total ethernet header is 42 octets and then followed by the
--   payload. This dp_pad_insert can insert 2 or 6 octets in front of the
--   header to make it 48 octets and hence make the payload start at a 32 or
--   64 bit word boundary respectively.
-- Remark:
-- . Ready latency (RL) = 1
-- . The padding preserves the sync and BSN from the snk_in.
-- . Reverse operation is dp_pad_remove, which is merely dp_split.
-- . This dp_pad_insert now can only insert 1 word at the head, this could be
--   extended to allow also g_nof_padding > g_data_w/g_symbol_w


ENTITY dp_pad_insert IS
  GENERIC (
    g_data_w          : NATURAL := 64;
    g_symbol_w        : NATURAL := 8;   -- g_data_w/g_symbol_w must be integer
    g_nof_padding     : NATURAL := 6;   -- must be > 0 and  <= g_data_w/g_symbol_w
    g_internal_bypass : BOOLEAN := FALSE -- can be used to eliminate external GENERATE-statements/wiring
  );
  PORT (
    rst     : IN  STD_LOGIC;
    clk     : IN  STD_LOGIC;
    -- ST sinks
    snk_out : OUT t_dp_siso;
    snk_in  : IN  t_dp_sosi;
    -- ST source
    src_in  : IN  t_dp_siso;
    src_out : OUT t_dp_sosi
  );
END dp_pad_insert;


ARCHITECTURE str OF dp_pad_insert IS

  CONSTANT c_symbols_per_beat     : NATURAL := g_data_w / g_symbol_w;
  CONSTANT c_pad_empty            : NATURAL := c_symbols_per_beat - g_nof_padding;
  
  -- dp padding is one word with g_nof_padding octets and c_pad_empty empty
  CONSTANT c_pad_sosi    : t_dp_sosi := ('0', (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), '1', '1', '1', TO_DP_EMPTY(c_pad_empty), (OTHERS=>'0'), (OTHERS=>'0'));
  
  SIGNAL pad_siso        : t_dp_siso;
  SIGNAL pad_sosi        : t_dp_sosi;
  SIGNAL nxt_pad_sosi    : t_dp_sosi;
  
  SIGNAL in_siso         : t_dp_siso;
  SIGNAL next_in_sosi    : t_dp_sosi;
  SIGNAL pend_in_sosi    : t_dp_sosi;
  SIGNAL in_sosi         : t_dp_sosi;
  
  -- element (0) becomes the head, element (1) becomes the tail
  SIGNAL concat_siso_arr : t_dp_siso_arr(0 TO 1);
  SIGNAL concat_sosi_arr : t_dp_sosi_arr(0 TO 1);
  
BEGIN

  no_bypass : IF g_internal_bypass = FALSE GENERATE

    -- Create the padding input word, while preserving the sync and BSN from the snk_in
    p_clk : PROCESS(rst, clk)
    BEGIN
      IF rst='1' THEN
        pad_sosi <= c_dp_sosi_rst;
        in_sosi  <= c_dp_sosi_rst;
      ELSIF rising_edge(clk) THEN
        pad_sosi <= nxt_pad_sosi;
        in_sosi  <= next_in_sosi;
      END IF;
    END PROCESS;
    
    p_pad_sosi : PROCESS(pad_siso, pend_in_sosi)
    BEGIN
      nxt_pad_sosi <= c_dp_sosi_rst;
      -- Provide the padding word when the output is ready and the input is pending, use the sync, bsn and channel from the input
      IF pad_siso.ready='1' AND pend_in_sosi.sop='1' THEN
        nxt_pad_sosi         <= c_pad_sosi;
        nxt_pad_sosi.sync    <= pend_in_sosi.sync;
        nxt_pad_sosi.bsn     <= pend_in_sosi.bsn;
        nxt_pad_sosi.channel <= pend_in_sosi.channel;
      END IF;
    END PROCESS;
    
    u_hold_snk_in : ENTITY work.dp_hold_input
    PORT MAP (
      rst              => rst,
      clk              => clk,
      -- ST sink
      snk_out          => snk_out,
      snk_in           => snk_in,
      -- ST source
      src_in           => in_siso,
      next_src_out     => next_in_sosi,
      pend_src_out     => pend_in_sosi,
      src_out_reg      => in_sosi
    );
    
    -- Prepend the padding octets to snk_in
    pad_siso           <= concat_siso_arr(0);
    in_siso            <= concat_siso_arr(1);
    concat_sosi_arr(0) <= pad_sosi;   -- = head frame with in_sosi sop info thanks to p_pad_sosi
    concat_sosi_arr(1) <= in_sosi;    -- = tail frame with in_sosi eop info
    
    u_concat : ENTITY work.dp_concat  -- RL = 1
    GENERIC MAP (
      g_data_w    => g_data_w,
      g_symbol_w  => g_symbol_w
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,
      -- ST sinks
      snk_out_arr => concat_siso_arr,
      snk_in_arr  => concat_sosi_arr,
      -- ST source
      src_in      => src_in,
      src_out     => src_out
    );  

  END GENERATE;

  gen_bypass : IF g_internal_bypass = TRUE GENERATE
    src_out <= snk_in;
    snk_out <= src_in;
  END GENERATE;  
  
END str;
