-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE work.dp_stream_pkg.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE work.tb_dp_pkg.ALL;

-- Purpose: Test bench to check deinterleave function on DP level
-- Usage:
--   > as 6
--   > run -all
-- Remark: 
--   This TB is only used to visually inspect the wave window to see
--   if all functions well on DP level. The actual deinterleaving itself
--   is verified more thoroughly in tb_common_reinterleave.

ENTITY tb_dp_deinterleave IS
  GENERIC (
    g_dat_w       : NATURAL := 16;
    g_nof_out     : NATURAL := 2;
    g_block_size  : NATURAL := 3;
    g_use_complex : BOOLEAN := TRUE
 );
END;

ARCHITECTURE tb OF tb_dp_deinterleave IS

  TYPE t_dsp_data_arr IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(g_dat_w/2-1 DOWNTO 0);
  TYPE t_data_arr     IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
  TYPE t_val_arr      IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC;

  -----------------------------------------------------------------------------
  -- Standard TB clocking, RST and control
  -----------------------------------------------------------------------------
  CONSTANT c_clk_period : TIME := 10 ns;

  SIGNAL clk            : STD_LOGIC := '1';
  SIGNAL rst            : STD_LOGIC;
  SIGNAL tb_end         : STD_LOGIC := '0';

  SIGNAL cnt_ena        : STD_LOGIC;

  -----------------------------------------------------------------------------
  --I/O streams
  -----------------------------------------------------------------------------
  CONSTANT c_input_val   : NATURAL := 192;
  CONSTANT c_input_inval : NATURAL := 64;

  SIGNAL snk_in  : t_dp_sosi;
  SIGNAL snk_out : t_dp_siso;

  SIGNAL src_out_arr : t_dp_sosi_arr(g_nof_out-1 DOWNTO 0);
  
  -----------------------------------------------------------------------------
  --Monitor signals for wave window
  -----------------------------------------------------------------------------
  SIGNAL in_data      :  STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
  SIGNAL in_re        :  STD_LOGIC_VECTOR(g_dat_w/2-1 DOWNTO 0);
  SIGNAL in_im        :  STD_LOGIC_VECTOR(g_dat_w/2-1 DOWNTO 0);
  SIGNAL in_val       :  STD_LOGIC;
  SIGNAL out_data_arr :  t_data_arr(g_nof_out-1 DOWNTO 0);
  SIGNAL out_re_arr   :  t_dsp_data_arr(g_nof_out-1 DOWNTO 0);
  SIGNAL out_im_arr   :  t_dsp_data_arr(g_nof_out-1 DOWNTO 0);
  SIGNAL out_val_arr  :  t_val_arr(g_nof_out-1 DOWNTO 0);
  
BEGIN

  -----------------------------------------------------------------------------
  -- Standard TB clocking, RST and control
  -----------------------------------------------------------------------------
  clk    <= NOT clk OR tb_end AFTER c_clk_period/2;
  rst    <= '1', '0' AFTER 3*c_clk_period;

  tb_end <= '0', '1' AFTER 2 us;
  
  -----------------------------------------------------------------------------
  -- Generate test data stream with counter data
  -----------------------------------------------------------------------------
  cnt_ena <= '0', '1' AFTER 20*c_clk_period;

  snk_out.ready <= '1';
  snk_out.xon <= '1';

  p_stimuli : PROCESS
    VARIABLE v_bsn : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0) := (OTHERS=>'0');
  BEGIN
    WHILE TRUE LOOP
      WAIT UNTIL rising_edge(clk);
      proc_dp_gen_block_data(1, sel_a_b(g_use_complex, FALSE, TRUE), g_dat_w, g_dat_w, 0, 0, 0, c_input_val, 0, 0, '1', v_bsn, clk, cnt_ena, snk_out, snk_in);
      proc_common_wait_some_cycles(clk, c_input_inval);
      v_bsn := INCR_UVEC(v_bsn, 1);
    END LOOP;
   END PROCESS;

  -----------------------------------------------------------------------------
  -- DUT
  -----------------------------------------------------------------------------
  u_deinterleave : ENTITY work.dp_deinterleave
  GENERIC MAP (
    g_nof_out      => g_nof_out,
    g_block_size   => g_block_size,
    g_dat_w        => g_dat_w,
    g_use_complex  => g_use_complex
  )
  PORT MAP (
    rst        => rst,
    clk        => clk,
    
    snk_in      => snk_in, 
    src_out_arr => src_out_arr
  );

  
  -----------------------------------------------------------------------------
  -- Wave window monitor
  -----------------------------------------------------------------------------
  in_data <= snk_in.data(g_dat_w-1 DOWNTO 0);
  in_re   <= snk_in.re(g_dat_w/2-1 DOWNTO 0);
  in_im   <= snk_in.im(g_dat_w/2-1 DOWNTO 0);
  in_val  <= snk_in.valid;

  gen_out : FOR I IN g_nof_out-1 DOWNTO 0 GENERATE
    out_data_arr(I)<= src_out_arr(I).data(g_dat_w-1 DOWNTO 0);
    out_re_arr(I)  <= src_out_arr(I).re(g_dat_w/2-1 DOWNTO 0);
    out_im_arr(I)  <= src_out_arr(I).im(g_dat_w/2-1 DOWNTO 0);
    out_val_arr(I) <= src_out_arr(I).valid;
  END GENERATE;
  
END tb;
