--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Keep a running sum for each element in snk_in_arr[i]
-- Description:
-- . A FIFO is used to store the running sums as wel as to output the
--   corresponding previous running sum so both align at the adder inputs.

ENTITY corr_accumulator IS
  GENERIC (
    g_sim                : BOOLEAN;
    g_nof_inputs         : NATURAL; -- Number of input streams
    g_nof_accumulators   : NATURAL; -- Number of accumulators to keep per input stream
    g_integration_period : NATURAL; -- Number of timesamples to accumulate
    g_data_w             : NATURAL; -- Complex input data width
    g_sim_model          : BOOLEAN := FALSE -- TRUE simulates a fast behavioural model instead of RTL (outputs only one int. period!)
  ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
    src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0)
  );
END corr_accumulator;

ARCHITECTURE str OF corr_accumulator IS

  -- Complex accumulator data width: wide enough to support g_integration_period accumulations
  CONSTANT c_acc_data_w               : NATURAL := g_data_w + ceil_log2(g_integration_period);
  CONSTANT c_adder_latency            : NATURAL := 2;
  CONSTANT c_fifo_latency             : NATURAL := 1;

  SIGNAL reg_snk_in_arr               :t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  -- Counter to keep track of the total numer of accumulations so we know when to reset accumulator to zero
  TYPE t_acc_cnt_arr IS ARRAY (g_nof_inputs-1 DOWNTO 0) OF STD_LOGIC_VECTOR(ceil_log2(g_nof_accumulators*g_integration_period)-1 DOWNTO 0);
  SIGNAL acc_cnt_arr                  : t_acc_cnt_arr;
  SIGNAL nxt_acc_cnt_arr              : t_acc_cnt_arr;

  SIGNAL corr_adder_snk_in_2arr_2     : t_dp_sosi_2arr_2(g_nof_inputs-1 DOWNTO 0); -- Array of pairs
  SIGNAL nxt_corr_adder_snk_in_2arr_2 : t_dp_sosi_2arr_2(g_nof_inputs-1 DOWNTO 0); -- Array of pairs
  SIGNAL corr_adder_src_out_arr       : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  TYPE t_dp_fifo_sc_usedw_arr      IS ARRAY(g_nof_inputs-1 DOWNTO 0) OF STD_LOGIC_VECTOR(ceil_log2(g_nof_accumulators)-1 DOWNTO 0);
  TYPE t_dp_fifo_sc_used_words_arr IS ARRAY(g_nof_inputs-1 DOWNTO 0) OF STD_LOGIC_VECTOR(ceil_log2(g_nof_accumulators)   DOWNTO 0);
  SIGNAL dp_fifo_sc_usedw_arr         : t_dp_fifo_sc_usedw_arr;
  SIGNAL dp_fifo_sc_used_words_arr    : t_dp_fifo_sc_used_words_arr;
  SIGNAL dp_fifo_sc_wr_ful_arr        : STD_LOGIC_VECTOR(g_nof_inputs-1 DOWNTO 0);

  SIGNAL dp_fifo_sc_src_in_arr        : t_dp_siso_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL dp_fifo_sc_src_out_arr       : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  SIGNAL dp_block_gen_snk_in_arr      : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL dp_block_gen_src_out_arr     : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  SIGNAL sim_snk_in_val_cnt           : NATURAL;
  SIGNAL nxt_sim_snk_in_val_cnt       : NATURAL;
  SIGNAL dp_split_src_out_2arr        : t_dp_sosi_2arr_2(g_nof_inputs-1 DOWNTO 0);
 
BEGIN

  gen_dsp: IF g_sim_model=FALSE GENERATE
    -----------------------------------------------------------------------------
    -- Adder inputs: current snk_in_arr + corresponding previous running sum
    -- from the FIFO buffer.
    -- When  we've completed one integration period for all the channels,
    -- we'll feed zeros into the second adder input to reset the accumulator value
    -- for each channel.
    -----------------------------------------------------------------------------
   gen_acc_counters : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      nxt_acc_cnt_arr(i) <= (OTHERS=>'0') WHEN TO_UINT(acc_cnt_arr(i))=g_nof_accumulators*g_integration_period-1 ELSE
                            INCR_UVEC(acc_cnt_arr(i), 1) WHEN reg_snk_in_arr(i).valid = '1' ELSE acc_cnt_arr(i);
    END GENERATE;
  
    gen_adder_inputs : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      -- Adder input 0: multiplier output
      nxt_corr_adder_snk_in_2arr_2(i)(0).valid <= reg_snk_in_arr(i).valid;
      nxt_corr_adder_snk_in_2arr_2(i)(0).re    <= RESIZE_DP_DSP_DATA(reg_snk_in_arr(i).re(g_data_w-1 DOWNTO 0));
      nxt_corr_adder_snk_in_2arr_2(i)(0).im    <= RESIZE_DP_DSP_DATA(reg_snk_in_arr(i).im(g_data_w-1 DOWNTO 0));
  
      -- Adder input 1: FIFO buffer output
      nxt_corr_adder_snk_in_2arr_2(i)(1).valid <= dp_fifo_sc_src_out_arr(i).valid;
      nxt_corr_adder_snk_in_2arr_2(i)(1).re    <= (OTHERS=>'0') WHEN TO_UINT(acc_cnt_arr(i))<=g_nof_accumulators-1 ELSE dp_fifo_sc_src_out_arr(i).re;
      nxt_corr_adder_snk_in_2arr_2(i)(1).im    <= (OTHERS=>'0') WHEN TO_UINT(acc_cnt_arr(i))<=g_nof_accumulators-1 ELSE dp_fifo_sc_src_out_arr(i).im;
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Complex adder stage
    -----------------------------------------------------------------------------
    u_corr_adder : ENTITY work.corr_adder
    GENERIC MAP (
      g_nof_inputs => g_nof_inputs,
      g_data_w     => c_acc_data_w
    )
    PORT MAP (
      clk            => clk,
      rst            => rst,
  
      snk_in_2arr_2  => corr_adder_snk_in_2arr_2,
      src_out_arr    => corr_adder_src_out_arr
    );  
  
    -----------------------------------------------------------------------------
    -- FIFOs to store the accumulator sums
    -----------------------------------------------------------------------------
    gen_dp_fifo_sc : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      u_dp_fifo_sc : ENTITY dp_lib.dp_fifo_sc
      GENERIC MAP (
        g_data_w         => 2*c_acc_data_w,
        g_use_ctrl       => FALSE,
        g_use_complex    => TRUE,
        g_fifo_size      => g_nof_accumulators, --Note: Even though we start reading the FIFO at g_nof_accumulators-c_adder_latency-c_fifo_latency-1 words,
        g_fifo_af_margin => 1                   --we still need g_nof_accumulators of capacity as 3 cycles of valid data is stored in the reg_snk_in and adder registers.
      )                                         --These 3 words may fill the FIFO while it is not being read out (in case of gaps in the input data).
      PORT MAP (
        rst         => rst,
        clk         => clk,

        wr_ful      => dp_fifo_sc_wr_ful_arr(i),
        usedw       => dp_fifo_sc_usedw_arr(i),
  
        snk_in      => corr_adder_src_out_arr(i),
        src_in      => dp_fifo_sc_src_in_arr(i),
        src_out     => dp_fifo_sc_src_out_arr(i)
      );

      -- The FIFO does not output the MSbit in its usedw output; it is output as wr_ful. Restore the full number of words here.
      dp_fifo_sc_used_words_arr(i) <= dp_fifo_sc_wr_ful_arr(i) & dp_fifo_sc_usedw_arr(i);
  
      -- Shift out the first accumulated value when it aligns with the corresponding current value at the adder input
      -- . this produces the dp_fifo_sc_src_out_arr data in sync with reg_snk_in_arr data
      dp_fifo_sc_src_in_arr(i).ready <= snk_in_arr(i).valid WHEN TO_UINT(dp_fifo_sc_used_words_arr(i))>=g_nof_accumulators-c_adder_latency-c_fifo_latency-1 ELSE '0';
  
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Output g_integration_period per stream once per integration period
    -- . The first g_nof_accumulators words are output. At the same time, the
    --   accumulators are reset (zeros) at the adder inputs.
    -- . Make sure the FIFO output is valid too so we don't output a block
    --   of zeros initially.
    -----------------------------------------------------------------------------
    gen_src_out_arr : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      dp_block_gen_snk_in_arr(i).re <= RESIZE_DP_DSP_DATA(dp_fifo_sc_src_out_arr(i).re);
      dp_block_gen_snk_in_arr(i).im <= RESIZE_DP_DSP_DATA(dp_fifo_sc_src_out_arr(i).im);
      
      dp_block_gen_snk_in_arr(i).valid <= '1' WHEN TO_UINT(acc_cnt_arr(i))<g_nof_accumulators   AND dp_fifo_sc_src_out_arr(i).valid='1' ELSE '0';
    END GENERATE;  

  END GENERATE;

  -----------------------------------------------------------------------------
  -- Simulation: fast basic behavioural model of accumulator. This assumes the
  -- input time samples are repeated (t0 = t1 = 2t = ..) so we can simply 
  -- multiply by the integration period and provide the result immediately.
  -- This saves a lot of simulation time.
  -- . NOTE: ONLY ONE INTEGRATION PERIOD IS OUTPUT.
  -----------------------------------------------------------------------------
  gen_sim_model: IF g_sim_model=TRUE GENERATE
    -- Count the number of incoming valid words
    p_val_cnt : PROCESS(snk_in_arr(0), sim_snk_in_val_cnt)
    BEGIN
      nxt_sim_snk_in_val_cnt <= sim_snk_in_val_cnt;
      IF snk_in_arr(0).valid = '1' THEN
        nxt_sim_snk_in_val_cnt <= sim_snk_in_val_cnt+1;
      END IF;

      FOR i IN 0 TO g_nof_inputs-1 LOOP
        IF snk_in_arr(i).valid = '1' AND sim_snk_in_val_cnt<g_nof_accumulators THEN
          dp_block_gen_snk_in_arr(i).re <= RESIZE_DP_DSP_DATA(TO_SVEC(g_integration_period*TO_SINT(snk_in_arr(i).re),c_acc_data_w));
          dp_block_gen_snk_in_arr(i).im <= RESIZE_DP_DSP_DATA(TO_SVEC(g_integration_period*TO_SINT(snk_in_arr(i).im),c_acc_data_w));
          dp_block_gen_snk_in_arr(i).valid <= '1';
        ELSE
          dp_block_gen_snk_in_arr(i) <= c_dp_sosi_rst;
        END IF;
      END LOOP;
    END PROCESS;

  END GENERATE;

  -----------------------------------------------------------------------------
  -- Add SOP and EOP for easier verification
  -----------------------------------------------------------------------------
  gen_sim_sop_eop : IF g_sim = TRUE GENERATE
    gen_dp_block_gen : FOR i IN 0 TO g_nof_inputs-1 GENERATE
      u_dp_block_gen: ENTITY dp_lib.dp_block_gen
      GENERIC MAP (
        g_use_src_in         => FALSE,
        g_nof_data           => g_nof_accumulators,
        g_nof_blk_per_sync   => g_integration_period 
      )
      PORT MAP (
        rst        => rst,
        clk        => clk,
    
        snk_in     => dp_block_gen_snk_in_arr(i),  
        src_out    => dp_block_gen_src_out_arr(i)
      );
    END GENERATE;
  
    src_out_arr <= dp_block_gen_src_out_arr;
  END GENERATE;

  no_sim_sop_eop : IF g_sim = FALSE GENERATE
    src_out_arr <= dp_block_gen_snk_in_arr;
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Registers
  -----------------------------------------------------------------------------
  p_clk : PROCESS (clk, rst)
  BEGIN
    IF rst='1' THEN
      acc_cnt_arr              <= (OTHERS=>(OTHERS=>'0'));
      corr_adder_snk_in_2arr_2 <= (OTHERS=>(OTHERS=>c_dp_sosi_rst));
      reg_snk_in_arr           <= (OTHERS=>c_dp_sosi_rst);
      sim_snk_in_val_cnt       <= 0;
    ELSIF rising_edge(clk) THEN
      acc_cnt_arr              <= nxt_acc_cnt_arr;
      corr_adder_snk_in_2arr_2 <= nxt_corr_adder_snk_in_2arr_2;
      reg_snk_in_arr           <= snk_in_arr;
      sim_snk_in_val_cnt       <= nxt_sim_snk_in_val_cnt;
    END IF;
  END PROCESS;

END str;
