--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- Description:
-- . nof_outputs = ceil_div(g_nof_inputs, 2^(g_nof_folds)) for g_nof_folds>=0
--   . Examples:
--     . g_nof_inputs=10, g_nof_folds=0 -> nof_outputs=10
--     . g_nof_inputs=10, g_nof_folds=1 -> nof_outputs= 5
--     . g_nof_inputs=10, g_nof_folds=2 -> nof_outputs= 3
--     . g_nof_inputs=10, g_nof_folds=3 -> nof_outputs= 2
--     . g_nof_inputs=10, g_nof_folds=4 -> nof_outputs= 1
--     . g_nof_inputs=10, g_nof_folds<0 -> nof_outputs= 1

ENTITY corr_folder IS
  GENERIC (
    g_nof_inputs : NATURAL; -- Number of inputs
    g_nof_folds  : INTEGER := -1 -- >0: Number of folds;
  );                             --  0: Wire out to in; 
  PORT (                         -- <0: Fold until one output remains
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

    src_out_arr    : OUT t_dp_sosi_arr(sel_a_b(g_nof_folds>=0, ceil_div(g_nof_inputs, ceil_pow2(g_nof_folds)), 1)-1 DOWNTO 0)
  );
END corr_folder;

ARCHITECTURE str OF corr_folder IS

  COMPONENT corr_folder IS
    GENERIC (
      g_nof_inputs : NATURAL;
      g_nof_folds  : INTEGER := -1
    ); 
    PORT (
      rst            : IN  STD_LOGIC;
      clk            : IN  STD_LOGIC;
  
      snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  
      src_out_arr    : OUT t_dp_sosi_arr(sel_a_b(g_nof_folds>=0, ceil_div(g_nof_inputs, ceil_pow2(g_nof_folds)), 1)-1 DOWNTO 0)
    );
  END COMPONENT;

  CONSTANT c_nof_muxes : NATURAL := ceil_div(g_nof_inputs, 2); 

  SIGNAL mux_snk_in_arr      : t_dp_sosi_arr(2*c_nof_muxes-1 DOWNTO 0);
  SIGNAL mux_snk_in_2arr_2   : t_dp_sosi_2arr_2(c_nof_muxes-1 DOWNTO 0);

  SIGNAL mux_src_out_arr     : t_dp_sosi_arr(c_nof_muxes-1 DOWNTO 0);
  SIGNAL nxt_mux_src_out_arr : t_dp_sosi_arr(c_nof_muxes-1 DOWNTO 0);

BEGIN

  gen_arch: IF g_nof_folds/=0 GENERATE
    -----------------------------------------------------------------------------
    -- Wire input array to mux_snk_in_arr to make sure we have an even number
    -- of buses to work with in case of an odd number of inputs
    -- . We want an even number of buses because we will wire up 2-input muxes
    -----------------------------------------------------------------------------
    gen_even_nof_buses: FOR i IN 0 TO g_nof_inputs-1 GENERATE   
      mux_snk_in_arr(i) <= snk_in_arr(i);
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Wire inputs to the 2-input muxes
    -----------------------------------------------------------------------------
    gen_mux_inputs_0: FOR i IN 0 TO c_nof_muxes-1 GENERATE
      mux_snk_in_2arr_2(i)(0) <= mux_snk_in_arr(2*i);
      mux_snk_in_2arr_2(i)(1) <= mux_snk_in_arr(2*i+1);
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Simple 2-input mux logic
    -----------------------------------------------------------------------------
    gen_mux_comb: FOR i IN 0 TO c_nof_muxes-1 GENERATE
      nxt_mux_src_out_arr(i) <= mux_snk_in_2arr_2(i)(0) WHEN mux_snk_in_2arr_2(i)(0).valid='1' ELSE
                                mux_snk_in_2arr_2(i)(1) WHEN mux_snk_in_2arr_2(i)(1).valid='1' ELSE
                                c_dp_sosi_rst;
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- If c_nof_muxes=1 or g_nof_folds=1, this is the last stage. Otherwise,
    -- add a stage.
    -- . g_nof_folds <0 : user wants to fold all the way to one output
    -- . g_nof_folds >0 : user wants to fold n times
    -----------------------------------------------------------------------------
    gen_corr_folder: IF (g_nof_folds<0 AND c_nof_muxes>1) OR g_nof_folds>1 GENERATE
      u_corr_folder : corr_folder
      GENERIC MAP (
        g_nof_inputs => c_nof_muxes,
        g_nof_folds  => g_nof_folds-1
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
                    
        snk_in_arr  => mux_src_out_arr,                   
        src_out_arr => src_out_arr
      );
    END GENERATE;
  
    gen_src_out_arr: IF (g_nof_folds<0 AND c_nof_muxes=1) OR g_nof_folds=1 GENERATE
      src_out_arr <= mux_src_out_arr;
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Registers
    -----------------------------------------------------------------------------
    p_clk: PROCESS(clk, rst)
    BEGIN
      IF rst='1' THEN
        mux_src_out_arr <= (OTHERS=>c_dp_sosi_rst);
      ELSIF rising_edge(clk) THEN
        mux_src_out_arr <= nxt_mux_src_out_arr;
      END IF;
    END PROCESS;

  END GENERATE;

  -----------------------------------------------------------------------------
  -- Wire output to input if g_nof_folds=0
  -----------------------------------------------------------------------------
  gen_wire_out_to_in: IF g_nof_folds=0 GENERATE
    src_out_arr <= snk_in_arr;
  END GENERATE;

END str;

