--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Buffer the visibilities during folding 
-- Description:
-- . Consists of an array of FIFOs that are read out periodically.
-- . The output data rate is such that all output streams can be folded into
--   one stream (valid cycles/stream never occur in parallel)
-- . In addition, the FIFOs are read out such that the folder outputs channel
--   blocks of visibilities.
-- . Finally, the output stream is toggled to produce a constant flow
--   instead of a short burst (to the extent that it is allowed by the 
--   integration time - a short integration time leaves little headroom).

ENTITY corr_visibility_buffer IS
  GENERIC (
    g_nof_inputs          : NATURAL; -- Number of input streams
    g_buffer_depth        : NATURAL; -- Number of words to buffer
    g_data_w              : NATURAL; -- Complex input data width
    g_nof_pre_mult_folds  : NATURAL; -- Nof times the data has been folded
    g_inter_channel_delay : NATURAL  -- Nof delay cycles between output channels
   ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
    src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0)
  );
END corr_visibility_buffer;

ARCHITECTURE str OF corr_visibility_buffer IS

  ------------------------------------------------------------------------------
  -- Example input: 6 visibilities, non-folded (g_nof_inputs = 6, 
  -- g_nof_pre_mult_folds = 0)
  -- The improtance of timing the FIFOs is depiced below:
  -- ====================
  -- Untimed FIFO output:
  -- ====================
  --  dp_fifo_sc_src_out_arr(0) -> [ch0, ch1, .. chN-1] 
  --  dp_fifo_sc_src_out_arr(1) -> [ch0, ch1, .. chN-1] 
  --  dp_fifo_sc_src_out_arr(2) -> [ch0, ch1, .. chN-1] 
  --  dp_fifo_sc_src_out_arr(3) -> [ch0, ch1, .. chN-1] 
  --  dp_fifo_sc_src_out_arr(4) -> [ch0, ch1, .. chN-1] 
  --  dp_fifo_sc_src_out_arr(5) -> [ch0, ch1, .. chN-1] 
  -- ====================
  -- Timed FIFO output:
  -- ====================
  --  dp_fifo_sc_src_out_arr(0) -> [ch0]                   [ch1]                   
  --  dp_fifo_sc_src_out_arr(1) ->     [ch0]                   [ch1]               
  --  dp_fifo_sc_src_out_arr(2) ->         [ch0]                   [ch1]           
  --  dp_fifo_sc_src_out_arr(3) ->             [ch0]                   [ch1]       
  --  dp_fifo_sc_src_out_arr(4) ->                 [ch0]                   [ch1]   
  --  dp_fifo_sc_src_out_arr(5) ->                     [ch0]                   [ch1]
  --
  -- The above timed FIFO output allows the 6 streams to be interleaved into
  -- one stream, putting the channels back-to-back in the right order.

  -- The number of visibilities interleaved onto one input stream
  CONSTANT c_vis_block_len        : NATURAL := pow2(g_nof_pre_mult_folds);

  -- The minimum period to wait so the folder does not miss data
  CONSTANT c_channel_block_length : NATURAL := g_nof_inputs*c_vis_block_len;

  SIGNAL dp_fifo_sc_src_out_arr   : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL dp_fifo_sc_src_in_arr    : t_dp_siso_arr(g_nof_inputs-1 DOWNTO 0);

BEGIN

  ------------------------------------------------------------------------------
  -- Buffer the visibilities before the folding stage.
  -- . All accumulators output blocks of visibilities at the same time. We need
  --   to buffer the visibilities to source them to the folder at the right
  --   pace.
  ------------------------------------------------------------------------------
  gen_dp_fifo_sc : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_dp_fifo_sc : ENTITY dp_lib.dp_fifo_sc
    GENERIC MAP (
      g_data_w         => 2*g_data_w,
      g_use_ctrl       => FALSE,
      g_use_complex    => TRUE,
      g_fifo_size      => g_buffer_depth,
      g_fifo_af_margin => 0
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,

      wr_ful      => OPEN,
      usedw       => OPEN,
      rd_emp      => OPEN,

      snk_out     => OPEN,
      snk_in      => snk_in_arr(i),
      src_in      => dp_fifo_sc_src_in_arr(i),
      src_out     => dp_fifo_sc_src_out_arr(i)
    );
  END GENERATE;

  ------------------------------------------------------------------------------
  -- Read out the FIFOs such that dp_fifo_sc_src_out_arr only carries one valid 
  -- word per cycle. This allows folding (multiplexing) into one stream without 
  -- losing data (example for 16 streams with c_vis_block_len=1):            
  --                                                                            
  --     |<-----------------------g_block_period-------------------->|           
  --     |<---c_channel_block_length---->|<--g_inter_channel_delay-->|          
  --  0) [0]                             .                           [1]        
  --  1) | [0]                           .                           . [1]      
  --  2) |   [0]                         .                           .   [1]    
  --  3) |     [0]                       .                           .     [1]  
  --  4) |       [0]                     .                           .       [1]
  --  5) |         [0]                   .                           .        ..
  --  6) |           [0]                 .                           .          
  --  7) |             [0]               .                           .          
  --  8) |               [0]             .                           .          
  --  9) |                 [0]           .                           .          
  -- 10) |                   [0]         .                           .          
  -- 11) |                     [0]       .                           .          
  -- 12) |                       [0]     .                           .          
  -- 13) |                         [0]   .                           .          
  -- 14) |                           [0] .                           .          
  -- 15) |<---g_init_valid_delay(15)-->[0]                           .          
  ------------------------------------------------------------------------------
  gen_dp_src_out_timer : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_dp_src_out_timer : ENTITY dp_lib.dp_src_out_timer
    GENERIC MAP (
      g_init_valid_delay => i*c_vis_block_len, --relative to dp_fifo_sc_src_out_arr(i).valid
      g_block_period     => c_channel_block_length+g_inter_channel_delay,
      g_block_len        => c_vis_block_len
    )
    PORT MAP (
      rst                  => rst,
      clk                  => clk,

      init_valid_delay_ref => snk_in_arr(i).valid,

      snk_in               => dp_fifo_sc_src_out_arr(i),
      snk_out              => dp_fifo_sc_src_in_arr(i)
    );
  END GENERATE;

  src_out_arr <= dp_fifo_sc_src_out_arr;

END str;
