-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
-- Purpose:   Performing a poly phase prefilter function on (multiple) wideband data stream.
--
--
-- Description: The poly phase prefilter function is applied on multiple inputs.
--              The incoming data must be divided over the inputs as shown in 
--              the following example where the wideband factor is 4 (4 inputs).
--
--              Stream input 0 : t0, t4,  t8, t12, t16...
--              Stream input 1 : t1, t5,  t9, t13, t17...
--              Stream input 2 : t2, t6, t10, t14, t18...
--              Stream input 3 : t3, t7, t11, t15, t19...
--              
--              Every input stream will be filtered by a single channel poly
--              phase prefilter unit. 
--              It is also possible to offer multiple wideband input streams. Those
--              input streams will share the filtercoefficients. For a
--              system with wb_factor=4 and nof_streams=2 the following stream 
--              inputs should be offered (s0t0 means stream 0, timestamp 0):
--
--              Stream input 0 : s0t0, s0t4, s0t8 , s0t12, s0t16...
--              Stream input 1 : s1t0, s1t4, s1t8 , s1t12, s1t16...
--              Stream input 2 : s0t1, s0t5, s0t9 , s0t13, s0t17...
--              Stream input 3 : s1t1, s1t5, s1t9 , s1t13, s1t17...
--              Stream input 4 : s0t2, s0t6, s0t10, s0t14, s0t18...
--              Stream input 5 : s1t2, s1t6, s1t10, s1t14, s1t18...
--              Stream input 6 : s0t3, s0t7, s0t11, s0t15, s0t19...
--              Stream input 7 : s1t3, s1t7, s1t11, s1t15, s1t19...

--
-- Remarks:    .
--              

library IEEE, common_lib;
use IEEE.std_logic_1164.ALL;
use IEEE.numeric_std.ALL;
use common_lib.common_pkg.ALL; 
use common_lib.common_mem_pkg.ALL;  
use work.fil_pkg.ALL;

entity fil_ppf_wide is
  generic (
    g_fil_ppf           : t_fil_ppf          := c_fil_ppf;    
    g_fil_ppf_pipeline  : t_fil_ppf_pipeline := c_fil_ppf_pipeline; 
    g_file_index_arr    : t_nat_natural_arr  := array_init(0, 128, 1);  -- default use the instance index as file index 0, 1, 2, 3, 4 ...
    g_coefs_file_prefix : string             := "../../data/coef"       -- Relative path to the mif files that contain the initial data for the coefficients memories 
  );                                                                    -- The sequence number and ".mif"-extension are added within the entity.
  port (
    dp_clk         : in  std_logic;
    dp_rst         : in  std_logic;
    mm_clk         : in  std_logic;
    mm_rst         : in  std_logic;
    ram_coefs_mosi : in  t_mem_mosi;               
    ram_coefs_miso : out t_mem_miso := c_mem_miso_rst;
    in_dat_arr     : in  t_fil_slv_arr(g_fil_ppf.nof_streams*g_fil_ppf.wb_factor-1 downto 0);
    in_val         : in  std_logic;  
    out_dat_arr    : out t_fil_slv_arr(g_fil_ppf.nof_streams*g_fil_ppf.wb_factor-1 downto 0);
    out_val        : out std_logic
  ); 
end fil_ppf_wide;

architecture rtl of fil_ppf_wide is   

  type t_fil_ppf_arr      is array(integer range <> ) of t_fil_ppf;                       -- An array of t_fil_ppf's generics.                                                                              
  type t_nat_natural_arr2 is array(integer range <> ) of t_nat_natural_arr(127 downto 0); -- An array of arrays, used to point to the right .mif files for the coefficients
  type t_out_arr          is array(integer range <> ) of std_logic_vector(g_fil_ppf.nof_streams*g_fil_ppf.out_dat_w -1 downto 0);
  type t_in_arr           is array(integer range <> ) of std_logic_vector(g_fil_ppf.nof_streams*g_fil_ppf.in_dat_w  -1 downto 0);
  
  ----------------------------------------------------------
  -- This function creates an array of t_fil_ppf generics 
  -- for the single channel poly phase filters that are 
  -- used to compose the multichannel(wideband) poly phase
  -- filter. The array is based on the content of the g_fil_ppf
  -- generic that belongs to the fil_ppf_w entity. 
  -- Only the nof_bands is modified. 
  ----------------------------------------------------------
  function func_create_generics_for_ppfs(input: t_fil_ppf) return t_fil_ppf_arr is
    variable v_nof_bands : natural := input.nof_bands/input.wb_factor;                     -- The nof_bands for the single channel poly phase filters
    variable v_return    : t_fil_ppf_arr(input.wb_factor-1 downto 0) := (others => input); -- Variable that holds the return values
  begin
    for I in 0 to input.wb_factor-1 loop
      v_return(I).nof_bands := v_nof_bands;     -- The new number of bands
    end loop;                               
    return v_return; 
  end;
  
  ----------------------------------------------------------
  -- Function that divids the input file index array into 
  -- "wb_factor" new file index arrays. 
  ----------------------------------------------------------
  function func_create_file_index_array(input: t_nat_natural_arr; wb_factor: natural; nof_taps: natural) return t_nat_natural_arr2 is
    variable v_return : t_nat_natural_arr2(wb_factor-1 downto 0) := (others => input);    -- Variable that holds the return values
  begin
    for I in 0 to wb_factor-1 loop
      for J in 0 to nof_taps-1 loop
        v_return(I)(J) := input(I*nof_taps+J);
      end loop;
    end loop;                               
    return v_return; 
  end;
  
  constant c_fil_ppf_arr     : t_fil_ppf_arr(g_fil_ppf.wb_factor-1      downto 0) := func_create_generics_for_ppfs(g_fil_ppf);  
  constant c_file_index_arr2 : t_nat_natural_arr2(g_fil_ppf.wb_factor-1 downto 0) := func_create_file_index_array(g_file_index_arr, g_fil_ppf.wb_factor, g_fil_ppf.nof_taps);
  
  constant c_mem_addr_w      : natural := ceil_log2(g_fil_ppf.nof_bands * g_fil_ppf.nof_taps / g_fil_ppf.wb_factor); 

  signal ram_coefs_mosi_arr  : t_mem_mosi_arr(g_fil_ppf.wb_factor-1   downto 0);                               
  signal ram_coefs_miso_arr  : t_mem_miso_arr(g_fil_ppf.wb_factor-1   downto 0) := (others => c_mem_miso_rst); 
  signal out_val_i           : std_logic_vector(g_fil_ppf.wb_factor-1 downto 0);
  signal out_arr_i           : t_out_arr(g_fil_ppf.wb_factor-1        downto 0);
  signal in_dat_vect_arr     : t_in_arr(g_fil_ppf.wb_factor-1         downto 0);
  
begin
  ---------------------------------------------------------------
  -- COMBINE MEMORY MAPPED INTERFACES
  ---------------------------------------------------------------
  -- Combine the internal array of mm interfaces for the coefficents 
  -- memory to one array that is connected to the port of the fil_ppf_w
  u_mem_mux_coef : entity common_lib.common_mem_mux
  generic map (    
    g_nof_mosi    => g_fil_ppf.wb_factor,
    g_mult_addr_w => c_mem_addr_w
  )
  port map (
    mosi     => ram_coefs_mosi,
    miso     => ram_coefs_miso,
    mosi_arr => ram_coefs_mosi_arr,
    miso_arr => ram_coefs_miso_arr
  );

  gen_in_vect_wb : for I in 0 to g_fil_ppf.wb_factor-1 generate
    gen_in_vect_streams : for J in 0 to g_fil_ppf.nof_streams-1 generate
       --in_dat_vect_arr(I)((J+1)*g_fil_ppf.in_dat_w-1 downto J*g_fil_ppf.in_dat_w) <= in_dat_arr(J*g_fil_ppf.wb_factor+I)(g_fil_ppf.in_dat_w-1 downto 0);
       in_dat_vect_arr(I)((J+1)*g_fil_ppf.in_dat_w-1 downto J*g_fil_ppf.in_dat_w) <= in_dat_arr(I*g_fil_ppf.nof_streams+J)(g_fil_ppf.in_dat_w-1 downto 0);
    end generate;
  end generate;

  ---------------------------------------------------------------
  -- INSTANTIATE MULTIPLE SINGLE CHANNEL POLY PHASE FILTERS
  ---------------------------------------------------------------
  gen_ppf_singles : for I in 0 to g_fil_ppf.wb_factor-1 generate
    u_single_filter : entity work.fil_ppf_single 
    generic map (
      g_fil_ppf           => c_fil_ppf_arr(I),
      g_fil_ppf_pipeline  => g_fil_ppf_pipeline,
      g_file_index_arr    => c_file_index_arr2(I),
      g_coefs_file_prefix => g_coefs_file_prefix 
    )
    port map (
      dp_clk         => dp_clk,
      dp_rst         => dp_rst,
      mm_clk         => mm_clk,
      mm_rst         => mm_rst,
      ram_coefs_mosi => ram_coefs_mosi_arr(I), 
      ram_coefs_miso => ram_coefs_miso_arr(I), 
      in_dat         => in_dat_vect_arr(I),
      in_val         => in_val,
      out_dat        => out_arr_i(I),
      out_val        => out_val_i(I)
    ); 
    gen_in_vect_streams : for J in 0 to g_fil_ppf.nof_streams-1 generate   
      out_dat_arr(I*g_fil_ppf.nof_streams+J) <= RESIZE_SVEC(out_arr_i(I)((J+1)*g_fil_ppf.out_dat_w-1 downto J*g_fil_ppf.out_dat_w), out_dat_arr(I)'LENGTH);
    end generate;
  end generate;  
  
  out_val <= out_val_i(0);                             
  
end rtl;
