--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Declare IP libraries to ensure default binding in simulation. The IP library clause is ignored by synthesis.
LIBRARY ip_stratixiv_ddr3_uphy_4g_800_master_lib;
LIBRARY ip_stratixiv_ddr3_uphy_4g_800_slave_lib;

LIBRARY IEEE, common_lib, technology_lib, dp_lib, tech_ddr_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL; 
USE common_lib.common_mem_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE work.ddr3_pkg.ALL;

ENTITY ddr3 IS
  GENERIC(
    g_phy                     : NATURAL := 1;   -- 0: ALTMEMPHY  1: UNIPHY_MASTER 2: UNIPHY_SLAVE
    g_ddr                     : t_c_ddr3_phy;
    g_mts                     : NATURAL := 800; -- Megatransfers per second
    g_wr_data_w               : NATURAL := c_ddr3_ctlr_data_w;  
    g_wr_use_ctrl             : BOOLEAN := FALSE;              -- TRUE to allow filling the WR FIFO (by disabling flush) after an EOP
    g_wr_fifo_depth           : NATURAL := 128;                -- >=16 AND >c_ddr3_ctlr_maxburstsize                               , defined at read  side of write FIFO.
    g_rd_fifo_depth           : NATURAL := 256;                -- >=16 AND >c_ddr3_ctlr_maxburstsize > c_ddr3_ctrl_nof_latent_reads, defined at write side of read  FIFO. 
    g_rd_data_w               : NATURAL := c_ddr3_ctlr_data_w;
    g_flush_wr_fifo           : BOOLEAN := FALSE;              -- TRUE instantiates a dp_flush + controller to flush the write fifo when the driver is not ready to write
    g_flush_ext_ena           : BOOLEAN := FALSE;              -- TRUE enables the external flush_ena signal and discards flushing when driver is not ready. FALSE enables flushing when driver is not ready
    g_flush_sop               : BOOLEAN := FALSE;              -- When FALSE, flushing is stopped by valid data. When TRUE flushing is stopped by SOP
    g_flush_sop_sync          : BOOLEAN := FALSE;              -- When TRUE, flushing is stopped by receiving SOP and SYNC
    g_flush_sop_channel       : BOOLEAN := FALSE;              -- WHEN TRUE, flushing is stopped by receiving SOP and specified channel. 
    g_flush_sop_start_channel : NATURAL := 0;
    g_flush_nof_channels      : NATURAL := 0
  );                      
  PORT (
    -- MM clock + reset                                                             
    mm_rst             : IN    STD_LOGIC := '0';                                           
    mm_clk             : IN    STD_LOGIC := '0';                                                          
    
    ctlr_ref_clk       : IN    STD_LOGIC;
    ctlr_rst           : IN    STD_LOGIC; -- asynchronous reset input to controller

    ctlr_gen_clk       : OUT   STD_LOGIC; -- Controller generated clock
    ctlr_gen_rst       : OUT   STD_LOGIC;    
    ctlr_gen_clk_2x    : OUT   STD_LOGIC; -- Controller generated double frequency clock
    ctlr_gen_rst_2x    : OUT   STD_LOGIC; -- ctlr_gen_rst synchronized to ctlr_gen_clk_2x

    ctlr_init_done     : OUT   STD_LOGIC;
    ctlr_rdy           : OUT   STD_LOGIC;

    dvr_start_addr     : IN    t_ddr3_addr;
    dvr_end_addr       : IN    t_ddr3_addr;

    dvr_en             : IN    STD_LOGIC;
    dvr_wr_not_rd      : IN    STD_LOGIC;
    dvr_done           : OUT   STD_LOGIC;

    wr_clk             : IN    STD_LOGIC;
    wr_rst             : IN    STD_LOGIC; 
    
    -- MM register map for DDR controller status info
    reg_io_ddr_mosi    : IN    t_mem_mosi := c_mem_mosi_rst;
    reg_io_ddr_miso    : OUT   t_mem_miso := c_mem_miso_rst;
    
    flush_ena          : IN    STD_LOGIC; -- When toggled '1' the write-fifo flusher is enabled and stops flushing when configured trigger condition is met.

    wr_sosi            : IN    t_dp_sosi;
    wr_siso            : OUT   t_dp_siso;
  
    rd_sosi            : OUT   t_dp_sosi;
    rd_siso            : IN    t_dp_siso;
    
    rd_clk             : IN    STD_LOGIC;
    rd_rst             : IN    STD_LOGIC;

    rd_fifo_usedw      : OUT   STD_LOGIC_VECTOR(ceil_log2(g_rd_fifo_depth * (c_ddr3_ctlr_data_w/g_rd_data_w) )-1 DOWNTO 0);
    
    ser_term_ctrl_out  : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);
    par_term_ctrl_out  : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);

    ser_term_ctrl_in   : IN    STD_LOGIC_VECTOR(13 DOWNTO 0) := (OTHERS => '0');
    par_term_ctrl_in   : IN    STD_LOGIC_VECTOR(13 DOWNTO 0) := (OTHERS => '0');

    phy_in             : IN    t_tech_ddr3_phy_in;
    phy_io             : INOUT t_tech_ddr3_phy_io;
    phy_ou             : OUT   t_tech_ddr3_phy_ou
   );
END ddr3;


ARCHITECTURE str OF ddr3 IS  
 
  CONSTANT c_wr_fifo_depth  : NATURAL := g_wr_fifo_depth * (c_ddr3_ctlr_data_w/g_wr_data_w); -- Multiply fifo depth by the fifo's rd/wr width ratio to get write side depth
                            
  CONSTANT c_latency        : NATURAL := 1;

  CONSTANT c_ctlr_address_w : NATURAL := func_tech_ddr_ctlr_address_w(c_tech_ddr3_4g_800m_master);
  CONSTANT c_ctlr_data_w    : NATURAL := func_tech_ddr_ctlr_data_w(c_tech_ddr3_4g_800m_master);
                            
  SIGNAL ctlr_burst         : STD_LOGIC; 
  SIGNAL ctlr_burst_size    : STD_LOGIC_VECTOR(c_ddr3_ctlr_maxburstsize_w-1 DOWNTO 0);
  SIGNAL ctlr_rd_req        : STD_LOGIC;
  SIGNAL ctlr_wr_req        : STD_LOGIC;
                            
  SIGNAL ctlr_gen_rst_n     : STD_LOGIC;
                            
  SIGNAL i_ctlr_gen_clk     : STD_LOGIC;
  SIGNAL i_ctlr_gen_rst     : STD_LOGIC;
  SIGNAL i_ctlr_gen_clk_2x  : STD_LOGIC;
  SIGNAL i_ctlr_init_done   : STD_LOGIC;
  SIGNAL i_ctlr_rdy         : STD_LOGIC;
  SIGNAL i_dvr_done         : STD_LOGIC;
                            
  SIGNAL dvr_cur_addr       : t_ddr3_addr;
  SIGNAL dvr_flush          : STD_LOGIC := '0';
                            
  SIGNAL ctlr_wr_siso       : t_dp_siso := c_dp_siso_rdy;  -- default xon='1'
  SIGNAL ctlr_wr_sosi       : t_dp_sosi;
                            
  SIGNAL flush_wr_siso      : t_dp_siso;
  SIGNAL flush_wr_sosi      : t_dp_sosi;
                            
  SIGNAL ctlr_rd_siso       : t_dp_siso;
  SIGNAL ctlr_rd_sosi       : t_dp_sosi;
                            
  SIGNAL wr_fifo_usedw      : STD_LOGIC_VECTOR(ceil_log2(g_wr_fifo_depth)-1 DOWNTO 0);  -- read side depth of the write FIFO
                            
  SIGNAL local_cal_success  : STD_LOGIC;
  SIGNAL local_cal_fail     : STD_LOGIC;   
  
  CONSTANT c_mem_reg_io_ddr : t_c_mem := (c_mem_reg_rd_latency, 1 , 32 , 1, 'X');
  SIGNAL mm_reg_io_ddr      : STD_LOGIC_VECTOR(31 DOWNTO 0);                     

  SIGNAL term_ctrl_out      : t_tech_ddr3_phy_terminationcontrol;
  SIGNAL term_ctrl_in       : t_tech_ddr3_phy_terminationcontrol;

  SIGNAL ctlr_tech_mosi     : t_mem_ctlr_mosi;
  SIGNAL ctlr_tech_miso     : t_mem_ctlr_miso;
  
BEGIN 

  dvr_done <= i_dvr_done;

  ctlr_gen_clk    <= i_ctlr_gen_clk;
  ctlr_gen_rst    <= i_ctlr_gen_rst;
  ctlr_gen_clk_2x <= i_ctlr_gen_clk_2x;   
  ctlr_rdy        <= i_ctlr_rdy;  
  ctlr_init_done  <= i_ctlr_init_done;

  u_wr_fifo : ENTITY dp_lib.dp_fifo_dc_mixed_widths
  GENERIC MAP (
    g_wr_data_w         => g_wr_data_w,
    g_rd_data_w         => c_ddr3_ctlr_data_w,
    g_use_ctrl          => g_wr_use_ctrl,
    g_wr_fifo_size      => c_wr_fifo_depth,
    g_wr_fifo_af_margin => 4 + c_latency, --default (4) + c_latency to compensate for latency introduced by registering wr_siso.ready
    g_rd_fifo_rl        => 0
  )
  PORT MAP (
    wr_rst         => wr_rst,
    wr_clk         => wr_clk,
    rd_rst         => i_ctlr_gen_rst,
    rd_clk         => i_ctlr_gen_clk,

    snk_out        => wr_siso,
    snk_in         => wr_sosi,
  
    wr_usedw       => OPEN,
    rd_usedw       => wr_fifo_usedw,
    rd_emp         => OPEN,

    src_in         => flush_wr_siso,
    src_out        => flush_wr_sosi
  );

  u_dp_flush : ENTITY dp_lib.dp_flush -- Always instantiate the flusher as it also contains a RL adapter
  GENERIC MAP (
    g_ready_latency => 0,
    g_framed_xon    => g_wr_use_ctrl,  -- stop flushing when dvr_flush is low and a sop has arrived 
    g_framed_xoff   => FALSE           -- immediately start flushing when dvr_flush goes high
  )
  PORT MAP (
    rst      => i_ctlr_gen_rst,
    clk      => i_ctlr_gen_clk,
   
    snk_in   => flush_wr_sosi,
    snk_out  => flush_wr_siso,

    src_out  => ctlr_wr_sosi,
    src_in   => ctlr_wr_siso,  -- fixed streaming xon='1'

    flush_en => dvr_flush      -- memory mapped xon/xoff control
  );

  gen_flush : IF g_flush_wr_fifo = TRUE GENERATE  
    u_flush_ctrl : ENTITY work.ddr3_flush_ctrl
    GENERIC MAP (  
      g_ext_ena           => g_flush_ext_ena,
      g_sop               => g_flush_sop,
      g_sop_sync          => g_flush_sop_sync,
      g_sop_channel       => g_flush_sop_channel,
      g_sop_start_channel => g_flush_sop_start_channel,
      g_nof_channels      => g_flush_nof_channels     
    )
    PORT MAP (
      rst           => wr_rst,
      clk           => wr_clk,
  
      dvr_en        => dvr_en,
      dvr_wr_not_rd => dvr_wr_not_rd,
      dvr_done      => i_dvr_done,
  
      wr_sosi       => wr_sosi,
      flush_ena     => flush_ena,
  
      dvr_flush     => dvr_flush
    );
  END GENERATE;

  u_rd_fifo : ENTITY dp_lib.dp_fifo_dc_mixed_widths
  GENERIC MAP (
    g_wr_data_w         => c_ddr3_ctlr_data_w,
    g_rd_data_w         => g_rd_data_w,
    g_use_ctrl          => FALSE,
    g_wr_fifo_size      => g_rd_fifo_depth,
    g_wr_fifo_af_margin => c_ddr3_ctrl_nof_latent_reads, -- >=4 (required by dp_fifo)
    g_rd_fifo_rl        => 1
  )
  PORT MAP (
    wr_rst   => i_ctlr_gen_rst,
    wr_clk   => i_ctlr_gen_clk,
    rd_rst   => rd_rst,
    rd_clk   => rd_clk,

    snk_out  => ctlr_rd_siso,
    snk_in   => ctlr_rd_sosi,
  
    wr_usedw => OPEN,
    rd_usedw => rd_fifo_usedw,
    rd_emp   => OPEN,

    src_in   => rd_siso,
    src_out  => rd_sosi
  );

  u_ddr3_driver : ENTITY work.ddr3_driver
  GENERIC MAP (
    g_wr_fifo_depth => g_wr_fifo_depth,  
    g_ddr           => g_ddr
  )
  PORT MAP ( 
    rst             => i_ctlr_gen_rst,  
    clk             => i_ctlr_gen_clk,        

    ctlr_rdy        => i_ctlr_rdy,
    ctlr_init_done  => i_ctlr_init_done,
    ctlr_wr_req     => ctlr_wr_req,      
    ctlr_rd_req     => ctlr_rd_req,
    ctlr_burst      => ctlr_burst,
    ctlr_burst_size => ctlr_burst_size,

    wr_val          => ctlr_wr_sosi.valid, 
    wr_rdy          => ctlr_wr_siso.ready,
    rd_rdy          => ctlr_rd_siso.ready,

    cur_addr        => dvr_cur_addr,
    start_addr      => dvr_start_addr,
    end_addr        => dvr_end_addr, 

    dvr_en          => dvr_en,
    dvr_wr_not_rd   => dvr_wr_not_rd,
    dvr_done        => i_dvr_done,

    wr_fifo_usedw   => wr_fifo_usedw
  );

  u_reg_map : ENTITY common_lib.common_reg_r_w_dc
  GENERIC MAP (
    g_cross_clock_domain => TRUE,             -- : BOOLEAN := TRUE;  -- use FALSE when mm_clk and st_clk are the same, else use TRUE to cross the clock domain
    g_in_new_latency     => 0,                -- : NATURAL := 0;  -- >= 0
    g_readback           => FALSE,            -- : BOOLEAN := FALSE;  -- must use FALSE for write/read or read only register when g_cross_clock_domain=TRUE
    g_reg                => c_mem_reg_io_ddr, -- : t_c_mem := c_mem_reg;
    g_init_reg           => (OTHERS => '0')   -- : STD_LOGIC_VECTOR(c_mem_reg_init_w-1 DOWNTO 0) := (OTHERS => '0')
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst      => mm_rst,          --: IN  STD_LOGIC;   -- reset synchronous with mm_clk
    mm_clk      => mm_clk,          --: IN  STD_LOGIC;   -- memory-mapped bus clock
    st_rst      => i_ctlr_gen_rst,     --: IN  STD_LOGIC;   -- reset synchronous with st_clk
    st_clk      => i_ctlr_gen_clk,     --: IN  STD_LOGIC;   -- other clock domain clock
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in      => reg_io_ddr_mosi, --: IN  t_mem_mosi;  -- actual ranges defined by g_reg
    sla_out     => reg_io_ddr_miso, --: OUT t_mem_miso;  -- actual ranges defined by g_reg
    
    -- MM registers in st_clk domain
    reg_wr_arr  => OPEN,            --  : OUT STD_LOGIC_VECTOR(            g_reg.nof_dat-1 DOWNTO 0);
    reg_rd_arr  => OPEN,            --  : OUT STD_LOGIC_VECTOR(            g_reg.nof_dat-1 DOWNTO 0);
    in_new      => '1',             --  : IN  STD_LOGIC := '1';
    in_reg      => mm_reg_io_ddr,   --  : IN  STD_LOGIC_VECTOR(g_reg.dat_w*g_reg.nof_dat-1 DOWNTO 0);
    out_reg     => OPEN             --  : OUT STD_LOGIC_VECTOR(g_reg.dat_w*g_reg.nof_dat-1 DOWNTO 0)
  );

  mm_reg_io_ddr <= RESIZE_UVEC(local_cal_fail & local_cal_success & i_ctlr_gen_rst & flush_ena & i_ctlr_rdy & i_ctlr_init_done, 32);   


  ser_term_ctrl_out <= term_ctrl_out.seriesterminationcontrol;
  par_term_ctrl_out <= term_ctrl_out.parallelterminationcontrol;
  term_ctrl_in.seriesterminationcontrol   <= ser_term_ctrl_in;
  term_ctrl_in.parallelterminationcontrol <= par_term_ctrl_in;

  ctlr_tech_mosi.burstbegin <= ctlr_burst;
  ctlr_tech_mosi.address    <= RESIZE_MEM_CTLR_ADDRESS(dvr_cur_addr.chip & dvr_cur_addr.bank & dvr_cur_addr.row(g_ddr.a_w-1 DOWNTO 0) & dvr_cur_addr.column(g_ddr.a_col_w -1 DOWNTO c_ddr3_ctlr_rsl_w));
  ctlr_tech_mosi.wrdata     <= RESIZE_MEM_CTLR_DATA(ctlr_wr_sosi.data);
  ctlr_tech_mosi.rd         <= ctlr_rd_req;
  ctlr_tech_mosi.wr         <= ctlr_wr_req;
  ctlr_tech_mosi.burstsize  <= RESIZE_MEM_CTLR_BURSTSIZE(ctlr_burst_size);

  i_ctlr_rdy          <= ctlr_tech_miso.waitrequest_n;
  ctlr_rd_sosi.valid  <= ctlr_tech_miso.rdval;
  ctlr_rd_sosi.data   <= RESIZE_DP_DATA(ctlr_tech_miso.rddata);
  i_ctlr_init_done    <= ctlr_tech_miso.done;
  local_cal_success   <= ctlr_tech_miso.cal_ok;
  local_cal_fail      <= ctlr_tech_miso.cal_fail;


  u_tech_ddr : ENTITY tech_ddr_lib.tech_ddr
  GENERIC MAP (
    g_technology    => c_tech_stratixiv,
    g_tech_ddr      => c_tech_ddr3_4g_800m_master
  )
  PORT MAP (
    -- PLL reference clock
    ref_clk         => ctlr_ref_clk,
    ref_rst         => ctlr_rst,

    -- Controller user interface
    ctlr_gen_clk    => i_ctlr_gen_clk,
    ctlr_gen_rst    => i_ctlr_gen_rst,
    ctlr_gen_clk_2x => i_ctlr_gen_clk_2x,
    ctlr_gen_rst_2x => ctlr_gen_rst_2x,

    ctlr_mosi       => ctlr_tech_mosi,
    ctlr_miso       => ctlr_tech_miso,

    term_ctrl_out   => term_ctrl_out,
    term_ctrl_in    => term_ctrl_in,

    -- DDR3 PHY interface
    phy3_in         => phy_in,
    phy3_io         => phy_io,
    phy3_ou         => phy_ou
  );  

END str;

