
-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
--
-- Purpose:  Testbench for the seq_ddr3 unit 
--           To be used in conjunction with python testscript: ../python/tc_seq_ddr3.py
--
--
-- Usage:
--   > as 8
--   > run -all
--   > run python script in separate terminal: "python tc_seq_ddr3.py --unb 0 --fn 0 --sim"
--   > Stop the simulation manually in Modelsim by pressing the stop-button.
--   > Evalute the WAVE window. 

LIBRARY IEEE, tech_ddr_lib, common_lib, mm_lib, diag_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL; 
USE diag_lib.diag_pkg.ALL;  
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE work.ddr3_pkg.ALL;

ENTITY tb_seq_ddr3 IS 
  GENERIC (
    g_wr_chunksize    : POSITIVE := 240;
    g_wr_nof_chunks   : POSITIVE := 1;
    g_rd_chunksize    : POSITIVE := 16;
    g_rd_nof_chunks   : POSITIVE := 15;
    g_gapsize         : NATURAL  := 16;
    g_nof_blocks      : POSITIVE := 5   
 );
END tb_seq_ddr3;

ARCHITECTURE tb OF tb_seq_ddr3 IS
  
  CONSTANT c_sim                : BOOLEAN := TRUE;

  ----------------------------------------------------------------------------
  -- Clocks and resets
  ----------------------------------------------------------------------------   
  CONSTANT c_mm_clk_period      : TIME := 8 ns; -- 200 ps;
  CONSTANT c_dp_clk_period      : TIME := 5 ns;
  CONSTANT c_dp_pps_period      : NATURAL := 64;

  SIGNAL dp_pps                 : STD_LOGIC;

  SIGNAL mm_rst                 : STD_LOGIC := '1';
  SIGNAL mm_clk                 : STD_LOGIC := '0';

  SIGNAL dp_rst                 : STD_LOGIC;
  SIGNAL dp_clk                 : STD_LOGIC := '0';

  ----------------------------------------------------------------------------
  -- MM buses
  ----------------------------------------------------------------------------                                         
  -- TB
  SIGNAL reg_diag_bg_mosi          : t_mem_mosi;
  SIGNAL reg_diag_bg_miso          : t_mem_miso;
                                
  SIGNAL ram_diag_bg_mosi          : t_mem_mosi;
  SIGNAL ram_diag_bg_miso          : t_mem_miso;
                                
  SIGNAL ram_diag_data_buf_mosi : t_mem_mosi;
  SIGNAL ram_diag_data_buf_miso : t_mem_miso;
  
  SIGNAL reg_diag_data_buf_mosi : t_mem_mosi;
  SIGNAL reg_diag_data_buf_miso : t_mem_miso;

  -- DUT
  SIGNAL reg_ddr3_mosi             : t_mem_mosi := c_mem_mosi_rst;   
  SIGNAL reg_ddr3_miso             : t_mem_miso := c_mem_miso_rst;   
 
  ----------------------------------------------------------------------------
  -- Component declaration of mm_file
  ----------------------------------------------------------------------------
  COMPONENT mm_file
  GENERIC(
    g_file_prefix       : STRING;    
    g_update_on_change  : BOOLEAN := FALSE
  );
  PORT (
    mm_rst        : IN  STD_LOGIC;
    mm_clk        : IN  STD_LOGIC;
    mm_master_out : OUT t_mem_mosi;
    mm_master_in  : IN  t_mem_miso 
  );
  END COMPONENT;   
  
  -- Compose the Constants for the DUT   
  CONSTANT c_ddr3_seq_conf           : t_ddr3_seq := (g_wr_chunksize, 
                                                      g_wr_nof_chunks,
                                                      g_rd_chunksize, 
                                                      g_rd_nof_chunks,
                                                      g_gapsize,      
                                                      g_nof_blocks);   
  
  CONSTANT c_blocksize               : POSITIVE := g_wr_nof_chunks * g_wr_chunksize;  
  CONSTANT c_page_size               : POSITIVE := c_blocksize * g_nof_blocks;
  
  CONSTANT c_ddr                     : t_c_ddr3_phy := c_ddr3_phy_4g;
  CONSTANT c_mts                     : NATURAL := 1066;--1066; --800
  CONSTANT c_phy                     : NATURAL := 1;
  CONSTANT c_data_w                  : NATURAL := 64; 

  CONSTANT c_ctrl_ref_clk_period     : TIME  := 5000 ps; 

  -- Custom definitions of constants
  CONSTANT c_bg_block_len           : NATURAL  := c_page_size;
  CONSTANT c_db_block_len           : NATURAL  := c_page_size;
 
  -- Configuration of the block generator:
  CONSTANT c_bg_nof_output_streams  : POSITIVE := 1;    
  CONSTANT c_bg_buf_dat_w           : POSITIVE := c_nof_complex*8;
  CONSTANT c_bg_buf_adr_w           : POSITIVE := ceil_log2(c_bg_block_len);
  CONSTANT c_bg_data_file_prefix    : STRING   := "../../../src/hex/tb_bg_dat";
  CONSTANT c_bg_data_file_index_arr : t_nat_natural_arr := array_init(0, 1, 1);
  
  -- Configuration of the databuffers:
  CONSTANT c_db_nof_streams         : POSITIVE := 1;
  CONSTANT c_db_data_w              : POSITIVE := c_diag_db_max_data_w;
  CONSTANT c_db_buf_nof_data        : POSITIVE := c_db_block_len;
  CONSTANT c_db_buf_use_sync        : BOOLEAN  := FALSE;
  CONSTANT c_db_data_type           : t_diag_data_type_enum := e_data;
  

  SIGNAL bg_siso_arr                : t_dp_siso_arr(1 -1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  SIGNAL bg_sosi_arr                : t_dp_sosi_arr(1 -1 DOWNTO 0);

  SIGNAL out_sosi_arr               : t_dp_sosi_arr(1-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL out_siso_arr               : t_dp_siso_arr(1-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  
  -- Siganls to connect the memory driver with the mm register interface
  SIGNAL ctlr_ref_clk         : STD_LOGIC    := '0';
  SIGNAL ctlr_rst             : STD_LOGIC    := '1';
  SIGNAL ctlr_gen_clk         : STD_LOGIC;
  SIGNAL ctlr_gen_rst         : STD_LOGIC;

  SIGNAL ctlr_rdy             : STD_LOGIC;
  SIGNAL ctlr_init_done       : STD_LOGIC;

  SIGNAL dvr_start_addr       : t_ddr3_addr; 
  SIGNAL dvr_end_addr         : t_ddr3_addr;

  SIGNAL dvr_en               : STD_LOGIC;
  SIGNAL dvr_wr_not_rd        : STD_LOGIC;
  SIGNAL dvr_done             : STD_LOGIC;
  
  
  -- Signals to interface with the DDR3 memory model.
  SIGNAL phy_in               : t_tech_ddr3_phy_in;
  SIGNAL phy_io               : t_tech_ddr3_phy_io;
  SIGNAL phy_ou               : t_tech_ddr3_phy_ou;   
  
  SIGNAL ras_n : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL cas_n : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL we_n  : STD_LOGIC_VECTOR(0 DOWNTO 0);


BEGIN

  ----------------------------------------------------------------------------
  -- Clock and reset generation
  ----------------------------------------------------------------------------
  mm_clk <= NOT mm_clk AFTER c_mm_clk_period/2;
  mm_rst <= '1', '0' AFTER c_mm_clk_period*5;

  dp_clk <= NOT dp_clk AFTER c_dp_clk_period/2;
  dp_rst <= '1', '0' AFTER c_dp_clk_period*5;  
  
  ctlr_ref_clk <= NOT ctlr_ref_clk AFTER c_ctrl_ref_clk_period/2;
  ctlr_rst     <= '1', '0' AFTER c_ctrl_ref_clk_period*5;  

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(1, c_dp_pps_period, '1', dp_clk, dp_pps);

   ----------------------------------------------------------------------------
  -- Procedure that polls a sim control file that can be used to e.g. get
  -- the simulation time in ns
  ----------------------------------------------------------------------------
  mmf_poll_sim_ctrl_file(c_mmf_unb_file_path & "sim.ctrl", c_mmf_unb_file_path & "sim.stat");
 
  ----------------------------------------------------------------------------
  -- MM buses  
  ----------------------------------------------------------------------------
  -- TB
  u_mm_file_reg_diag_bg       : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "REG_DIAG_BG")
                                        PORT MAP(mm_rst, mm_clk, reg_diag_bg_mosi, reg_diag_bg_miso);

  u_mm_file_ram_diag_bg       : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_DIAG_BG")
                                        PORT MAP(mm_rst, mm_clk, ram_diag_bg_mosi, ram_diag_bg_miso);
  
  u_mm_file_ram_diag_data_buf : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_DIAG_DATA_BUFFER")
                                        PORT MAP(mm_rst, mm_clk, ram_diag_data_buf_mosi, ram_diag_data_buf_miso);

  u_mm_file_reg_diag_data_buf : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "REG_DIAG_DATA_BUFFER")
                                           PORT MAP(mm_rst, mm_clk, reg_diag_data_buf_mosi, reg_diag_data_buf_miso);

  ----------------------------------------------------------------------------
  -- Source: block generator
  ---------------------------------------------------------------------------- 
  u_bg : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP(
    g_nof_streams        => c_bg_nof_output_streams,
    g_buf_dat_w          => c_bg_buf_dat_w,
    g_buf_addr_w         => c_bg_buf_adr_w,            
    g_file_index_arr     => c_bg_data_file_index_arr,
    g_file_name_prefix   => c_bg_data_file_prefix
  )
  PORT MAP(
    -- System
    mm_rst               => mm_rst,
    mm_clk               => mm_clk,
    dp_rst               => dp_rst,
    dp_clk               => dp_clk,
    en_sync              => dp_pps,
    -- MM interface      
    reg_bg_ctrl_mosi     => reg_diag_bg_mosi,
    reg_bg_ctrl_miso     => reg_diag_bg_miso,
    ram_bg_data_mosi     => ram_diag_bg_mosi,
    ram_bg_data_miso     => ram_diag_bg_miso,
    -- ST interface      
    out_siso_arr         => bg_siso_arr,
    out_sosi_arr         => bg_sosi_arr
  );
   
  ----------------------------------------------------------------------------
  -- DUT: Device Under Test
  ---------------------------------------------------------------------------- 
  u_dut: ENTITY work.seq_ddr3
  GENERIC MAP(
    g_ddr      => c_ddr,
    g_mts      => c_mts, 
    g_phy      => c_phy,
    g_data_w   => c_data_w,
    g_ddr3_seq => c_ddr3_seq_conf
  )                          
  PORT MAP (        
    ctlr_ref_clk       => ctlr_ref_clk,
    ctlr_rst           => ctlr_rst,

    ctlr_gen_clk       => open,
    ctlr_gen_rst       => open,      

    wr_clk             => dp_clk,
    wr_rst             => dp_rst,

    wr_sosi            => bg_sosi_arr(0), 
    wr_siso            => bg_siso_arr(0),
    
    flush_ena          => '0',
  
    rd_sosi            => out_sosi_arr(0),
    rd_siso            => out_siso_arr(0),

    rd_clk             => dp_clk,
    rd_rst             => dp_rst,

    ddr3_in            => phy_in,
    ddr3_io            => phy_io,     
    ddr3_ou            => phy_ou
  );

  gen_u_800_model : IF c_phy > 0 GENERATE
    u_4gb_800_ddr3_model : component alt_mem_if_ddr3_mem_model_top_ddr3_mem_if_dm_pins_en_mem_if_dqsn_en
		generic map (
			MEM_IF_ADDR_WIDTH            => 15,
			MEM_IF_ROW_ADDR_WIDTH        => 15,
			MEM_IF_COL_ADDR_WIDTH        => 10,
			MEM_IF_CS_PER_RANK           => 1,
			MEM_IF_CONTROL_WIDTH         => 1,
			MEM_IF_DQS_WIDTH             => 8,
			MEM_IF_CS_WIDTH              => 2,
			MEM_IF_BANKADDR_WIDTH        => 3,
			MEM_IF_DQ_WIDTH              => 64,
			MEM_IF_CK_WIDTH              => 2,
			MEM_IF_CLK_EN_WIDTH          => 2,
			DEVICE_WIDTH                 => 1,
			MEM_TRCD                     => 6,
			MEM_TRTP                     => 3,
			MEM_DQS_TO_CLK_CAPTURE_DELAY => 100,
			MEM_CLK_TO_DQS_CAPTURE_DELAY => 100000,
			MEM_IF_ODT_WIDTH             => 2,
			MEM_MIRROR_ADDRESSING_DEC    => 0,
			MEM_REGDIMM_ENABLED          => false,
			DEVICE_DEPTH                 => 1,
			MEM_GUARANTEED_WRITE_INIT    => false,
			MEM_VERBOSE                  => true,
			MEM_INIT_EN                  => false,
			MEM_INIT_FILE                => "",
			DAT_DATA_WIDTH               => 32
		)
		port map (
			mem_a       => phy_ou.a(c_ddr.a_w-1 DOWNTO 0),      
			mem_ba      => phy_ou.ba,       
			mem_ck      => phy_ou.ck,      
			mem_ck_n    => phy_ou.ck_n,    
			mem_cke     => phy_ou.cke(c_ddr.cs_w-1 DOWNTO 0),   
			mem_cs_n    => phy_ou.cs_n(c_ddr.cs_w-1 DOWNTO 0),  
			mem_dm      => phy_ou.dm,       
			mem_ras_n   => ras_n,           
			mem_cas_n   => cas_n,           
			mem_we_n    => we_n,            
			mem_reset_n => phy_ou.reset_n, 
			mem_dq      => phy_io.dq,       
			mem_dqs     => phy_io.dqs,      
			mem_dqs_n   => phy_io.dqs_n,    
			mem_odt     => phy_ou.odt       
		);               
		
		ras_n(0) <= phy_ou.ras_n;
		cas_n(0) <= phy_ou.cas_n;
		we_n(0)  <= phy_ou.we_n;

  END GENERATE;
  
  ----------------------------------------------------------------------------
  -- Sink: data buffer real 
  ---------------------------------------------------------------------------- 
  u_data_buf : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams     => c_db_nof_streams, 
    g_data_type       => c_db_data_type,   
    g_data_w          => c_db_data_w,      
    g_buf_nof_data    => c_db_buf_nof_data,
    g_buf_use_sync    => c_db_buf_use_sync
  )
  PORT MAP (
    -- System
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,
     -- MM interface
    ram_data_buf_mosi => ram_diag_data_buf_mosi,
    ram_data_buf_miso => ram_diag_data_buf_miso,
    reg_data_buf_mosi => reg_diag_data_buf_mosi,
    reg_data_buf_miso => reg_diag_data_buf_miso,
    -- ST interface
    in_sync           => OPEN,
    in_sosi_arr       => out_sosi_arr
  );


END tb;
