--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose:
--   Drop-in simulation model for tech_transceiver_gx.vhd.
-- Description:
--   A fast serdes for simulation within tr_nonbonded.
-- Remarks:
--   None

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

ENTITY sim_transceiver_gx IS 
  GENERIC(
    g_data_w     : NATURAL;
    g_nof_gx     : NATURAL;
    g_mbps       : NATURAL;  -- line rate in Mbps so including 10/8 encoding
    g_tx         : BOOLEAN;
    g_rx         : BOOLEAN
  );      
  PORT(
    tb_end       : IN  STD_LOGIC := '0';  -- in simulation stop internal clocks when tb_end='1' to support 'run -all'
    
    tr_clk       : IN  STD_LOGIC;

    rx_clk       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    rx_rst       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    rx_sosi_arr  : OUT t_dp_sosi_arr(g_nof_gx-1 DOWNTO 0);
    rx_siso_arr  : IN  t_dp_siso_arr(g_nof_gx-1 DOWNTO 0);

    tx_clk       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_rst       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_sosi_arr  : IN  t_dp_sosi_arr(g_nof_gx-1 DOWNTO 0);
    tx_siso_arr  : OUT t_dp_siso_arr(g_nof_gx-1 DOWNTO 0);
   
    --Serial I/O
    rx_datain    : IN  STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_dataout   : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0)
  );

END sim_transceiver_gx;

ARCHITECTURE str OF sim_transceiver_gx IS

  CONSTANT c_tr_clk_period  : TIME := 1 us * g_data_w * 10 / 8 / g_mbps;

  TYPE t_ctrl_2arr IS ARRAY(g_nof_gx-1 DOWNTO 0) OF STD_LOGIC_VECTOR(g_data_w/c_byte_w-1 DOWNTO 0);
  
  SIGNAL tr_rst      : STD_LOGIC;

  SIGNAL tx_ready    : STD_LOGIC;

  SIGNAL tx_in_ctrl  : t_ctrl_2arr;
  SIGNAL rx_out_ctrl : t_ctrl_2arr;

BEGIN

  u_areset_tr_rst : ENTITY common_lib.common_areset
  GENERIC MAP(
    g_rst_level => '1'
  )
  PORT MAP(
    clk     => tr_clk,
    in_rst  => '0',
    out_rst => tr_rst
  );  

  tx_clk <= (OTHERS=>tr_clk);
  tx_rst <= (OTHERS=>tr_rst);
  rx_clk <= (OTHERS=>tr_clk);
  rx_rst <= (OTHERS=>tr_rst);

  p_tx_rdy : PROCESS
    VARIABLE v_tx_clk_cnt : NATURAL := 0;
  BEGIN
    tx_ready <= '0';
    WAIT FOR 1084.8 ns;              -- Time until tx_rst (in tr_clk domain) is deasserted: tx init done
    WHILE v_tx_clk_cnt < 1000 LOOP   -- 1000 tx clk cycles until alignment is completed
      WAIT UNTIL rising_edge(tr_clk);
      v_tx_clk_cnt := v_tx_clk_cnt + 1;
      tx_ready <= '1';
    END LOOP;
    WAIT;
  END PROCESS;

  gen_sim: FOR i IN 0 to g_nof_gx-1 GENERATE
    gen_tx : IF g_tx = TRUE GENERATE
      tx_siso_arr(i).ready <= tx_ready;
      tx_siso_arr(i).xon   <= tx_ready;

      tx_in_ctrl(i) <= (OTHERS=>tx_sosi_arr(i).valid);

      u_ser: ENTITY work.sim_transceiver_serializer
      GENERIC MAP (
        g_data_w         => g_data_w,
        g_tr_clk_period  => c_tr_clk_period
      )
      PORT MAP (
        tb_end        => tb_end,
        
        tr_clk        => tr_clk,  
        tr_rst        => tr_rst,
         
        tx_in_data    => tx_sosi_arr(i).data(g_data_w-1 DOWNTO 0),
        tx_in_ctrl    => tx_in_ctrl(i),
    
        tx_serial_out => tx_dataout(i)
      );
    END GENERATE;

    gen_rx : IF g_rx = TRUE GENERATE   
      u_des: ENTITY work.sim_transceiver_deserializer 
      GENERIC MAP (
        g_data_w         => g_data_w,
        g_tr_clk_period  => c_tr_clk_period
      )
      PORT MAP (
        tb_end        => tb_end,
        
        tr_clk        => tr_clk,  
        tr_rst        => tr_rst,
         
        rx_out_data   => rx_sosi_arr(i).data(g_data_w-1 DOWNTO 0),
        rx_out_ctrl   => rx_out_ctrl(i),
    
        rx_serial_in  => rx_datain(i)
      );
    END GENERATE;

    rx_sosi_arr(i).valid <= andv(rx_out_ctrl(i));

  END GENERATE;

END str;

